{-# OPTIONS -fglasgow-exts  #-}
-----------------------------------------------------------------------------
-- |
-- Stability   :  experimental
-- Portability :  experimental
--
-- The Stack Algebra example.
--
-----------------------------------------------------------------------------

module Camila.Examples.StackAlg where

import Camila.Prelude hiding (empty)

-----------------------------------------------------------------------------
-- * Types

-- | Type definition. 
--   A new type is used rather than a type synonym, because we want
--   to contrain the data type.
newtype Stack = Stack { theStack :: [Int] } deriving Show

-- Datatype invariant
instance CData Stack where
   inv s = all odd (theStack s)

-- | The initialization function is monadic,
--   because it checks the invariant.
initStack:: CamilaMonad m => m Stack
initStack = returnInv $ Stack []

-----------------------------------------------------------------------------
-- * Functions

-- | The empty predicate is not monadic, since it does not need
--   to check any invariants or conditions.
empty :: Stack -> Bool
empty s = theStack s == []

-- | The push function is monadic for two reasons. It has a pre
--   condition, and it needs to check the invariant of the returned
--   stack.
push :: CamilaMonad m => Int -> Stack -> m Stack
push n s = do
  pre (odd n)
  returnInv $ Stack (n: theStack s)

push' :: Int -> Stack -> Stack
push' n s = Stack (n: theStack s)
pop' :: Stack -> Stack
pop' s = Stack (tail (theStack s))
top' :: Stack -> Int
top' (Stack (n:s)) = n
-- | The pop function is monadic for two reasons. It has a pre
--   condition, and it needs to check the invariant of the returned
--   stack.
pop :: CamilaMonad m => Stack -> m Stack
pop s = do 
  pre (not $ empty s)
  returnInv $ Stack $ tail $ theStack s
  
-- | The top function is monadic for two reasons. It has a pre
--   condition, and it has a post condition.
top :: CamilaMonad m => Stack -> m Int
top s = do 
  pre (not $ empty s)
  let result = head $ theStack s
  post (odd result)
  post ( ((result : (tail $ theStack s)) == theStack s) 
         && (odd result) )
  return result
  
-- | The top function is monadic for two reasons. It has a pre
--   condition, and it needs to check the invariant of the returned
--   stack.
add :: CamilaMonad m => Stack -> m Stack
add (Stack s) = do 
  pre (length s >= 2)
  returnInv $ Stack $ [s!!0 + s!!1] ++ tail (tail s)

-----------------------------------------------------------------------------
-- * Tests

-- | Demonstrate error that will be fatal if it goes unchecked.
testTopEmptyStack :: CamilaMonad m => m Int
testTopEmptyStack = do
  s <- initStack
  n <- top s
  if n==n then return n else return n    -- Make strict
  
-- | Demonstrate error that will go unnoticed if unchecked.
testPushEvenOnStack :: CamilaMonad m => m Int
testPushEvenOnStack = do
  s <- initStack
  s <- push 0 s
  n <- top s
  if n==n then return n else return n    -- Make strict
  
-----------------------------------------------------------------------------

