{-# OPTIONS -fglasgow-exts -fallow-overlapping-instances -fallow-undecidable-instances -fno-monomorphism-restriction #-}

-----------------------------------------------------------------------------
-- |
-- Maintainer  :  Joost Visser
-- Stability   :  experimental
-- Portability :  experimental
--
-- This module captures monadic support for checking of pre-conditions,
-- post-conditions, and datatype invariants.
--
-----------------------------------------------------------------------------

module Camila.Monad where

import Camila.DTInvariants

import Control.Monad.Error
import Control.Monad.Identity
import Control.Monad.State
import Control.Monad
import IO

-----------------------------------------------------------------------------
-- * Camila monad

-- | A CamilaMonad is a monad with additional operations that
--   support condition and invariant checking.
class Monad m => CamilaMonad m where
  -- | Check precondition
  pre :: Bool -> m ()
  -- | Check postcondition
  post :: Bool -> m ()
  -- | Check invariant before returning data in monad
  returnInv :: CData a => a -> m a

-- | Data type to represent different possible assertion violations.
data CamilaViolation
  = PreConditionViolation
  | PostConditionViolation
  | InvariantViolation
  deriving (Show,Eq,Read)
instance Error CamilaViolation
  
-----------------------------------------------------------------------------
-- * Instances for various modes of pure operation

-- ** Safe operation, silent. 
--
-- When a violation is detected, mzero is returned.
-- No specific error messages.
-- Instances: Maybe, [].

instance CamilaMonad Maybe where
  pre p = guard p
  post p = guard p
  returnInv a = guard (inv a) >> return a
instance CamilaMonad [] where
  pre p = guard p
  post p = guard p
  returnInv a = guard (inv a) >> return a

testCamila_SafeSilent_Maybe :: Maybe String
testCamila_SafeSilent_Maybe = do
  pre False
  return "Reached the finish"

testCamila_SafeSilent_List :: [] String
testCamila_SafeSilent_List = do
  pre False
  return "Reached the finish"

-- ** Safe operation, messages. 
--
-- When a violation is detected, a specific
-- error is returned, embedded in Either.Left.
instance CamilaMonad (Either CamilaViolation) where
  pre p = if p then return () else Left PreConditionViolation
  post p = if p then return () else Left PostConditionViolation
  returnInv a = if (inv a) then return a else Left InvariantViolation

testCamila_SafeMessage_Either :: Either CamilaViolation String
testCamila_SafeMessage_Either = do
  pre False
  return "Reached the finish"

-- ** Unsafe operation, silent.
-- 
-- When an invariant or condition is violated, a
-- runtime error MAY occur. This is the default
-- behaviour for all monads that do not have
-- a specific instance, e.g. Identity.
instance Monad m => CamilaMonad m where
  pre p = return ()
  post p = return ()
  returnInv a = return a

testCamila_UnsafeSilent_Identity :: Identity String
testCamila_UnsafeSilent_Identity = do
  pre False
  return "Reached the finish"

-- ** Checked operation, unsafe, messages.
--
-- | When a violation is detected, a messages is emitted on
-- standard error, but operation continues. A runtime error
-- MAY occur.
instance CamilaMonad IO where
  pre p = unless p $ hPutStrLn stderr $ show PreConditionViolation
  post p = unless p $ hPutStrLn stderr $ show PostConditionViolation
  returnInv a = do
    unless (inv a) $ hPutStrLn stderr $ show InvariantViolation
    return a

testCamila_UnsafeMessage_IO :: IO ()
testCamila_UnsafeMessage_IO = do
  pre False
  putStrLn "Reached the finish"

-----------------------------------------------------------------------------
-- * Convenience functions

-- | Function application inside a monad, i.e. 
--   functorial map for monads.
(=<) :: Monad m => (a -> b) -> m a -> m b
f =< ma = ma >>= (return . f)

-- | Check precondition (monadic variation)
preM :: CamilaMonad m => m Bool -> m ()
preM mp = mp >>= pre

-----------------------------------------------------------------------------
-- * Convenience functions for dealing with state.

-- | Check invariant before putting data in state.
putInv :: (MonadState a m, CData a, CamilaMonad m) => a -> m ()
putInv s = pre (inv s) >> put s

-- | Check invariant while modifying data in state.
modifyInv :: (MonadState s m, CData s, CamilaMonad m) => (s -> s) -> m ()
modifyInv f = do
  s <- get
  putInv (f s)

-- | Check invariant on data in the current state.
preState :: (CamilaMonad m, MonadState s m) => (s -> Bool) -> m ()
preState p = do
  state <- get
  pre (p state)

-----------------------------------------------------------------------------


-----------------------------------------------------------------------------
