{-# OPTIONS -fglasgow-exts -fallow-overlapping-instances -fallow-undecidable-instances #-}

-----------------------------------------------------------------------------
-- |
-- Maintainer  :  Joost Visser
-- Stability   :  experimental
-- Portability :  experimental
--
-- This module captures monadic support for checking of pre-conditions,
-- post-conditions, and datatype invariants.
--
-----------------------------------------------------------------------------

module Camila.MonadT where

import Camila.Monad
import Camila.DTInvariants
import Control.Monad.Identity
import Control.Monad.Error
import Control.Monad.Writer
import Control.Monad.State
import Data.IORef
import IO

-----------------------------------------------------------------------------
-- * Monad transformer

-- | The CamilaT monad transformer adds checking effects to a given
--   base monad. Which effects exactly depends on the phantom mode
--   argument.
data CamilaT mode m a = CamilaT {runCamilaT :: m a} deriving Show
instance Monad m => Monad (CamilaT mode m) where
  return a = CamilaT $ return a
  ma >>= f = CamilaT (runCamilaT ma >>= runCamilaT . f)

-----------------------------------------------------------------------------
-- * Mode types

-- | Free fall mode: no checking whatsoever.
data FreeFall

-- | Warn mode: perform checks, emit warning on violation, keep running.
data Warn

-- | Fail mode: perform checks, fatal error on violation.
data Fail

-- | Error mode: perform checks, monadic error on violation.
--   Instead of this error type constructor, the error type itself can
--   be supplied directly.
type ErrorMode error = error

-----------------------------------------------------------------------------
-- * Instances for various modes of monadic operation

-- ** Free fall (works for any monad).

instance Monad m => CamilaMonad (CamilaT FreeFall m) where
  pre p = return ()
  post p = return ()
  returnInv a = return a

-- | Enforce free-falling mode.
freeFall :: CamilaT FreeFall m a -> CamilaT FreeFall m a
freeFall = id
 
-- ** Warn (works for monads that can write, e.g. IO, Writer)

-- instance CamilaMonad (CamilaT Warn IO) where
instance MonadIO m => CamilaMonad (CamilaT Warn m) where
  pre p = unless p $ liftIO $ hPutStrLn stderr $ show PreConditionViolation
  post p = unless p $ liftIO $ hPutStrLn stderr $ show PostConditionViolation
  returnInv a = do
    unless (inv a) $ liftIO $ hPutStrLn stderr $ show InvariantViolation
    return a    
instance MonadIO m => MonadIO (CamilaT mode m) where
  liftIO = CamilaT . liftIO

instance CamilaMonad (CamilaT Warn (Writer [CamilaViolation])) where
  pre p = if p then return () else tell [PreConditionViolation]
  post p = if p then return () else tell [PostConditionViolation]
  returnInv a = if (inv a) then return a else do 
    tell [InvariantViolation]
    return a
instance Monad m => CamilaMonad (CamilaT Warn (WriterT [CamilaViolation] m)) where
  pre p = if p then return () else tell [PreConditionViolation]
  post p = if p then return () else tell [PostConditionViolation]
  returnInv a = if (inv a) then return a else do 
    tell [InvariantViolation]
    return a
instance MonadWriter w m => MonadWriter w (CamilaT mode m) where
  tell w = CamilaT $ tell w
  listen ma = CamilaT $ listen $ runCamilaT ma
  pass maf = CamilaT $ pass $ runCamilaT maf

-- | Enforce warning mode.
warn :: CamilaT Warn m a -> CamilaT Warn m a 
warn = id

-- ** Fail (works for any monad)

instance Monad m => CamilaMonad (CamilaT Fail m) where
  pre p = if p then return () else fail $ show PreConditionViolation
  post p = if p then return () else fail $ show PostConditionViolation
  returnInv a = if (inv a) then return a else fail $ show InvariantViolation

-- | Enforce failing mode.
fatal :: CamilaT Fail m a -> CamilaT Fail m a 
fatal = id

-- ** Error (works for any error monad, e.g IO, Either)

instance MonadError CamilaViolation m 
      => CamilaMonad (CamilaT CamilaViolation m) where
  pre p = if p then return () else throwError PreConditionViolation
  post p = if p then return () else throwError PostConditionViolation
  returnInv a = if (inv a) then return a else throwError InvariantViolation
instance MonadError e m => MonadError e (CamilaT e m) where
  throwError e = CamilaT $ throwError e
  catchError ma f = CamilaT $ catchError (runCamilaT ma) (runCamilaT . f)

instance MonadError IOError m 
      => CamilaMonad (CamilaT IOError m) where
  pre p = if p then return () else throwError $ userError $ show PreConditionViolation
  post p = if p then return () else throwError $ userError $ show PostConditionViolation
  returnInv a = if (inv a) then return a else throwError $ userError $ show InvariantViolation

-- | Enforce error mode
errorMode :: MonadError e m => CamilaT e m a -> CamilaT e m a
errorMode = id

-- | Enforce use of CamilaViolation in error mode
camilaViolation :: MonadError CamilaViolation m 
                => CamilaT CamilaViolation m a -> CamilaT CamilaViolation m a
camilaViolation = id

-- | Class MonadRef 

class MonadRef m where
    readRef :: IORef a -> m a
    modifyRef :: IORef a -> (a -> a) -> m ()
    writeRef :: IORef a -> a -> m ()
    newRef :: a -> m (IORef a)

instance MonadIO m => MonadRef m where
    readRef = liftIO . readIORef
    modifyRef r f = liftIO $ modifyIORef r f
    writeRef r a = liftIO $ writeIORef r a
    newRef = liftIO . newIORef
    
-- | instance of MonadFix for CamilaT mode m 

instance MonadFix m => MonadFix (CamilaT mode m) where
    mfix f =  CamilaT $ mfix (\g -> runCamilaT $ f g)
-----------------------------------------------------------------------------
-- * Test functions to demonstrate different modes of operation

testFreeFall_IO :: IO ()
testFreeFall_IO = runCamilaT $ freeFall testProgram

testWarn_IO :: IO ()
testWarn_IO = runCamilaT $ warn testProgram

testFail_IO :: IO ()
testFail_IO = runCamilaT $ fatal testProgram

testError_IO :: IO ()
testError_IO = runCamilaT $ errorMode testProgram

testProgram :: (CamilaMonad m, MonadIO m) => m ()
testProgram = do
  pre False
  liftIO $ putStrLn "So far so good"
  return ()

-----------------------------------------------------------------------------
-- * Instances to interact with state monad transformer

-- Propagation of the state effect through the Camila monad transformer.  
instance MonadState s m => MonadState s (CamilaT mode m) where
  get = CamilaT get
  put a = CamilaT $ put a
  
-- Propagation of the camila effects through the state monad transformer.
instance CamilaMonad m => CamilaMonad (StateT s m) where
  pre p = StateT $ \s -> pre p >> return ((),s)
  post p = StateT $ \s -> post p >> return ((),s)
  returnInv a = StateT $ \s -> returnInv a >>= \a -> return (a,s)

-----------------------------------------------------------------------------

