-----------------------------------------------------------------------------
-- |
-- Module      :  Camila.State
-- Copyright   :  (c) Alexandra Mendes, João Ferreira
-- License     :  GPL
--
-- Maintainer  :  João Ferreira, Alexandra Mendes
-- Stability   :  experimental
-- Portability :  experimental
--
-- Module that defines State structure. It uses State Monad.
-- 
-----------------------------------------------------------------------------

module Camila.State (InterpState,
                     Instances,
                     execute,
                     preOn,
                     preOff,
                     dtinvOn,
                     dtinvOff,
                     checkDTInv,
                     checkPre,
                     instances,
                     CamilaState(CamilaState),camilaInitial,getTypeAsString,
                     addToInstances,getInstance,setState,
                     liftE,runInstance,ST,
                     create,getId) where

import Data.Map as Map
import Data.FiniteMapExtras
import Camila.Error
import Control.Monad.Error
import Control.Monad.State
import AltData.Dynamic
import AltData.Typeable
import Maybe

-----------------------------------------------------------------------------


-- | The type of the Camila interpreter state. It holds a map from
--   instance references to instances.
data CamilaState = CamilaState { checkDTInv :: Bool, -- ^ Not used, currently
                                 checkPre   :: Bool, -- ^ Not used, currently
                                 instances  :: Instances
                               }
                               
-- | The type of an instances store.
type Instances = Map.Map String (String,Dynamic)                                


-- | The initial state of the interpreter.
camilaInitial :: CamilaState
camilaInitial = CamilaState { checkDTInv = False, checkPre = False, instances = Map.empty }


-- | Monad that combines all needed interpreter effects.
type ST ivars a = ErrorT CamilaError' (StateT ivars IO) a

-- | Monad instantiated with the interpreter state.
type InterpState a = ST CamilaState a 
               -- = ErrorT CamilaError' (StateT CamilaState IO) a 

--camilaState :: InterpState a -> CamilaState 
--camilaState = 

-- | Execute a computation that delivers a showable type.
execute :: (Show a) => ST CamilaState a -> IO ()
execute p = do
  result <- evalStateT (runErrorT p) camilaInitial
  putStrLn $ show result

-----------------------------------------------------------------------------
-- * Functions that change the state 

-- | Turn on data type invariant checking.
dtinvOn:: InterpState ()
dtinvOn = modify (\s -> s {checkDTInv = True})

-- | Turn off data type invariant checking.
dtinvOff:: InterpState ()
dtinvOff = modify (\s -> s {checkDTInv = False})

-- | Turn on precondition checking.
preOn :: InterpState ()
preOn = modify (\s -> s {checkPre = True})

-- | Turn off precondition checking.
preOff :: InterpState ()
preOff = modify (\s -> s {checkPre = False})

-- | Create a new instance.
create :: (Typeable a) 
       => String    -- ^ Instance identifier 
       -> String    -- ^ Class name
       -> a         -- ^ Initial state of the instance
       -> InterpState Instances
create id t obj = do modify (\s -> s { instances = Map.insert id (t,(toDyn obj)) (instances s) })
                     insts <- gets instances
                     return insts

-- | Destroy an existing instance.
destroy :: String  -- ^ Instance identifier. 
        -> InterpState ()
destroy id = modify (\s -> s { instances = Map.delete id (instances s) })

-- | Retrieve an instance from a given store.
getInstance :: String            -- ^ Instance identifier.
            -> Instances         -- ^ The instance store.
            -> CamilaError Dynamic
getInstance id map | isJust $ m map id = Right (snd $ fromJust $ m map id) -- change to let, in order to evaluate function m only once
                   | otherwise = fail "Error getting instance..."

-- | Retrieve an instance from the store.
getId :: String         -- ^ Instance identifier.
      -> InterpState Dynamic
getId id = do insts <- gets instances
              liftE $ getInstance id insts

-- | Returns the class name of an instance.
getTypeAsString :: String   -- ^ Instance identifier.
                -> InterpState String
getTypeAsString id = do insts <- gets instances
                        if (isJust $ m insts id) then return $ fst $ fromJust $ (m insts id) 
                                                 else fail "Error getting instance..."

-- | Add an instance to the current store.
setState :: String              -- ^ Instance identifier.
         -> (String, Dynamic)   -- ^ Pair of class name and object.
         -> InterpState ()
setState id obj = modify (\s -> s { instances = Map.insert id obj (instances s) })

-- | Add several instances to the current store.
addToInstances :: Instances -> InterpState ()
addToInstances obj = modify (\s -> s { instances = (instances s) +++ obj})

-- | Lift an Either term into the error monad.
liftE :: (MonadError e (ErrorT e m), Monad m) 
      => Either e a -> ErrorT e m a
liftE f = catchError (ErrorT $ return f) throwError
--liftE f = (ErrorT $ return f)

-----------------------------------------------------------------------------
-- * Instances

-- | Perform a given operation on a given initial state.
runInstance :: (Show a) 
            => ST ivars a       -- ^ Computation. 
            -> ivars            -- ^ Initial state
            -> IO (CamilaError a)
runInstance p initial = evalStateT (runErrorT p) initial

-----------------------------------------------------------------------------

