-----------------------------------------------------------------------------
-- |
-- Module      :  Data.MapExtras
-- Copyright   :  (c) Alexandra Mendes, Joao Ferreira
-- License     :  GPL
--
-- Maintainer  :  Joao Ferreira, Alexandra Mendes
-- Stability   :  experimental
-- Portability :  experimental
--
-- Extra functions to use with Maps (includes all VDM-SL functions)
-- 
-----------------------------------------------------------------------------


module Data.FiniteMapExtras (
-- * Maps' basic functions
   dom,             -- | VDM: dom m
   rng,             -- | VDM: rng m
   munion,            -- | VDM: munion m1 m2
   munionRel,   
   (+++), override,   -- | VDM: m1 ++ m2
   merge,             -- | VDM: merge ms
   (<:),              -- | VDM: s <: m 
   (<-:),             -- | VDM: s <-: m 
   (>:),              -- | VDM: m :> s
   (>-:),             -- | VDM: m :-> s
   comp,            -- | VDM: m1 comp m2
   (***),             -- | VDM: m ** n 
   inverse,           -- | VDM: inverse m 
   inverse2,          -- | VDM: inverse m 
   m,                 -- | VDM: m(d) 
   injective,

-- * Extra functions
   mkr,fmToSet,
   setOfKeys,
   setOfElts,

-- *  File IO
   readFile_,
   interact_
) where


import Maybe(fromJust)
import qualified Data.Map as Map
import Data.Map(Map)
import Data.Set as Set hiding (map)
import Data.Relation.SetOfPairs (Rel) 

import Text.ReadExtras

infix 5 +++ -- Override
infix 5 *** -- Map iteration
infix 5 <:  -- Domain restrict to
infix 5 <-: -- Domain restrict by
infix 5 >:  -- Range restrict to
infix 5 >-: -- Range restrict by


-- | Yields the domain (the set of keys) of a map.
dom :: Map a b -> Set a
dom = Map.keysSet 


-- Yields the range (the set of information values) of a map.
rng :: (Ord a,Ord b) => Map a b -> Set b
rng = (Set.fromList).(Map.elems)


-- | Yields a map combined by two other maps, such that the resulting map
--   maps the elements of the domain of both maps. The two maps must have
--   disjoint domains.
munion :: (Ord a) => Map a b -> Map a b -> Maybe (Map a b)
munion m1 m2 | Set.null (intersection (dom m1) (dom m2)) = Just (m1 +++ m2)
             | otherwise = Nothing

-- | Yields a relation that has all pairs (key,elt) of the two given maps.
munionRel :: (Ord a,Ord b) => Map a b -> Map a b -> Rel a b
munionRel m1 m2 = (fmToSet m1) `union` (fmToSet m2)

-- | Overrides and merges two maps. It is like munion, except that both maps
--   don't need to be compatible; the values of the second map override the 
--   ones of the first.
(+++) :: (Ord a) => Map a b -> Map a b -> Map a b
(+++) = flip Map.union


-- | Same as (+++).
override :: (Ord a) => Map a b -> Map a b -> Map a b
override = (+++)


-- | Given a set of maps, yields the map that is contructed by merging them all.
--   The maps must be compatible.
merge :: (Ord a) => Set (Map a b) -> Maybe (Map a b)
merge s = foldr maux (Just Map.empty) (Set.elems s)
   where
      maux _ Nothing = Nothing
      maux a (Just c) = munion a c


-- | Given a set and a map, creates the map consisting of the elements whose key is in the set.
--   The set don't need to be a subset of the given map's domain.
(<:) :: (Ord a) => Set a -> Map a b -> Map a b
(<:) s = Map.filterWithKey (\a b -> member a s)  --- MAIS EFICIENTE :-)


-- | Given a set and a map, creates the map consisting of the elements whose key is not in the set.
--   The set don't need to be a subset of the given map's domain.
(<-:) :: (Ord a) => Set a -> Map a b -> Map a b
(<-:) s m = let l=(Set.elems s)
            in delete' l m
           
delete' [] m = m
delete' (a:as) m = delete' as (Map.delete a m)


-- | Given a map and a set, creates the map consisting of the elements whose information value is in the set.
--   The set don't need to be a subset of the given map's range.
(>:) :: (Ord a,Ord b) => Map a b -> Set b -> Map a b
(>:) m s = Map.filterWithKey (\a b -> member b s) m


-- | Given a map and a set, creates the map consisting of the elements whose information value is not in the set.
--   The set don't need to be a subset of the given map's range.
(>-:) :: (Ord a, Ord b) => Map a b -> Set b -> Map a b
(>-:) m s = Map.filterWithKey (\a b -> not (member b s)) m


-- | Given a map and a key, yields the information value associated with that key, which must be in 
--   the domain of m.
m :: (Ord key) => Map key elt -> key -> Maybe elt
m = flip Map.lookup


-- | Given two maps m1 and m2, yields the map that is created by composing m2
--   elements with m1 elements. The resulting map is a map with the same domain as m2.
--   The information value corresponding to a key is the one found by first applying m2 to the
--   key and then applying m1 to the result. rng m2 must be a subset of dom m1.
comp :: (Ord a) => Map a a -> Map a a -> Maybe (Map a a)
comp m1 m2 | intersection (rng m2) (dom m1) == (rng m2) = Just (Map.fromList (caux (Map.toList m1) (Map.toList m2)))
             | otherwise = Nothing
 where
  caux _ [] = []
  caux l2 ((a,b):ps) = let value = lookup b l2
                       in if value /= Nothing then (a, fromJust (lookup b l2)) : caux ps l2
                                              else caux ps l2


-- | Given a map m and a positive integer n, yields the map where m is composed with itself
--   n times. n=0 yields the identity map where each element of dom m is map into itself;
--   n=1 yields m itself. For n>1, the range of m must be a subset of dom m.
(***) :: (Ord a, Num b) => Map a a -> b -> Maybe (Map a a)
(***) m 0 = Just (Map.fromList [(a,a) | a<-Set.elems (dom m)])
(***) m 1 = Just m
(***) m n = comp m (fromJust (m *** (n-1)))


-- Equality
-- VDM: m1 = m2
-- Map module defines an instance of Eq


-- Inequality
-- VDM: m1 <> m2
-- Map module defines an instance of Eq

-- | Given a map m, yields the inverse map of m. m must be a 1-to-1 mapping.
inverse :: (Ord key,Ord elt) => Map key elt -> Maybe (Map elt key)
inverse m | injective m = Just $ Map.fromList.(map swap).Map.toList $ m 
          | otherwise = Nothing
    where swap (a,b) = (b,a)

-- | Given a map m, returns true if m is injective.
injective :: (Ord key,Ord elt) => Map key elt -> Bool
injective m = (size.rng $ m) == (size.dom $ m)


-- | Given a map m, yields the inverse map of m. m must be a 1-to-1 mapping.
--   This is a slightly more efficient version than inverse. 
inverse2 :: (Ord key,Ord elt) => Map key elt -> Maybe (Map elt key)
inverse2 m | injective m = Just $ Map.foldWithKey (\ key elt rest -> Map.insert elt key rest) Map.empty $ m
           | otherwise = Nothing


-- Show and Ord instances for Map

{- Now provided by Data.Map
instance (Show a, Show b) => Show (Map a b) where
 showsPrec _ fm | isEmpty fm = showString "{|->}"
                | otherwise = showChar '{' . showList (Map.toList fm) . showChar '}'
   where showList [] = showString ""
         showList ((k,e):t) = shows k . showString "|->" . shows e . showTail t

         showTail [] = showString ""
         showTail l = showChar ',' . showList l
-}

{- Now provided by Data.Map
instance (Ord a, Ord b) => Ord (Map a b) where
 m1 <= m2  =  (Map.toList m1) <= (Map.toList m2)
-}


-- | Given a map m, yields the set of pairs (key,elt) where m(key)=elt, ie,
--   builds the relation defined by the map. mkr means 'make relation'.
mkr :: (Ord key, Ord elt) => Map key elt -> Rel key elt
mkr fm = Map.foldWithKey (\ key elt rest -> flip insert rest (key,elt)) Set.empty fm


-- |  Same as mkr. 
fmToSet :: (Ord key, Ord elt) => Map key elt -> Set (key,elt)
fmToSet = mkr 


-- | Given a map, yields the set of keys. It is the same as dom.
setOfKeys :: (Ord key,Ord elt) => Map key elt -> Set key
setOfKeys = dom


-- | Given a map, yields the set of elements. It is the same as rng.
setOfElts :: (Ord key,Ord elt) => Map key elt -> Set elt
setOfElts = rng


{- Now provided by Data.Map
instance Functor (Map a) where
 fmap = map . \f x y->f y
-}

instance (Read a, Read b, Ord a) => Read (Map a b) where
  readsPrec n = map (\(x,y) -> (Map.fromList x, y)) .
                 braces (readSeqSep "," readElem)
   where
    --readElem :: (Read a,Read b) => ReadS (a,b)
    readElem = map (\ (((a,_),b),r) -> ((a,b),r) ) . 
               ((reads `after` readConst "|->") `after` reads)


-- | Applies a given function to a map read from a given file.  
readFile_ :: (Read a, Read b, Ord a, Show c) =>
         FilePath -> ((Map a b) -> c) -> IO c
readFile_ file f = readFile file >>= return.f.read


-- | Applies readFile_ and writes the result in a given file.
interact_ :: (Read a, Read b, Ord a, Show c) =>
         FilePath -> FilePath -> ((Map a b) -> c) -> IO ()
interact_ inn out f = readFile_ inn f >>= writeFile out.show
