-----------------------------------------------------------------------------
-- |
-- Maintainer  :  Alexandra Silva, Joost Visser
-- Stability   :  experimental
-- Portability :  portable
--
-- Representation of Relational Databases, including referenctial integrity check.
--
-----------------------------------------------------------------------------

module Data.HDB.Databases where

import qualified Data.Map as Map
import Data.Map (Map)
import qualified Data.Set as Set
import Data.Set (Set)
import Data.HListExtras
import Data.HDB.RelationalAlgebra
import Data.HDB.Representation
import Prelude hiding (null)

-----------------------------------------------------------------------------
-- * Representation of databases



-- | Well-formedness predicate for relational databases.

data (NoRepeatedAttributes db, CheckRI db) => DB db = DB db

class RDB rdb
instance (NoRepeatedAttributes rdb, CheckRI rdb) => RDB rdb

rdb :: RDB lts => lts -> ()
rdb lts = ()


-- | Global check for referential integrity.
class CheckRI rdb where
  checkRI :: rdb -> Bool
instance (
  CheckRI' (Record l) l
 ) => CheckRI (Record l) where
  checkRI rdb@(Record l) = checkRI' rdb l
  
-- | Helper that has complete rdb as context, and recurses over the contained tables.
class CheckRI' rdb l where
  checkRI' :: rdb -> l -> Bool
instance CheckRI' rdb HNil where
  checkRI' _ _ = True
instance (
  CheckFKs rdb t fks,
  CheckRI' rdb ts
 ) => CheckRI' rdb (HCons (tn,(t,fks)) ts) where
  checkRI' rdb (HCons (_,(t,fks)) ts) 
   = checkFKs rdb t fks && checkRI' rdb ts

-- | Check all FK declarations for a given table.
class CheckFKs rdb t fks where
  checkFKs :: rdb -> t -> fks -> Bool
instance CheckFKs rdb t (Record HNil) where
  checkFKs _ _ _ = True
instance (
  CheckFK rdb t fk tn pk,
  CheckFKs rdb t (Record fks)
 ) => CheckFKs rdb t (Record (HCons (fk, (tn, pk)) fks)) where
  checkFKs rdb t (Record (HCons (fk, (tn, pk)) fks)) 
   = checkFK rdb t (fk, (tn, pk)) && checkFKs rdb t (Record fks)

-- | Check one FK declaration for a given table.
class CheckFK rdb t fk tn pk where
 checkFK :: rdb -> t -> (FK fk tn pk) -> Bool
instance (
  HeaderFor (a,b) k v,
  HasField tn rdb (t,fks),
  HAppend k v kv,
  Row (a,b) k v (Record l),
  LookupMany fk (Record l) l',
  LineUp pk l' x r,
  AttributesFor pk x,
  KeysPresent (ResultList pk x) t
 ) => CheckFK rdb (Table (a,b) k v) fk tn pk where
  checkFK rdb (Table (a,b) m) (fk, (tn, pk))
    = keysPresent r t
      where
        (t,_) = hLookupByLabel tn rdb
        r = ResultList pk xs
        xs = Map.foldWithKey (\k v pv -> (fst $ lineUp pk $ lookupMany fk $ row (a,b) k v):pv) [] m

-- | Check that all values in a given result list 
--   indeed occur as primary keys in a given table.
class KeysPresent r t where
  keysPresent :: r -> t -> Bool
instance  (
   AttributesFor a' k',
   Permute a' k' a k,
   HeaderFor (a,b) k v
 ) => KeysPresent (ResultList a' k') (Table (a,b) k v) where
  keysPresent (ResultList a' vs) (Table (a,_) m)
    = mapHasKeys (foldr (\k' -> Set.insert (permute a' k' a) ) Set.empty vs) m

-- | Check that a given map has a given key.
mapHasKey :: Ord k => k -> Map k v -> Bool
mapHasKey k m = maybe False (const True) $ Map.lookup k m

-- | Check that a given map has a given set of keys.
mapHasKeys :: Ord k => Set k -> Map k v -> Bool
mapHasKeys ks m = Set.fold (\k b -> b && mapHasKey k m) True ks

-- | Permute a list of attributes and corresponding list of
--   according to a second list of attributes.
class Permute as vs as' vs' | as vs as' -> vs'
 where
  permute :: as -> vs -> as' -> vs'
instance
      Permute HNil HNil HNil HNil
 where
  permute _ _ _ = HNil
instance (
      AttributeFor at v,
      HFind at as n,
      HLookupByHNat n vs v,
      HDeleteAtHNat n as as',
      HDeleteAtHNat n vs vs',
      Permute as' vs' as'' vs'' 
 ) => Permute as vs (HCons at as'') (HCons v vs'')
 where
  permute as vs (HCons a as'') = HCons v vs''
   where
    n  = hFind a as 
    v = hLookupByHNat n vs
    as' = hDeleteAtHNat n as
    vs' = hDeleteAtHNat n vs
    vs'' = permute as' vs' as''

-- | Check if all attributes in a database are distinct
class NoRepeatedAttributes rdb 
instance (CollectAttributes rdb ats, NoRepeats ats) => NoRepeatedAttributes rdb 

-- | Collect all attributes used in a relational database
class CollectAttributes rdb ats | rdb -> ats where 
    collectAttributes :: rdb -> ats
instance CollectAttributes (Record HNil) HNil where
    collectAttributes rdb = HNil
instance (HAppend a b ab,  
          CollectAttributes (Record ts) ats', 
          HAppend ab ats' ats,
          HeaderFor (a,b) k v
        ) => CollectAttributes (Record (HCons (l,(Table (a,b) k v,fks)) ts)) ats where
    collectAttributes (Record (HCons (l,(Table (a,b) _,_)) ts)) = hAppend ab (collectAttributes (Record ts))
                        where ab = hAppend a b

-- | Insertion of a table guaranteing well-formedness
insertTableRDB :: (
  HExtend e db db',
  RDB db,
  RDB db'
  ) => e -> db  -> db'
insertTableRDB e db = e .*. db

----------------------------------------------------------------------------- 
-- * Representation of foreign keys

-- | fk references pk in table t
type FK fk t pk = (fk, (t, pk))

----------------------------------------------------------------------------- 

