{-# OPTIONS -fglasgow-exts -fallow-overlapping-instances #-}
{-# OPTIONS -fallow-undecidable-instances #-}
-----------------------------------------------------------------------------
-- |
-- Maintainer  :  Claudia Necco, Joost Visser
-- Stability   :  experimental
-- Portability :  portable
--
-- Generalization where general relations rather than Maps are
-- the underlying data structure.
--
-----------------------------------------------------------------------------
module Data.HDB.Polytypic where

import Data.HDB.Representation
import qualified Data.Map as Map
import Data.Map (Map)
import qualified Data.Set as Set
import Data.Set (Set)
import Data.HListExtras
import Prelude hiding (product)

----
-- * Representation

-- | Tables take three arguments: 
--   header, key type, and value type.
data HeaderFor h ks vs 
  => TableG h r ks vs 
      = TableG h (r ks vs) 
      deriving (Eq,Show,Read)

----
-- * Operations (relational algebra)

-- | Add a single row to a table.
addT :: (
   HeaderFor h ks vs,
   Relation r ks vs
 ) => (ks, vs) -> TableG h r ks vs -> TableG h r ks vs
addT (k,v) (TableG h m) = TableG h m'
  where
    m' = insert k v m

----
-- * Operations (SQL)

productT (TableG (a,b) m) (TableG (a',b') m') = TableG h'' m''
   where
    h'' = (hAppend a a', hAppend b b')
    m'' = product m m'



------
-- * Relations

class Relation r a b where
  insert :: a -> b -> r a b -> r a b
  fold :: (a -> b -> c -> c) -> c -> r a b -> c
  empty :: r a b
instance (Ord a) => Relation Map a b where
  insert = Map.insert
  fold = Map.foldWithKey
  empty = Map.empty
instance (Ord a,Ord b) => Relation Rel a b where
  insert a b (Rel r) = Rel (Set.insert (a,b) r)
  fold f c (Rel r) = Set.fold (uncurry f) c r
  empty = Rel Set.empty
  
class Relation' rab a b where
  insert' :: a -> b -> rab -> rab
instance (Ord a) => Relation' (Map a b) a b where
  insert' = Map.insert
instance (Ord a) => Relation' (Set (a,b)) a b where
  insert' = undefined
  
newtype Rel a b = Rel (Set (a,b)) deriving (Eq,Show)
unRel (Rel r) = r

product :: (
  Relation r a b, Relation r' a' b', Relation r'' a'' b'',
  HAppend a a' a'', HAppend b b' b''
 ) => r a b -> r' a' b' -> r'' a'' b''
product m m'
  = fold worker empty m
    where
      worker k v m'' 
        = fold 
                (\k' v' m'' -> insert (hAppend k k') (hAppend v v') m'') 
                m'' m'


----------
-- * Tests

data ID; atID = undefined :: Attribute String ID
data YN; atYN = undefined :: Attribute Bool YN

h0 = (atID .*. HNil, atYN .*. HNil)

t0 = addT ("foo" .*. HNil,True .*. HNil) $
     addT ("foo" .*. HNil,False .*. HNil) $
       (TableG h0 Map.empty)

t1 = addT ("foo" .*. HNil,True .*. HNil) $
     addT ("foo" .*. HNil,False .*. HNil) $
       (TableG h0 (Rel Set.empty))



----------
