{-# OPTIONS -fglasgow-exts -fallow-overlapping-instances -fallow-undecidable-instances #-}
-----------------------------------------------------------------------------
-- Maintainer  :  Alexandra and Joost Visser
-- Stability   :  experimental
-- Portability :  portable
--
-- 
--
-----------------------------------------------------------------------------

module Data.HDB.Read where

import Data.HListExtras
import Data.HDB.Representation
import Data.HDB.FunctionalDependencies
import Data.HDB.Advanced ()
import Data.HDB.Databases
import Language.Gnumeric.Syntax
import Language.Gnumeric.Parser (readXml')
import qualified Data.Map as Map
import Data.Map (Map)
import Data.List ((\\))
import GHC.List as List
import Text.XML.HaXml.Xml2Haskell hiding (Default)


----
import StrategyLib
import System
---


-- * Read Gnumeric File

-- | reads Gnumeric File to a database : 
        -- Each sheet in the workbook is supposed to have a single table 
        -- with the first row being the header (and therefore ignored)
        -- Referential integrety of resulting database is checked

readRDB fp h fds tns fks = do (wb ::  Gmr'Workbook) <- readGnumeric' fp
                              let cs = map cells (sheets wb )
                              let cols = hLengths h
                              let lCells = map (tail.readTable') $ zip cs cols
                              let db = Record $ hZip tns $ hZip (readTables h fds lCells) fks
                              if (not $ checkRI db) then error "Referential integrity error"
                                                    else return $ DB db


class HLengths l where
    hLengths :: l -> [Int] 
instance HLengths HNil where
    hLengths _ = []
instance (
    HLength l n,
    HNat2Integral n,
    HLengths ls
    ) => HLengths (HCons l ls) where
    hLengths (HCons l ls) = (hNat2Integral $ hLength l): (hLengths ls)
instance (
    HAppend a b l,
    HLength l n,
    HNat2Integral n,
    HLengths ls
    ) => HLengths (HCons (a,b) ls) where
    hLengths (HCons (a,b) ls) = (hNat2Integral $ hLength $ hAppend a b): (hLengths ls)
-- Maps a list of Cells to a list of Values (Strings) in the rigth position (Row,Collumn)    
readTable' (Gmr'Cells cells, max) = Map.elems (Map.map Map.elems (paddM max $ foldr worker Map.empty cells))
                   where worker (Gmr'Cell ats value) m = let (i :: Int) = read $ gmr'CellRow ats
                                                             (j :: Int) = read $ gmr'CellCol ats
                                                         in Map.insertWith Map.union i (Map.singleton j value) m

-- | Function that performs padding in a matrix of Strings, guaranteeing that all positions
--   are covered
--paddM :: Map Int (Map Int String) -> Int -> Map Int (Map Int String)
paddM max m = Map.map paddRow m 
    where paddRow m' = let ks = Map.keys m'
                           ks' = [1..max] \\ ks
                           newElems = foldr (\a m -> Map.insert a "" m) Map.empty ks'
                       in Map.union m' newElems                      
-- Read a matrix of strings into a list of tables
class ReadTables h fds t | h fds -> t where
    readTables :: h -> fds -> [[[String]]] -> t

instance ReadTables HNil fds HNil where
    readTables _ _ _ = HNil

instance (
            ReadTables hs fds' ts, 
            ReadTable (a,b) fds (Table (a,b) k v)
          ) 
       => ReadTables (HCons (a,b) hs) (HCons fds fds') (HCons (Table (a,b) k v) ts) where
    readTables (HCons (a,b) hs) (HCons fds fds') (t:ts) = HCons (readTable (a,b) fds t) (readTables hs fds' ts)

-- | Class that allows the transformation of a matrix of Strings to a table according a given header
     -- When the table is not according to fds it returns empty table 
     -- there was no need to have this function inside a class. however to avoid huge constraint headers 
     -- I've used this class to gather some of teh constraints
class ReadTable h fds t | h -> t where
    readTable :: h -> fds -> [[String]] -> t

instance (CheckFDs fds (Table (a,b) k v), HeaderFor (a,b) k v, ReadRow (a,b) (k,v)) => ReadTable (a,b) fds (Table (a,b) k v) where
    readTable (a,b) fds rows | checkFDs fds t = t
                             | otherwise = Table (a,b) Map.empty
            where t = Table (a,b) (Map.fromList $ map (readRow (a,b)) rows) 

-- | read of a single row
class ReadRow h v | h -> v where
    readRow :: h -> [String] -> v

instance ReadRow HNil HNil where
    readRow _ _ = HNil

instance (
    ReadShow v,
    ReadRow ats vs
    ) => ReadRow (HCons (Attribute v nr) ats) (HCons v vs) where
    readRow (HCons at ats) (s:ss) = HCons v (readRow ats ss)
                where v = fromString s

instance (
    ReadShow v,
    ReadRow ats vs
    ) => ReadRow (HCons (NullableAttribute v nr) ats) (HCons (Maybe v) vs) where
    readRow  (HCons at ats) ("":ss) = HCons Nothing  (readRow ats ss)
    readRow  (HCons at ats) (s:ss) =  HCons v (readRow ats ss)
                where v = Just $ fromString s

instance (
    ReadShow v,
    ReadRow ats vs
    ) => ReadRow (HCons (DefaultableAttribute v nr) ats) (HCons v vs) where
    readRow (HCons (Default v) ats) ("":ss) = HCons v (readRow ats ss)
    readRow (HCons at ats) (s:ss) = HCons v (readRow ats ss)
                where v = fromString s

-- | Lifting to pairs
instance (ReadRow a k, ReadRow b v, HLength a n, HNat2Integral n) => ReadRow (a,b) (k,v) where
    readRow (a,b) l = let (k,v) = splitAt (hNat2Integral $ hLength a) l
                      in (readRow a k, readRow b v) 

class ReadShow t where
    fromString :: String -> t
    toString :: t -> String

instance (Read t, Show t) => ReadShow t where
    fromString = read
    toString = show

instance ReadShow String where
    fromString = id
    toString = id
    
---------------------
-- * Auxiliary stuff for gnumeric

-- | Reads a gnumeric file into a workbook
readGnumeric' :: (Term a, XmlContent a) => FilePath -> IO a
readGnumeric' fileName = do
  let tmpFile = "/tmp/TestGnumericTEMPFILE"
  let systemCmd = "zcat \""++fileName++"\" > "++tmpFile
  system systemCmd
  content <- readFile tmpFile
  let parseResult = {- parseFormulas $ -} readXml' content
  case (parseResult) of
    (Just workbook) -> return workbook
    Nothing -> error "NO PARSE"

-- | Selector for list stored in field sheets
sheets (Gmr'Workbook a b c d e f g h i (Gmr'Sheets sts) k l) = sts

-- | Selector for field cells
cells (Gmr'Sheet a b c d e f g h i j k l cs n o p q)
 = cs


