{-# OPTIONS -fglasgow-exts -fallow-overlapping-instances -fallow-undecidable-instances #-}
-----------------------------------------------------------------------------
-- |
-- Maintainer  :  Alexandra Silva, Joost Visser
-- Stability   :  experimental
-- Portability :  portable
--
-- Implementation of relational algebra operators on tables.
--
-----------------------------------------------------------------------------
module Data.HDB.RelationalAlgebra where

import Data.HDB.Representation
import qualified Data.Map as Map
import Data.Map (Map)
import qualified Data.Set as Set
import Data.Set (Set)
import Data.HListExtras

-----------------------------------------------------------------------------
-- * Row operations
-- | Selection of a column value from a row
--   using a column type.

get :: (
   HFind a c n, 
   HLookupByHNat n r v
 ) => a -> c -> r -> v
get a c r = v
  where
    n = hFind a c
    v = hLookupByHNat n r
    
-----------------------------------------------------------------------------
-- * Operations (primitive)
--
-- Here the split between key columns and non-key columns is still
-- exposed in the arguments.

-- | Add a single row to a table.
add :: (
   HeaderFor h ks vs
 ) => (ks, vs) -> Table h ks vs -> Table h ks vs
add (k,v) (Table h m) = Table h m'
  where
    m' = Map.insert k v m

-- | Insert operation.
--   The order of the values to be inserted need not coincide exactly
--   with the order of the attributes in the table. Only when attributes
--   of the same type occur, the values for these attributes need to be
--   supplied in the right order.
insert :: (
  HeaderFor h ks vs,
  Split vs' h ks vs
 ) => vs' -> Table h ks vs -> Table h ks vs
insert vs (Table h mp) = Table h mp'
 where
  mp' = Map.insert kvs nkvs mp
  (kvs,nkvs) = split vs h

-- | Similar to insert.
updateV :: (
  TABLE (Table h ks vs),
  HeaderFor h ks vs,
  Split vs' h ks vs
 ) => vs' -> Table h ks vs -> Table h ks vs
updateV vs (Table h mp) = Table h mp'
 where
  mp' = Map.update (\_ -> Just nkvs) kvs mp
  (kvs,nkvs) = split vs h


-- * Multiple table operations

unionT :: HeaderFor h k v => Table h k v -> Table h k v -> Table h k v
unionT (Table h m) (Table _ m') = Table h (Map.union m m')

intersectionT :: HeaderFor h k v => Table h k v -> Table h k v -> Table h k v
intersectionT (Table h m) (Table _ m') = Table h (Map.intersection m m')
 
differenceT :: HeaderFor h k v => Table h k v -> Table h k v -> Table h k v
differenceT (Table h m) (Table _ m') = Table h (Map.difference m m')




-- | Filter rows using a predicate on keys and values.
--   Keep rows for which the predicate holds.
filter :: (
   HeaderFor h ks vs
 ) => ((ks, vs) -> Bool) -> Table h ks vs -> Table h ks vs
filter p (Table s m) = Table s m'
  where m' = Map.filterWithKey (\x y -> p (x,y)) m

-- | Project, using a list of KEY attributes.
--   All non-key attributes are retained.
--   Selected key attributes and non-key attributes all move to the key.
projectKeys :: (
   HeaderFor (a,b) k v,
   RESULTSET a'b k'v,
   LineUp a' k k' r,
   HAppend a' b a'b,
   HAppend k' v k'v
 ) => a' -> Table (a,b) k v -> ResultSet a'b  k'v
projectKeys a' (Table (a,b) m) = ResultSet a'b m'
  where 
    a'b = hAppend a' b
    m' = Map.foldWithKey worker Set.empty m
    worker k v s 
      = Set.insert (hAppend (fst $ lineUp a' k) v) s

-- | Project, using a list of NON-KEY attributes.
--   In the result, all key columns are retained as such.
projectValues :: (
   HeaderFor (a,b) k v,
   HeaderFor (a,b') k v',
   LineUp b' v v' r
 ) => b' -> Table (a,b) k v -> Table (a,b') k v'
projectValues b' (Table (a,b) m) = Table (a,b') m'
  where 
    m' = Map.map (fst . lineUp b') m

fold :: HeaderFor h k v 
   => (h -> k -> v -> b -> b) -> b -> Table h k v -> b
fold f a (Table h m) = a'
  where
    a' = Map.foldWithKey (\k v a -> f h k v a) a m

-----------------------------------------------------------------------------
-- * Operations (flat column specs)
--
-- Here the split between key columns and non-key columns is no longer
-- exposed in the arguments. Instead, a flat list of columns is used.


-- | Project using a list of attributes (column specification).
--   Keep specified columns.
--   Order of columns can deviate from order in the table header.
--   In the result, all columns are keys.

project :: (
   HeaderFor h k v,
   RESULTSET ab' kv',
   LookupMany ab' r kv',
   Row h k v r
 ) => ab' -> Table h k v -> ResultSet ab' kv'
project ab' (Table h m) = ResultSet ab' m'
  where 
    m' = Map.foldWithKey worker Set.empty m
    worker k v mp 
      = Set.insert (lookupMany ab' $ row h k v) mp

groupBy :: (
   RESULTSET a v,
   Ord k, 
   LookupMany a' (Record r) k,
   HZip a v r
 ) => a' -> ResultSet a v -> Map k (Set v)
groupBy a' (ResultSet a v) = Set.fold worker Map.empty v
 where
  worker v
   = Map.insertWith Set.union (lookupMany a' $ Record $ hZip a v) (Set.singleton v)



-- | A list of values can be split into a list of 
--   key values and a list of non-key values, in 
--   accordance with a given header.
--   We use a class because to avoid repeating 
--   the list of constraints in functions that
--   use splitKey. We add the Ord constaint on 
--   key types, to ensure they can be used in a map.

class Ord kvs => Split' vs h kvs nkvs | vs h -> kvs nkvs
 where
  split' :: vs -> h -> (kvs,nkvs)
instance (
      LineUp kas vs kvs r,
      LineUp nkas r nkvs (Record HNil),  -- No elements are allowed to remain.
      Ord kvs
 ) => Split' vs (kas,nkas) kvs nkvs
 where
   split' vs (kas,nkas) = (kvs,nkvs)
    where
     (kvs,r)  = lineUp kas vs  -- hProject' vs
     (nkvs,_) = lineUp nkas r -- hProject' vs   

-- | Split keys, with insertion of NULL or Default where necessary.
class Ord kvs => Split vs h kvs nkvs | vs h -> kvs nkvs
 where
  split :: vs -> h -> (kvs,nkvs)
instance (
      LineUp kas vs kvs r,
      LineUpPadd nkas r nkvs HNil,  -- No elements are allowed to remain.
      Ord kvs
 ) => Split vs (kas,nkas) kvs nkvs
 where
   split vs (kas,nkas) = (kvs,nkvs)
    where
     (kvs,r)  = lineUp kas vs  
     (nkvs,_) = lineUpPadd nkas r 




{- Illegal (GOOD)
-- | Putting a row that is too long.

customerTable'' =
  insert ( "Oracle" .*. (1::Integer) .*. "New York" .*. "Seattle" .*. HNil) $ myTable
-}


-----------------------------------------------------------------------------
-- Line up a list of values according a given header
-- Value list to line up can be longer than the specification list
-- Any remaining values are also returned

class LineUp h v v' r | h v -> v' r
 where
  lineUp :: h -> v -> (v',r)

-- If a header is a pair then the first component of the pair represents
-- keys and, therefore, cannot be padded. The second component of the 
-- pair are normal values which are lined up and padding is performed
instance (
    LookupMany a (Record row) ks,
    DeleteMany a (Record row) (Record row'),
    LineUpPadd b (Record row') vs (Record r)
  ) => LineUp (a,b) (Record row) (ks,vs) (Record r)
 where
  lineUp (a,b) row = ((ks,vs), r)
        where ks = lookupMany a row
              row' = deleteMany a row
              (vs,r) = lineUpPadd b row'
instance(
    LookupMany a (Record row) ks,
    DeleteMany a (Record row) (Record row'),
    LineUpPadd b (Record row') vs (Record r)
  ) => LineUp (a,b) row (ks,vs) (Record r)
 where
  lineUp (a,b) row = ((ks,vs), r)
        where ks = lookupMany a (Record row)
              row' = deleteMany a (Record row)
              (vs,r) = lineUpPadd b row'

instance
      LineUp HNil vs HNil vs
 where
  lineUp _ vs = (HNil,vs)

-- LineUp according a plain list of attributes: no padding is performed
instance (
      AttributeFor at v,
      HFind v vs n,
      HLookupByHNat n vs v,
      HDeleteAtHNat n vs vs',
      LineUp as vs' vs'' r
 ) => LineUp (HCons at as) vs (HCons v vs'') r
 where
  lineUp (HCons a as) vs = (HCons v vs'',r)
   where
    n  = hFind (mkDummy a) vs 
    v = hLookupByHNat n vs
    vs' = hDeleteAtHNat n vs
    (vs'',r) = lineUp as vs'



-- | Line up and do padding.

class LineUpPadd b row vs r | b row -> vs r
 where
  lineUpPadd :: b -> row -> (vs,r)

-- | Lineup of a row 
-- The row where the lineup is performed is presented as a 
-- record where attributes identifiers are linked to values
-- The result is presented not as a Record but as a normal 
-- heterogeneous list
instance (
   AttributeFor at v,
   HMember (at,v) row b,
   LineUpPadd' b at (Record row) v row',
   LineUpPadd ats row' vs r
 ) => LineUpPadd (HCons at ats) (Record row) (HCons v vs) r
 where
  lineUpPadd (HCons at ats) (Record row) = (HCons v vs,r)
    where
     (v,row')  = lineUpPadd' (hMember (at,(mkDummy at)) row) at (Record row)
     (vs,r) = lineUpPadd ats row'

instance
      LineUpPadd HNil vs HNil vs
 where
  lineUpPadd _ vs = (HNil,vs)

-- | LineUp of a plain list of values
instance (
   AttributeFor at v,
   HMember v vs b,
   LineUpPadd' b at vs v r,
   LineUpPadd as r vs' r'
 ) => LineUpPadd (HCons at as ) vs (HCons v vs') r'
 where
  lineUpPadd (HCons a as) vs
   = (HCons v' vs',r')
         where v = mkDummy a
               (v',r)= lineUpPadd' (hMember v vs) a vs
               (vs',r') = lineUpPadd as r 

-- | Helper function to lineUp that returns the value to insert in 
--   the list and the remaining list to lineUp
class LineUpPadd' b at r v r' | b at r -> v r'
 where
  lineUpPadd' :: b -> at -> r -> (v,r')

-- | If the element searched is not in the list of Values and corresponds
--   to a nullable attribute Nothing is returned together with the same list
--   since no value was deleted from it
instance LineUpPadd' HFalse (NullableAttribute v nr) r (Maybe v) r
 where
  lineUpPadd' b at r
   = (Nothing,r)

-- | If the element searched is not in the list of Values and corresponds
--   to a defaultable attribute, user declared default value is returned 
--   together with the same list since no value was deleted from it
instance LineUpPadd' HFalse (DefaultableAttribute v nr) r v r
 where
  lineUpPadd' b (Default v) r
   = (v,r)

-- | If the element searched is not in the list of Values and corresponds
--   to a normal attribute, but its type is system default, that declared 
--   value is is returned together with the same list
--   since no value was deleted from it
instance 
    Defaultable v 
    => LineUpPadd' HFalse (Attribute v nr) r v r
 where
  lineUpPadd' b at r
   = (defaultValue ,r)

-- | If the element searched is in the list of values the element and the list 
--   whithout the element is returned
instance (
    HasField at r v, 
    H2ProjectByLabels (HCons at HNil) r rin r'
  ) => LineUpPadd' HTrue at (Record r) v (Record r')
 where
  lineUpPadd' b at r = (r .!. at, hDeleteAtLabel at r)

instance (
      AttributeFor at v,
      HFind v vs n,
      HLookupByHNat n vs v,
      HDeleteAtHNat n vs vs'
 ) => LineUpPadd' HTrue at vs v vs'
 where
  lineUpPadd' b at vs = (v', vs')
   where
    v = mkDummy at
    n  = hFind v vs 
    v' = hLookupByHNat n vs
    vs' = hDeleteAtHNat n vs


