{-# OPTIONS -fglasgow-exts -fallow-overlapping-instances -fallow-undecidable-instances #-}
-----------------------------------------------------------------------------
-- |
-- Maintainer  :  Alexandra Silva, Joost Visser
-- Stability   :  experimental
-- Portability :  portable
--
-- Extensions to the HUnit Library.
-- Re-exports useful parts of the HUnit Library for convenient import.
--
-----------------------------------------------------------------------------

module Data.HListExtras (

 module Data.HListExtras,
 module HListPrelude,
 module FakePrelude,
 module HOccurs,
 module GhcSyntax,
 module HTypeIndexed,
 module HArray,
 module TypeEqGeneric1,
 module TypeCastGeneric1,
 module Record,
 module HZip
 
) where

import HListPrelude
import FakePrelude
import HOccurs
import GhcSyntax ((.*.),(:*:),(.=.),(:=:),(.!.),(:+:))
import HTypeIndexed (hProjectByProxies,HTypes2HNats)
import HArray (HProjectByHNats,hProjectByHNats,
               HLookupByHNat,hLookupByHNat,
               HDeleteAtHNat,hDeleteAtHNat,
               HUpdateAtHNat, HLength, hLength)
import TypeEqGeneric1 ()
import TypeCastGeneric1 ()
import Record -- hiding (HasField, hLookupByLabel)
import HZip

-----------------------------------------------------------------------------
-- * Numbers

-- | Named values.
one = hSucc hZero
two = hSucc one
three = hSucc two
four = hSucc three
five = hSucc four
six = hSucc five
seven = hSucc six
eigth = hSucc seven
nine = hSucc eigth
ten = hSucc nine

-- | Named types
type Zero    = HZero
type One = HSucc Zero
type Two = HSucc One
type Three = HSucc Two
type Four = HSucc Three
type Five = HSucc Four
type Six = HSucc Five
type Seven = HSucc Six
type Eigth = HSucc Seven
type Nine = HSucc Eigth
type Ten = HSucc Nine

-----------------------------------------------------------------------------
-- * List operations.

-- | All elements in one list also occur in the other.
--   Duplicates are allowed in both lists.
class OccurAll sb sp
instance OccurAll HNil sp
instance (
   HFind a sp n,
   OccurAll sb sp
 ) => OccurAll (HCons a sb) sp

hElem :: (HOccursMany e l, Eq e) => e -> l -> Bool
hElem e l = elem e $ hOccursMany l 

-----------------------------------------------------------------------------
-- * Set operations

-- | Set union on lists.
class (
   HSet l,
   HSet l'
 ) => Union l l' l'' | l l' -> l''
 where
  union :: l -> l' -> l''
instance (
   HSet l
 ) => Union HNil l l
 where
  union _ l = l
instance (
   HSet (HCons e l),
   HSet l',
   HMember e l' b,
   Union' b e l l' l''
 ) => Union (HCons e l) l' l''
 where
  union (HCons e l) l' = l''
   where
    b = hMember e l'
    l'' = union' b e l l'

-- | Auxiliary to Union
class Union' b e l l' l'' | b e l l' -> l''
 where
  union' :: b -> e -> l -> l' -> l''
instance (
   Union l l' l''
 ) => Union' HTrue e l l' l'' 
 where
  union' _ _ l l' = union l l' 
instance (
   Union l l' l''
 ) => Union' HFalse e l l' (HCons e l'') 
 where
  union' _ e l l' = HCons e (union l l')

-- | Set intersection on lists.
class (
   HSet l,
   HSet l'
 ) => Intersect l l' l'' | l l' -> l''
 where
  intersect :: l -> l' -> l''
instance (
   HSet l
 ) => Intersect HNil l HNil
 where
  intersect _ l = hNil
instance (
   HSet (HCons e l),
   HSet l',
   HMember e l' b,
   Intersect' b e l l' l''
 ) => Intersect (HCons e l) l' l''
 where
  intersect (HCons e l) l' = l''
   where
    b = hMember e l'
    l'' = intersect' b e l l'

-- | Auxiliary to Intersect
class Intersect' b e l l' l'' | b e l l' -> l''
 where
  intersect' :: b -> e -> l -> l' -> l''
instance (
   Intersect l l' l''
 ) => Intersect' HFalse e l l' l'' 
 where
  intersect' _ _ l l' = intersect l l' 
instance (
   Intersect l l' l''
 ) => Intersect' HTrue e l l' (HCons e l'') 
 where
  intersect' _ e l l' = HCons e (intersect l l')

-- | Set difference on lists.
class (
   HSet l,
   HSet l'
 ) => Difference l l' l'' | l l' -> l''
 where
  difference :: l -> l' -> l''
instance (
   HSet l
 ) => Difference HNil l HNil
 where
  difference _ l = hNil
instance (
   HSet (HCons e l),
   HSet l',
   HMember e l' b,
   Difference' b e l l' l''
 ) => Difference (HCons e l) l' l''
 where
  difference (HCons e l) l' = l''
   where
    b = hMember e l'
    l'' = difference' b e l l'

-- | Auxiliary to Difference
class Difference' b e l l' l'' | b e l l' -> l''
 where
  difference' :: b -> e -> l -> l' -> l''
instance (
   Difference l l' l''
 ) => Difference' HTrue e l l' l'' 
 where
  difference' _ _ l l' = difference l l' 
instance (
   Difference l l' l''
 ) => Difference' HFalse e l l' (HCons e l'') 
 where
  difference' _ e l l' = HCons e (difference l l')

-- The union of a list of sets

class Unions ll l | ll -> l

instance Unions HNil HNil
instance (
        Unions ll l', 
        Union e l' l
      ) => Unions (HCons e ll) l

-- | Test if an heterogeneous list is empty
class IsEmpty l where isEmpty :: l -> Bool
instance IsEmpty HNil where isEmpty _ = True
instance IsEmpty (HCons a l) where isEmpty _ = False

----------------------------------------------------------------------
-- | Selecting specific attributes in a single tuple according a schema

class HFilter s s' t t' | s s' t -> t' where
    hFilter :: s-> s' -> t -> t'

instance (HFindMany s s' lns,  HProjectByHNats lns t t') 
    => HFilter s s' t t'
    where
        hFilter  s s' t = t'
            where lns = hFindMany s s' 
                  t' = hProjectByHNats lns t 

-----------------------------------------------------
-- | Function that gives the positions of elements in a List 

class HFindMany e l l' | e l -> l'
 where
  hFindMany :: e -> l -> l'

instance (HFindMany e (HCons x xs) m, HFindMany l (HCons x xs) l', MapHSucc l' l'', HAppend m l'' l''') 
        => HFindMany (HCons e l) (HCons x xs) l'''
  where
   hFindMany  (HCons e l) xs = hAppend m l'' 
        where m = hFindMany e xs
              l'= hFindMany l xs 
              l'' = mapHSucc l'

instance HFindMany e HNil HNil
 where
  hFindMany _ HNil = HNil

instance ( HList l
         , TypeEq e e' b
         , HFindManyCase b e e' l l'
         )
      =>   HFindMany e (HCons e' l) l'
 where
  hFindMany e (HCons e' l) = l'
   where
    b  = typeEq e e'
    l' = hFindManyCase b e e' l


class HFindManyCase b e e' l l' | b e e' l -> l'
 where
  hFindManyCase :: b -> e -> e' -> l -> l'

instance (HFindMany e l l'
         , MapHSucc l' l'')
      => HFindManyCase HTrue e e l (HCons HZero l'')
 where
  hFindManyCase _ e _ l = HCons hZero (mapHSucc (hFindMany e l))


instance HFindMany e l l'
      => HFindManyCase HFalse e e' l l'
 where
  hFindManyCase _ e e' l = (hFindMany e l)


{------------------------------------------------------------------------------}
-- | Function that maps all naturals in a list to its successor

class HNats l => MapHSucc l l' | l -> l' where
    mapHSucc :: l -> l'

instance MapHSucc HNil HNil
 where
  mapHSucc _ = hNil
instance (HNat e, MapHSucc l l') => MapHSucc (HCons e l) (HCons (HSucc e) l')
 where
  mapHSucc (HCons e l) = HCons (hSucc e) (mapHSucc l)

{--------------------------------------------------------------------------------------}
-- | instance of HStagedEq for any two elements based on TypeEq on TypeEqGeneric2.hs
instance (TypeEq e e' b
         , HStagedEq' b e e') => HStagedEq e e' where
     hStagedEq e e' = hStagedEq' (typeEq e e') e e'
{---------------------------------------------------------------------------------------}

-- | instances of SubType for HNil, HCons, Pairs and Tuple

instance HList l => SubType HNil l
instance (HOccursMany1 e l, SubType l' l)
      =>  SubType (HCons e l') l
instance (SubType l l', SubType m m')
      =>  SubType (l,m) (l', m')

-- | Complement of Booleans
class Not b b' | b->b'
instance Not HTrue HFalse
instance Not HFalse HTrue

-- | Check if a a list l is contained in l' (it is a strict contained -- the lists cannot be equal)
class Contained l l' b | l l' -> b
instance Contained HNil (HCons e l) HTrue
instance Contained HNil HNil HFalse
instance (HMember at l' b', Contained l l' b'', HAnd b' b'' b)
        => Contained (HCons at l) l' b

contained :: Contained l l' b => l -> l' -> b
contained = undefined

class ContainedEq l l' b | l l' -> b
instance ContainedEq HNil HNil HTrue
instance ContainedEq HNil (HCons e l) HTrue
instance (HMember at l' b', Contained l l' b'', HAnd b' b'' b)
        => ContainedEq (HCons at l) l' b

containedEq :: ContainedEq l l' b => l -> l' -> b
containedEq = undefined


-- | Convert type-level booleans to value-level booleans.
class HBool b => BooleanValue b where
    booleanValue :: b -> Bool
instance BooleanValue HTrue where
    booleanValue _ = True
instance BooleanValue HFalse where
    booleanValue _ = False 


-- SubLists of a List

class PowerSet l ls | l -> ls 

instance PowerSet HNil (HCons HNil HNil)
instance (
    PowerSet l l', 
    MapHAppend e l' l'',
    HAppend l' l'' l'''
  ) => PowerSet (HCons e l) l'''

powerSet :: (PowerSet l l') => l -> l'
powerSet = undefined

class ProperSubsets l ls | l -> ls
instance (
    PowerSet l l', 
    FilterEmptyTotal l l' l''
  ) => ProperSubsets l l''

properSubsets :: (ProperSubsets l l') => l -> l'
properSubsets = undefined


-- | Eliminate Empty and Total set from PowerSet list --> getting proper subsets only
class FilterEmptyTotal a ps ps' | a ps -> ps'

instance FilterEmptyTotal a HNil HNil
instance (
   HEq all s b,
   HEq HNil s b',
   HOr b b' b'',
   FilterEmptyTotal' b'' all (HCons s l) l'
 ) => FilterEmptyTotal all (HCons s l) l'


-- | Eliminate Empty set from PowerSet list
class FilterEmptySet ps ps' | ps -> ps'

instance FilterEmptySet HNil HNil
instance (
   HEq HNil s b,
   FilterEmptySet' b (HCons s l) l'
 ) => FilterEmptySet (HCons s l) l'

-- * Auxiliary to FilterEmptySet
class FilterEmptySet' b ps ps' | b ps -> ps'
instance FilterEmptySet l l'=> FilterEmptySet' HTrue (HCons e l) l'
instance FilterEmptySet l l'=> FilterEmptySet' HFalse (HCons e l) (HCons e l')

-- testing only
fet :: FilterEmptyTotal a ps ps' => a -> ps -> ps'
fet = undefined
fe :: FilterEmptySet ps ps' =>  ps -> ps'
fe = undefined

-----
-- * Auxiliary to FilterEmptyTotal
class FilterEmptyTotal' b a ps ps' | b a ps -> ps'

instance FilterEmptyTotal all l l' => FilterEmptyTotal' HTrue all (HCons s l) l'
instance FilterEmptyTotal all l l' => FilterEmptyTotal' HFalse all (HCons s l) (HCons s l')


-- * Auxiliary for PowerSet -- appending an element to every list 
class MapHAppend e lls lls' | e lls -> lls'
instance MapHAppend e HNil HNil
instance MapHAppend e ls ls' => MapHAppend e (HCons l ls) (HCons (HCons e l) ls')

mapHAppend :: (MapHAppend e l l') => e -> l -> l'
mapHAppend = undefined

-- * Auxiliary to IsPrime 
class MapHMember e lls b | e lls -> b
instance MapHMember e HNil HFalse
instance (
    HMember e l b', 
    MapHMember e ls b'',
    HOr b' b'' b 
  ) => MapHMember e (HCons l ls) b

mapHMember :: (MapHMember e l b) => e -> l -> b
mapHMember = undefined


-----------------------------------------------------------------------------
-- * General type-level functionality.

-- | Type-level predicate that tests that no type is repeated
--   in a list.
class NoRepeats l 
instance NoRepeats HNil 
instance (
      HOccursNot e l,
      NoRepeats l
 ) => NoRepeats (HCons e l) 



-- | Create a list of proxies out of a list
class ToProxies xs ps | xs -> ps
 where
  toProxies :: xs -> ps
instance ToProxies HNil HNil
 where
  toProxies _ = HNil
instance (
      ToProxies xs ps
 ) => ToProxies (HCons x xs) (HCons (Proxy x) ps)
 where
  toProxies (HCons x xs) = HCons (toProxy x) (toProxies xs)


-- | Now we can do projection via proxies:

myproject l
 = l'
   where -- Note the circularity:
     ps = toProxies l'
     l' = hProjectByProxies ps l

-- | Variant on HProject that allows repeated types.
--   In case of repeated types, order is preserved,
--   and later occurrences will be dropped in the projection
--   if the number of repetitions is lower.
class HProject' l l'
 where
  hProject' :: l -> l'
instance (
      HTypes2HNats ps l ns,
      HProjectByHNats ns l l',
      ToProxies l' ps
 ) => HProject' l l'
 where
  hProject' = myproject  

-- | Record modification
class HasField l r v => ModifyAtLabel l v v' r r' | l r v' -> v r' where
  modifyAtLabel :: l -> (v->v') -> r -> r'
instance (
   HasField l r v,
   HZip x y r,
   HFind l x n,
   HUpdateAtHNat n v' y y',
   HZip x y' r'
 ) => ModifyAtLabel l v v' (Record r) (Record r') where
   modifyAtLabel l f r = r'
     where
       v = r .!. l
       r' = hUpdateAtLabel l (f v) r

-- Record Update based on a second record, possibly smaller
class UpdateWith r s 
 where
  updateWith :: r -> s -> r
instance (
  UpdateWith r s
 ) => UpdateWith r (Record s)
 where
  updateWith r (Record s) = updateWith r s
instance UpdateWith r HNil
 where
  updateWith r _ = r
instance (
   HasField l (Record r) v,
   UpdateWith (Record r) s,
   HZip ls vs r, HFind l ls n, HUpdateAtHNat n v vs vs
 ) => UpdateWith (Record r) (HCons (l,v) s)
 where
  updateWith r (HCons (l,v) s)
   = updateWith (hTPupdateAtLabel l v r) s

-- | Lookup the values of a list of labels in a record.
class LookupMany ls r vs | ls r -> vs
 where
  lookupMany :: ls -> r -> vs
instance LookupMany HNil r HNil
 where
  lookupMany _ r = HNil
instance (
   HasField l r v,
   LookupMany ls r vs
 ) => LookupMany (HCons l ls) r (HCons v vs)
 where
  lookupMany (HCons l ls) r = HCons v vs
    where
     v  = r .!. l
     vs = lookupMany ls r

-- | Delete the fields in record corresponding to a list of labels 
class DeleteMany ls r vs | ls r -> vs
 where
  deleteMany :: ls -> r -> vs

instance DeleteMany HNil r r 
 where
  deleteMany _ r = r
instance (
   H2ProjectByLabels (HCons l HNil) r rin rout,
   HasField l (Record r) v,
   DeleteMany ls (Record rout) r'
 ) => DeleteMany (HCons l ls) (Record r) r'
 where
  deleteMany (HCons l ls) r = r''
    where
     r' = hDeleteAtLabel l r
     r'' = deleteMany ls r'

class HasLabel l r b | l r -> b where
    hasLabel :: l -> r -> b

instance (
    HZip ls vs r,
    HTMember l ls b
    ) => HasLabel l (Record r) b where
    hasLabel l (Record r) = hTMember l ls
        where (ls,_) = hUnzip r
            
class HasTEqField l r v | l r -> v where
    teqLookupByLabel:: l -> r -> v

instance HasTEqField l r v => HasTEqField l (Record r) v where
    teqLookupByLabel l (Record r) = teqLookupByLabel l r

class HasTEqField' b l r v | b l r -> v where
    teqLookupByLabel':: b -> l -> r -> v

instance (TypeEq l l' b, HasTEqField' b l (HCons (l',v') r) v)
    => HasTEqField l (HCons (l',v') r) v where
    teqLookupByLabel l r@(HCons (l',_) _) = teqLookupByLabel' (typeEq l l') l r

instance HasTEqField' HTrue l (HCons (l,v) r) v where 
    teqLookupByLabel' _ _ (HCons (_,v) _) = v
instance HasTEqField l r v => HasTEqField' HFalse l (HCons (l',v') r) v where
    teqLookupByLabel' _ l (HCons _ r) = teqLookupByLabel l r
