module Data.Relation.Folds where

import Data.Set hiding (fold,empty)
import Data.Monoid
import Data.FiniteMap

{-

Under construction.

Attempt hide relation datatype under an interface.

-}

-- Several types of relations:

-- Set of pairs
type SoP a b = Set (a,b)

-- Predicate of pairs
type PoP a b = (a,b) -> Bool

-- Map to Sets
type MoP a b = FiniteMap a (Set b)

-- Defining some basic operations:

emptySoP :: SoP a b
emptySoP = emptySet

emptyPoP :: PoP a b
emptyPoP = const False

emptyMoP :: MoP a b
emptyMoP = emptyFM

--

invSoP :: (Ord a, Ord b) => SoP a b -> SoP b a
invSoP = mapSet flip where flip (a,b) = (b,a)

invPoP :: PoP a b -> PoP b a
invPoP p = p . flip where flip (a,b) = (b,a)

invMoP :: (Ord a, Ord b) => MoP a b -> MoP b a
invMoP = foldFM f emptyFM
  where 
    f a bs r = addListToFM_C union r [(b,unitSet a) | b <- setToList bs ]

--

compSoP :: (Ord a, Eq b, Ord c) => SoP b c -> SoP a b -> SoP a c
compSoP yz xy
 = mkSet [ (x,z) | (x,y) <- setToList xy, (y',z) <- setToList yz, y==y' ]

compPoP :: PoP b c -> PoP a b -> PoP a c
compPoP p q = \(a,c) -> error "STUCK"

compMoP :: (Ord a, Ord c, Ord b) => MoP b c -> MoP a b -> MoP a c
compMoP yz xy = mapFM f xy
  where
    f _ ys 
      = unionManySets [ lookupWithDefaultFM yz emptySet y | y <- setToList ys ]
    
--

unionSoP :: (Ord a, Ord b) => SoP a b -> SoP a b -> SoP a b
unionSoP = union

unionMoP :: (Ord a, Ord b) => MoP a b -> MoP a b -> MoP a b
unionMoP = plusFM_C union

--

instance (Ord a, Ord b) => Monoid (SoP a b) where
  mempty = emptySoP
  mappend = unionSoP
  
instance (Ord a, Ord b) => Monoid (MoP a b) where
  mempty = emptyMoP
  mappend = unionMoP
  
--

{-
class Inv r r' where
  inv :: r -> r' 
instance (Ord a, Ord b) => Inv (SoP a b) (SoP b a) where
  inv = invSoP
instance (Ord a, Ord b) => Inv (MoP a b) (MoP b a) where
  inv = invMoP
-}

{-
class Comp r r' r'' where
  comp :: r -> r' -> r''
instance (Ord a, Ord c, Ord b) => Comp (SoP b c) (SoP a b) (SoP a c) where
  comp = compSoP
instance (Ord a, Ord c, Ord b) => Comp (MoP b c) (MoP a b) (MoP a c) where
  comp = compMoP
-}

--

foldSet :: (a -> v -> v) -> v -> Set a -> v
foldSet f e = foldr f e . setToList

foldSoP :: (a -> b -> v -> v) -> v -> SoP a b -> v
foldSoP f e = (foldSet (uncurry f) e)

foldMoP :: (a -> b -> v -> v) -> v -> MoP a b -> v
foldMoP f e = foldFM (\a bs v -> foldSet (\b -> f a b) v $ bs) e

--

class Fold r a b | r -> a b where
  fold :: (a -> b -> v -> v) -> v -> r -> v
instance Fold (SoP a b) a b where
  fold = foldSoP

-- 

unfoldSet :: (Ord a) => (u -> Either () (a,u)) -> u -> Set a
unfoldSet f u = case f u of
  Left ()        -> emptySet
  Right (a,u') -> addToSet (unfoldSet f u') a 

unfoldSoP :: (Ord a, Ord b) => (u -> Either () (a,b,u)) -> u -> SoP a b
unfoldSoP f u = case f u of
  Left ()        -> emptySet
  Right (a,b,u') -> addToSet (unfoldSoP f u') (a,b) 

unfoldMoP :: (Ord a, Ord b) => (u -> Either () (a,b,u)) -> u -> MoP a b
unfoldMoP f u = case f u of
  Left ()        -> emptyFM
  Right (a,b,u') -> addToFM_C union (unfoldMoP f u') a (unitSet b) 

--

class Unfold r a b | r -> a b where
  unfold :: (u -> Either () (a,b,u)) -> u -> r
instance (Ord a, Ord b) => Unfold (SoP a b) a b where
  unfold = unfoldSoP
instance (Ord a, Ord b) => Unfold (MoP a b) a b where
  unfold = unfoldMoP

--

empty :: Unfold r a b => r
empty = unfold Left ()

add :: Unfold r a b  => a -> b -> u -> r
add a b r = unfold (\_ -> Right (a,b,r)) r

inv :: (Unfold r' b a, Fold r a b) => r -> r'
inv = unfold f . fold (\a b l -> (a,b):l) []
  where
    f ((a,b):l) = Right (b,a,l)
    f []        = Left ()
    
inv' :: (Unfold r' b a, Fold r a b) => r -> r'
inv' r = fold (\a b r -> add b a r) empty r

pairs :: Fold r a b => r -> [(a, b)]
pairs = fold (\a b l -> (a,b):l) [] 

comp :: (Fold xy x y, Fold yz y z, Eq y, Unfold xz x z) =>
        yz -> xy -> xz
comp yz xy = fold f empty xy
  where
--    f x y r = fold (\y' z r' -> if y==y' then add x z r' else r') r yz
    f x y r = fold (\y' z -> if y==y' then add x z else id) r yz

--

unionr :: (Fold r' a b, Unfold r a b) => r -> r' -> r
unionr r r' = fold (\a b r'' -> add a b r'') r r'

unionl :: (Fold r a b, Unfold r' a b) => r -> r' -> r'
unionl r r' = fold (\a b r'' -> add a b r'') r' r

--

