{-# OPTIONS -fglasgow-exts -fallow-undecidable-instances #-}
-----------------------------------------------------------------------------
-- |
-- Module      :  Data.TermLanguage
-- Copyright   :  (c) PURe team @ Univ.Minho 2004
-- License     :  LGPL
-- 
-- Maintainer  :  jba @ di . uminho . pt
-- Stability   :  experimental
-- Portability :  portable
--
-- A module for handling Terms, Variables and Substitutions .
--
--
-----------------------------------------------------------------------------


module Data.TermLanguage where


import Data.List(nub, intersect)
  

-- | A class that handle terms (/t a/) with variables (/a/) and the
-- associated notion of substitution (/s a t/).
--
-- Note that /a/ has kind /*/, /t/ has kind /*->*/ and /s/ has 
-- kind /*->*->*/.

class (Eq a) => TermLanguage t s a | t->s where
  -- | Build a term from a variable
  varT :: a -> t a
  -- | Vars in a term
  varsIn :: t a -> [a]
  -- | Applies a substitution to a term
  appS :: s a t -> t a -> t a
  -- | Null Substitution
  nullS :: s a t
  -- | Singleton substitution
  unitS :: a -> t a -> s a t
  -- | Composition of two substitutions
  compS :: s a t -> s a t -> s a t
  -- | Gets the Domain of a substitution
  domS :: s a t -> [a]
  -- | 'unifyT' @t1@ @t2@ computes the "most general unifier" for terms 
  -- @t1@ and @t2@ (if it exists).
  -- That is, if 'unify' @t1@ @t2@ returns @Just s@ then
  --     @(appS s t1)==(appS s t2)@
  -- and, for every substitution @s1@ that equalizes both terms, it factors
  -- through @s@ (i.e. exists @s2@ such that @s1==comp s s2@).
  unifyT :: t a -> t a -> Maybe (s a t)
  -- | 'matchT' @t1@ @t2@ computes the "matcher substitution" for terms
  -- @t1@ and @t2@ (if it exists).
  -- That is, if 'match' @t1@ @t2@ returns @Just s@ then
  --     @(appS s t1)==t2@.
  matchT :: t a -> t a -> Maybe (s a t)


-- | (@s2@ '@@' @s1@) returns the substitution  @s2@ AFTER @s1@.
-- It is the infix operator for substitution composition.
-- However, remember that the arguments are swaped when compared with
-- 'compS', that is @compS s1 s2==s2 @@ s1@

--infixr 4 aa
--(@@) :: (TermLanguage t s a) => s a t -> s a t -> s a t
--s2 @@ s1 = compS s1 s2

-- | Get all vars in the range of a substitution.

vRngS :: (TermLanguage t s a) => s a t -> [a]
vRngS s = (nub . concat . map (varsIn.(appS s).varT) . domS) s 


-- | 'isIdempotentS' @s@ checks if @s@ is an idempotent substitution.

isIdempotentS :: (TermLanguage t s a) => s a t -> Bool
isIdempotentS s = null ((domS s) `intersect` (vRngS s))



-- | 'isInstanceOfT' @t1@ @t2@ checks if @t1@ is an instance of @t2@.

isInstanceOfT :: (TermLanguage t s a) => t a -> t a -> Bool
isInstanceOfT t1 t2 = case matchT t1 t2 of {Nothing -> False
					   ;Just _ -> True}
