{-# OPTIONS -fglasgow-exts -fallow-undecidable-instances #-}
-----------------------------------------------------------------------------
-- |
-- Module      :  Data.TermLanguage.AutoTermLanguage
-- Copyright   :  (c) PURe team @ Univ.Minho 2004
-- License     :  LGPL
-- 
-- Maintainer  :  jba @ di . uminho . pt
-- Stability   :  experimental
-- Portability :  portable
--
-- A generic implementations for Substitutions and Unification in
-- a TermLanguage.
--
-- Some /conventions/ used in this module:
--
--   *) functions with suffix @_@ are default methods that will instanciate
--   the corresponding methods on the "TermLanguage" class.
--
--   *) generic functions have prefix @g@.
--
--   *) in generic transversals with two arguments, a single @g@ in the
--   prefix means that both arguments have the same (generic) type, and
--   two @gg@s in the prefix mean a /trully generic/ variant.
--
-----------------------------------------------------------------------------

module Data.TermLanguage.AutoTermLanguage 
  (
   -- * Classes
   AutoTermLanguage,
   varConstr,
   gVarsIn,
   gGetVar,
   gAppS,
   gUnifyT,
   ggUnifyT,
   gMatchT,
   ggMatchT,
   -- * Types
   Subst,
   -- * Reexported Modules
   -- | The end-user will be interessed in using the functions from the "TermLanguage"
   -- module.
   module Data.TermLanguage,
   -- * Reexports from @Data.Generics@ to avoid extra imports clauses...
   -- | @Typeable,Data@
   Typeable,
   Data
   -- * Usage
   -- $usage
  )
where


import Data.Maybe
import Data.List(nub, intersect)
import Data.Generics
import Data.FiniteMap
import Control.Monad
import Control.Monad.State

import Data.TermLanguage


-- | Substitutions are implemented as @FiniteMap@s.

newtype Subst a t = Subst (FiniteMap a (t a))

instance (Show a,Show (t a)) => Show (Subst a t) where
  showsPrec _ (Subst s) = showString "{"
			  . showsFM (fmToList s)
			  . showString "}"
    where showsFM [] = id
	  showsFM [(k,e)] = shows k . showString "->"
			    . shows e
	  showsFM ((k,e):xs) = shows k . showString "->"
			       . shows e . showString " ;" . showsFM xs



-- | Main class for this module. The only method that does not have a default
-- definition is the 'varConstr' method. This method specifies the constructor
-- for variables in the "TermLanguage"

class (Ord a,Data a,Data (t a)) => AutoTermLanguage t a where
  -- | 'varConstr' is the name for the constructor that injects
  -- Variables in Terms.
  varConstr :: a -> t a
  -- | 'gGetVar' @_@ @x@ is a generic function that extracts a variable
  -- from a Term. That is, it returns @Just v@ when @x==(Var v)@ and
  -- @Nothing@ otherwise.
  -- The first argument for @gGetVar@ is a /dummy argument/ used to
  -- establish the intended Term type. It can be instanciated with
  -- @undefined@.
  gGetVar :: t a -> GenericQ (Maybe a)
  gGetVar tipo x | isVar_ tipo x = do (r::t a) <- cast x
				      (vx::a) <- gfindtype r
				      return vx
		 | otherwise = Nothing
  -- | 'gVarsIn' @_@ @t@ is a generic function that computes the list of
  -- variables in @t@. The first argument is /dummy/.
  gVarsIn :: t a -> (GenericQ [a])
  gVarsIn tipo t = nub $ map getArg (listify isVar t)
    where getArg x = case gfindtype x of
		       (Just y)->y
		       _ -> error "gVarsIn: unexpected error..."
          isVar :: t a -> Bool
	  isVar = isVar_ tipo
  -- | Null substituion
  nullS_ :: Subst a t
  nullS_ = Subst emptyFM
  -- | Unit substitution
  unitS_ :: a -> t a -> Subst a t
  unitS_ v t = Subst (unitFM v t)
  -- | Application of a substitution to a term.
  appS_ :: Subst a t -> t a -> t a
  appS_ = gAppS
  -- | Composition of two substituions
  compS_ :: Subst a t -> Subst a t -> Subst a t
  compS_ (Subst s1) (Subst s2) =
    Subst (foldFM (\k e s->addToFM s k (appS_ (Subst s2) e)) s2 s1)
  -- | Domain of a substitution
  domS_ :: Subst a t -> [a]
  domS_ (Subst s) = keysFM s
  -- | Unification of two terms
  unifyT_ :: t a -> t a -> Maybe (Subst a t)
  unifyT_ t1 t2 = ggUnifyT t1 t1 t2
  -- | Matching of two terms
  matchT_ :: t a -> t a -> Maybe (Subst a t)
  matchT_ t1 t2 = ggMatchT t1 t1 t2
  -- | One-layer substitution application
  atomicSubst_ :: Subst a t -> t a -> t a
  atomicSubst_ (Subst s) t =
    case gGetVar (undefined::t a) t of
      (Just z) -> case lookupFM s z of Nothing -> t
				       Just t' -> t'
      Nothing -> t



-------------------------------------------------------------------------
-- | Instanciation for the "TermLanguage" class
--

instance (AutoTermLanguage t a) => TermLanguage t Subst a
  where 
    varT = (varConstr::a ->t a)
    varsIn = gVarsIn (undefined::t a)
    appS = appS_
    nullS = nullS_
    unitS = unitS_
    compS = compS_
    domS = domS_
    unifyT = unifyT_
    matchT = matchT_

-- | Auxiliary generic function that checks if a term is a variable
-- (first argument is dummy)

isVar_ :: (AutoTermLanguage t a) => t a -> GenericQ Bool
{- no percebo porqu..., mas se tentarmos realizar a comparao ao nvel
dos constructores o programa entra em ciclo infinito...
        (toConstr t) == toConstr (varConstr tipo undefined)
-}
{- Nor do I understand, but with GHC 6.4.1 the string trick does not work anymore
isVar_ (tipo::t a) t = (conString (toConstr t))
		 == (conString (toConstr ((varConstr undefined)::t a)))
-}
isVar_ (tipo::t a) t 
   = (toConstr t) == (toConstr (varConstr::a -> t a))

-- | Generic application of a substitution

gAppS :: (AutoTermLanguage t a) => Subst a t -> GenericT
gAppS s = everywhere (mkT (atomicSubst_ s))


-- | Generic unification (first argument is dummy)

gUnifyT :: (AutoTermLanguage t a, Data x) => 
	   t a -> x -> x -> Maybe (Subst a t)
gUnifyT = ggUnifyT_ []


-- | Generic unification (first argument is dummy)

ggUnifyT :: (AutoTermLanguage t a) => 
	    t a -> GenericQ (GenericQ (Maybe (Subst a t)))
ggUnifyT tipo t1 t2 = ggUnifyT_ [] tipo t1 t2


-- | Generic unification and matching (first argument is dummy).

ggUnifyT_ :: (AutoTermLanguage t a) => 
	     [a] -> t a -> GenericQ (GenericQ (Maybe (Subst a t)))
ggUnifyT_ fset tipo t1 t2
  | isVar_ tipo t1 || isVar_ tipo t2
  = do --Maybe
      { v1 <- gGetVar tipo t1
      ; if case (cast t2) of Nothing -> False
			     Just t2' -> geq t1 t2'
        then return nullS_
        else do { guard (v1 `notElem` fset)
		; (x::t a) <- cast t2
		; guard (v1 `notElem` gVarsIn tipo x)
		; return (unitS v1 x)
		}
      }
    `mplus`
    do --Maybe
      { v2 <- gGetVar tipo t2
      ; guard (v2 `notElem` fset)
      ; (x::t a) <- cast t1
      ; guard (v2 `notElem` gVarsIn tipo x)
      ; return (unitS v2 x)
      }
  | toConstr t1 /= toConstr t2 = fail "Term do not unify"
  | otherwise = execStateT (gzipWithM aux t1 t2) nullS_
 where aux ta tb = do --State
		     { s1 <- get
		     ; let s2 = ggUnifyT_ fset tipo (gAppS s1 ta) (gAppS s1 tb)
		     ; guard (isJust s2)
		     ; put (compS_ s1 (fromJust s2))
		     ; return tb
		     }

-- | Generic matching (first argument is dummy)

gMatchT :: (AutoTermLanguage t a, Data x) => 
	   t a -> x -> x -> Maybe (Subst a t)
gMatchT tipo t1 t2 = ggUnifyT_ (gVarsIn tipo t2) tipo t1 t2


-- | Generic matching (first argument is dummy)

ggMatchT :: (AutoTermLanguage t a) => 
	    t a -> GenericQ (GenericQ (Maybe (Subst a t)))
ggMatchT tipo t1 t2 = ggUnifyT_ (gVarsIn tipo t2) tipo t1 t2



-----------------------------------------------------------------------------
-- $usage
-- To use the class "AutoTermLanguage" it is necessary:
--
--   (1) Define term constructor data type with the variable type has parameter
--   and make it instance of the @Data@ and @Typeable@ classes (include these
--   classes in the @deriving@ clause).
--
--   (2) Intanciate the class "AutoTermLanguage" specifying the
--   /var-constructor/.
--
-- /EXAMPLE OF USE:/
--
-- > import Data.TermLanguage.AutoTermLanguage
-- >
-- > data Term a = Var a | Const String | BOp String (Term a) (Term a) 
-- >   deriving (Eq, Show, Data, Typeable)
-- >
-- > instance AutoTermLanguage Term Int
-- >   where varConstr = Var
-- >
-- > t1,t2 :: Term Int
-- > ...
-- > mgu = unifyT t1 t2    --computes the most general unifier
-- >
--
-----------------------------------------------------------------------------

