{-# OPTIONS -fglasgow-exts #-}
{-# OPTIONS -fallow-undecidable-instances #-}

-----------------------------------------------------------------------------
-- |
-- Maintainer  :  joost.visser@di.uminho.pt, alcino@di.uminho.pt
-- Stability   :  experimental
-- Portability :  portable
--
-- See
--
--     Type-safe Two-level Data Transformation
-- 
-- by
--
--     Alcino Cunha, Jose Nuno Oliveira, Joost Visser
--
-----------------------------------------------------------------------------

module Data.Transform.TwoLevel where

import Data.Set as Set
import Data.Map as Map
import Data.List as List
import Control.Monad.State
import Data.Maybe
import qualified Data.HDB.CoddFish as CoddFish
import qualified Data.HListExtras as HList

-----------------------------------------------------------------------------
-- * Representation of Types

-- | A type-safe representation of types.
data Type a where

    -- Without fixpoints
    Int :: Type Int
    Bool :: Type Bool
    Char :: Type Char
    String :: Type String
    One :: Type ()
    List :: Type a -> Type [a]
    Set :: Type a -> Type (Set a)
    Map :: Type a -> Type b -> Type (Map a b)
    Either :: Type a -> Type b -> Type (Either a b)
    Prod :: Type a -> Type b -> Type (a,b)
    Tag :: String -> Type a -> Type a

    Mu :: Dist f => Fctr f -> Type (Mu f)
    
    X :: Type X   -- Only for printing!
    
    -- For databases
    HNil :: Type HList.HNil
    HCons :: Type a -> Type l -> Type (HList.HCons a l)

-- Implementing a type-level predicate on the value level
gIsList :: Type a -> Bool
gIsList HNil = True
gIsList (HCons a l) = gIsList l
gIsList _ = False

-- Implementing a type-level relation on the value level
gListContainsType :: Type l -> Type e -> Bool
gListContainsType HNil e = False
gListContainsType (HCons x l) e = (maybe False (const True) $ teq x e)||(gListContainsType l e)
gListContainsType _ _ = False

-- Implementing type-level functions on the value level
lastFromList :: Type l -> l -> Maybe Dynamic
lastFromList (HCons ta HNil) (HList.HCons a _) = Just (Dyn ta a)
lastFromList (HCons ta tl) (HList.HCons a l) = lastFromList tl l
lastFromList _ _ = Nothing

lastFromList' :: Typeable l => l -> Maybe Dynamic
lastFromList' = lastFromList typeof

appendLists :: (Type l,l) -> (Type l',l') -> Maybe Dynamic
appendLists (HNil,_) (tl',l') 
  | gIsList tl' = Just (Dyn tl' l')
  | otherwise = Nothing
appendLists (HCons tx tl,HList.HCons x l) y
  = case (appendLists (tl,l) y) of
      Just (Dyn tl' l') -> Just $ Dyn (HCons tx tl') (HList.HCons x l') 
      Nothing -> Nothing
appendLists _ _ = Nothing

{-
appendLists' :: (HList.HAppend l l' l'') => (Type l,l) -> (Type l',l') 
             -> Maybe (HAPPEND l l')
appendLists' (HNil,_) (tl',l') 
  | gIsList tl' = Just (HAPPEND l')
  | otherwise = Nothing
appendLists' _ _ = Nothing
-}
{-
gIsList' :: Type a -> Maybe (HLIST a)
gIsList' HNil = Just (HLIST HList.hNil)
gIsList' (HCons a l) = maybe Nothing aux (gIsList' l)
  where 
    aux (HLIST ll) = Just (HLIST (HList.HCons undefined ll))
gIsList' _ = Nothing
-}
data HList.HList x => HLIST x where
  HLIST :: (HList.HList l) => l -> HLIST l

data HAPPEND x y where
  HAPPEND :: (HList.HAppend l l' l'') => l'' -> HAPPEND l l'
  
data X

data Fctr f where
    ID    :: Fctr ID
    K     :: Type a -> Fctr (K a)
    (:+:) :: Fctr g -> Fctr h -> Fctr (g:+:h)
    (:*:) :: Fctr g -> Fctr h -> Fctr (g:*:h)
    (:@:) :: Functor g => (forall a . Type a -> Type (g a)) -> Fctr h -> Fctr (g:@:h)
    
myType = Mu ((K Int) :*: (List :@: ID))
    
instance Show (Type a) where
    show Int = "Int"
    show Bool = "Bool"
    show Char = "Char"
    show String = "String"
    show One = "One"
    show (List x) = "(List "++(show x)++")"
    show (Set x) = "(Set "++(show x)++")"
    show (Map x y) = "(Map "++(show x)++" "++(show y)++")"
    show (Either x y) = "(Either "++(show x)++" "++(show y)++")"
    show (Prod x y) = "(Prod "++(show x)++" "++(show y)++")"
    show (Tag n t) = "(Tag "++(show n)++" "++(show t)++")"

    show (Mu f) = "(Mu "++show f
    show X = "<X>"
    
instance Show (Fctr f) where
  show ID = "ID"
  show (K t) = "(K "++show t++")"
  show (f:*:g) = "("++show f++":*:"++show g++")"
  show (f:+:g) = "("++show f++":+:"++show g++")"
  show (f:@:g) = "(X."++show (f X)++":@:"++show g++")"
    
----------------------------------------
-- ** Some generic function

-- | Explicitly contrain a value to a particular type.
constrain :: Type a -> a -> a
constrain _ a = a

-- | Generic show function.
gshow :: Type a -> a -> String
gshow Int i = show i
gshow One _ = "()"
gshow Bool x = show x
gshow Char x = show x
gshow String x = show x
gshow (Map a b) m = "{"++(concat $ intersperse "," $ fmap aux (Map.toList m))++"}"
    where aux (x,y) = (gshow a x)++":="++(gshow b y)
gshow (Prod a b) (x,y) = "("++(gshow a x)++","++(gshow b y)++")"
gshow (Either a b) (Left y) = "Left ("++(gshow a y)++")"
gshow (Either a b) (Right y) = "Right ("++(gshow b y)++")"
gshow (List a) l = "["++(concat $ intersperse "," $ fmap (gshow a) l)++"]"
gshow (Set a) s = "{"++(concat $ intersperse "," $ fmap (gshow a) (Set.elems s))++"}"
gshow (Tag n a) x = n++"="++(gshow a x)

gshow (Mu f) (In x) = maybe "Mu <f>" id aux
  where
    aux :: Maybe String
    aux = do
      (View rep fmf) <- applyF f (Mu f)
      return $ gshow fmf $ (to rep) x

-- gshow ((:*:) g h a) (Pair ga ha) = "(Pair "++(gshow (g a) ga)++" "++(gshow (h a) ha)++")"
-- gshow (K a b) (Const x) = "(Const "++(gshow a x)++")"
-- gshow (ID a) (Id x) = "(Id "++(gshow a x)++")"
-- gshow ((:@:) g h a) (Comp x) = "(Comp "++(gshow (g (h a)) x)++")"

-- gshow (MU _ fx) (IN x) = "(IN x . "++(gshow fx x)++")"
-- gshow (X a) (PlaceHolder x) = "<"++(gshow a x)++">"
-- gshow Organization x = show x


-- | Generic comparison function.
--   Currently defined only for products and integers....
gcompare :: Type a -> a -> a -> Ordering
gcompare Int x y = compare x y
gcompare (Prod a b) (x1,y1) (x2,y2) 
 = let t = gcompare a x1 x2
   in case t of 
     EQ -> gcompare b y1 y2
     otherwise -> t

----------------------------------------
-- ** Equality of types

-- | Data type to express type-equality.    
data Equal a b where
    Eq :: Equal a a

-- | Test equality of types.
--   If the types are equal, a witness 'Eq' of the equality is returned.
teq :: Type a -> Type b -> Maybe (Equal a b)
teq Int Int = return Eq
teq Bool Bool = return Eq
teq Char Char = return Eq
teq One One = return Eq
teq String String = return Eq
teq (List a) (List b) = 
    do Eq <- teq a b
       return Eq
teq (Map a b) (Map c d) =
    do Eq <- teq a c
       Eq <- teq b d
       return Eq
teq (Prod a b) (Prod c d) =
    do Eq <- teq a c
       Eq <- teq b d
       return Eq
teq (Either a b) (Either c d) =
    do Eq <- teq a c
       Eq <- teq b d
       return Eq
teq (Tag _ a) b =
    do Eq <- teq a b
       return Eq
teq a (Tag _ b) =
    do Eq <- teq a b
       return Eq
teq (Mu f) (Mu g) = do
{-
  (View _ fx) <- applyF f X
  (View _ gx) <- applyF g X 
  Eq <- teq fx gx
  return Eq
-}
  error "Cannot determine equality for Mu's"
teq X X = return Eq
teq _ _ = Nothing


-----------------------------------------------------------------------------

data SQL a b = SQL (Alter a b) (Alter b a)

data Alter a b where
  NOP :: Alter a a
  ADDFIELD :: Alter a (a,b)



-----------------------------------------------------------------------------
-- * Encapsulation of type-changing transformations

-- | Pairs of bi-directional value-level data transformation functions.
data Rep a b = Rep {to :: a -> b, from :: b -> a}

-- | The View constructor expreses that a type a can be represented by
--   a type b, as witnessed by value-level functions that convert 
--   between these types. Note that the target type does not escape, 
--   which means that it is implicitly existentially qualified.
data View a where
    View :: Rep a b -> Type b -> View (Type a)

instance Show (View a) where
    show (View _ b) = "(View (Rep <to> <from>) "++(show b)++")"

-- | Print the string representation of the target type encapsulated in a view.
showType :: View a -> String
showType (View _ b) = show b

-- | Type-changing rewrite rules.
type Rule = forall a . Type a -> Maybe (View (Type a))

-----------------------------------------------------------------------------
-- * Strategy combinators for two-level transformations

-- | Sequential composition.
(>>>) :: Rule -> Rule -> Rule
(f >>> g) a = do View r b <- f a
                 View s c <- g b
                 return (View (comprep r s) c)

-- | Left-biased choice.
(|||) :: Rule -> Rule -> Rule
(f ||| g) x = f x `mplus` g x

-- | Do-nothing operation. 
--   Unit for sequential composition. 
--   Left-zero and right-unit for left-biased choice.
nop :: Rule
nop x = Just (View idrep x)

-- | Repeat until failure.
--   Zero or more repetitions.
many :: Rule -> Rule
many r = (r >>> many r) ||| nop

-- | Apply argument rule exactly once, at arbitrary depth.
once :: Rule -> Rule
once r Int  = r Int
once r Bool = r Bool
once r Char = r Char
once r One  = r One
once r String = r String
once r (List a) = r (List a) `mplus` 
                  (do View s b <- once r a
                      return (View (maprep s) (List b)))
once r (Set a) = r (Set a) `mplus` 
                 (do View s b <- once r a
                     return (View (maprep s) (Set b)))
once r (Map a b) = r (Map a b) `mplus`
                   (do View s c <- once r a
                       return (View (bimaprep s idrep) (Map c b))) `mplus`
                   (do View s c <- once r b
                       return (View (bimaprep idrep s) (Map a c)))
once r (Either a b) = r (Either a b) `mplus`
                      (do View s c <- once r a
                          return (View (bimaprep s idrep) (Either c b))) `mplus`
                      (do View s c <- once r b
                          return (View (bimaprep idrep s) (Either a c)))
once r (Prod a b) = r (Prod a b) `mplus`
                    (do View s c <- once r a
                        return (View (bimaprep s idrep) (Prod c b))) `mplus`
                    (do View s c <- once r b
                        return (View (bimaprep idrep s) (Prod a c)))
once r (Tag n a) = r (Tag n a) `mplus` 
                   (do View s b <- once r a
                       return (View s (Tag n b)))
once r X = mzero
once r x = r x

----------------------------------------
-- ** Value-level combinators used to define two-level combinators.

-- | Pair of identity value-level tranformations. Used to define 'nop'.
idrep :: Rep a a
idrep = Rep {to = id, from = id}

-- | Sequential composition of value-level transformations.
--   Used to define '(>>>)'.
comprep :: Rep a b -> Rep b c -> Rep a c
comprep f g = Rep {from = (from f) . (from g), to = (to g) . (to f)}

-- | Lift value-level transformations to a functor.
--   Used to define traversal strategies, like 'once'.
maprep :: Functor f => Rep a b -> Rep (f a) (f b)
maprep r = Rep {to = fmap (to r), from = fmap (from r)}

-- Additional instances of Functor. Needed because of maprep.
instance Functor Set where
    fmap = mapMonotonic

-- | Lift value-level transformations to a bi-functor.
--   Used to define traversal strategies, like 'once'.
bimaprep :: Bifunctor f => Rep a b -> Rep c d -> Rep (f a c) (f b d)
bimaprep x y = Rep {to = bimap (to x) (to y), from = bimap (from x) (from y)}

-- | Class of bi-functors. Needed for 'bimaprep'
class Bifunctor f where
    bimap :: (a -> b) -> (c -> d) -> f a c -> f b d

instance Bifunctor (,) where
    bimap = (><)

instance Bifunctor Either where
    bimap = (-|-)

instance Bifunctor Map where
    bimap f g = mapKeysMonotonic f . Map.map g

-----------------------------------------------------------------------------
-- * Unleashing composed data migration functions

-- | Apply the forward value-level function masked by a view.
forth :: View (Type a) -> Type b -> a -> Maybe b
forth (View rep tb') tb a = do {Eq <- teq tb tb'; return (to rep a)}

-- | Apply the backward value-level function masked by a view.
back :: View (Type a) -> Type b -> b -> Maybe a
back (View rep tb') tb b = do {Eq <- teq tb tb'; return (from rep b)}

-- | Take the forward and backward value-level functions out of
--   a view.
unView :: View (Type a) -> Type b -> (a -> Maybe b, b -> Maybe a)
unView (View rep tb') tb =
  ( \a -> do {Eq <- teq tb tb'; return (to rep a)}
  , \b -> do {Eq <- teq tb tb'; return (from rep b)} )

-----------------------------------------------------------------------------
-- * Rules for format evolution

----------------------------------------
-- ** Enrichment and removal

-- | Add a field (to the right).
--   The forward value-level function inserts a 
--   constant value for the new field.
addfield :: Type b -> b -> Rule
addfield b x a = Just (View rep (Prod a b))
    where rep = Rep {to = \y -> (y,x), from = fst}

-- | Add a field (to the left).
--   The forward value-level function inserts a 
--   constant value for the new field.
addfieldl :: Type b -> b -> Rule
addfieldl b x a = Just (View rep (Prod b a))
    where rep = Rep {to = \y -> (x,y), from = snd}

-- | Type of generic value-level queries.
type Query b = forall a . Type a -> a -> b

-- | Add a field.
--   The forward value-level function inserts a field
--   determined by a query on its input value.
addinfo :: Type b -> Query b -> Rule
addinfo b f a = Just (View rep (Prod a b))
    where rep = Rep {to = \y -> (y,f a y), from = fst}

----------------------------------------
-- ** Generalization and restriction

-- | Add alternative.
--   The backward value level function fails if the added 
--   alternative is present.
addalt :: Type b -> Rule
addalt b a = return $ View rep (Either a b)
    where rep = Rep {to = Left, from = \(Left x) -> x}

-- | Allow repetition.
--   Maps optional to a list.
allowRep :: Rule
allowRep (Either a One) = return $ View (Rep opt2seq seq2opt) (List a)
  where 
    opt2seq (Left x) = [x]
    opt2seq (Right ()) = []
    seq2opt []  = Right ()
    seq2opt [x] = Left x
    seq2opt _ = error "Too many elements in list for conversion to optional"
allowRep _ = mzero

-- | Lifted version of 'allowRep', works on functors.
allowRep1 :: Rule
allowRep1 a = return $ View rep (Mu ((K a) :+: ((K a) :*: ID)))
  where 
    rep = Rep singleton (nehead . out)
    singleton x = In $ Inl $ Const x
    nehead (Inl x) = unConst x
    nehead (Inr y) = error "Too many elements in list+"

----------------------------------------
-- ** Focus on tags

-- | Perform the argument rule inside the given tag.
inside :: String -> Rule -> Rule
inside n r (Tag m a) | n==m = do View r b <- r a
                                 return $ View r (Tag m b)
inside _ _ _ = Nothing

-- | Perform the argument rule on the given tag.
atTag :: String -> Rule -> Rule
atTag n r x@(Tag m a) | n==m = r x
atTag _ _ _ = Nothing

-----------------------------------------------------------------------------
-- * Rules for data mapping

-- | Flatten a hierarchical data type.
--   Does not include the rule for recursion elimination.
flatten :: Rule
flatten = many (aux >>> many aux >>> prods)
    where aux = once (listmap ||| mapsum ||| summap ||| mapprodmap ||| setmap ||| prodsum ||| sumone ||| prodone)

-- | Map a hierarchical datatype to a relational type.
toRDB :: Rule
toRDB = many (once fixastable) >>> many (once enum2int) >>> removetags >>> flatten

----------------------------------------
-- ** Elimination and introduction

-- | Eliminate list by refinement to map.
listmap :: Rule
listmap (List a) = Just (View rep (Map Int a))
  where
    rep = Rep {to = seq2index, from = list}
    list :: Map Int a -> [a]
    list = Map.elems
    seq2index :: [a] -> Map Int a
    seq2index = Map.fromList . (zip [0..])
listmap _ = Nothing

-- | Eliminate set by refinement to map.
setmap :: Rule
setmap (Set a) = return $ View rep (Map a One) 
  where 
    rep :: Rep (Set a) (Map a ())
    rep = Rep {to = set2fm, from = dom}
    set2fm s = Map.fromDistinctAscList (zip (Set.toAscList s) (repeat ()))
    dom = Set.fromDistinctAscList . keys
setmap _ = Nothing

-- | Eliminate optional by refinement to map.
sumone :: Rule
sumone (Either a One) = return $ View plusone (Map One a)
 where
  plusone :: Rep (Either a ()) (Map () a)
  plusone = Rep {to = optintro, from = optelim}
    where optintro (Left x) = Map.singleton () x
	  optintro (Right ()) = Map.empty
	  optelim m | Map.member () m = let Just x = Map.lookup () m in Left x
		    | otherwise = Right ()
sumone _ = fail "Not applicable"




----------------------------------------
-- ** Distribution

-- | Distribute map over sum.
mapsum :: Rule
mapsum (Map a (Either b c)) = return $ View rep (Prod (Map a b) (Map a c))
  where
    rep = Rep {to = uncojoin, from = cojoin a}
    uncojoin :: Map a (Either b c) -> (Map a b, Map a c)
    uncojoin = (  fmap (either id undefined) . Map.fromDistinctAscList >< 
                  fmap (either undefined id) . Map.fromDistinctAscList) .
               List.partition (isLeft . snd) . Map.toAscList
    -- Pre: non duplicate keys
    cojoin :: Type a -> (Map a b, Map a c) -> Map a (Either b c)
    cojoin t = Map.fromDistinctAscList . 
               sortBy (\x y -> gcompare t (fst x) (fst y)) .
               uncurry (++) . 
               (Map.toAscList . fmap Left >< Map.toAscList . fmap Right)
mapsum _ = Nothing

-- | Distribute product over sum.
prodsum :: Rule
prodsum (Prod a (Either b c)) = return $ View distr (Either (Prod a b) (Prod a c))
 where
  distr :: Rep (a,Either b c) (Either (a,b) (a,c))
  distr = Rep {from = undistr, to = distr}
    where distr (x, Left y) = Left (x,y)
          distr (x, Right y) = Right (x,y)
          undistr (Left (x,y)) = (x, Left y)
          undistr (Right (x,y)) = (x, Right y)
prodsum (Prod (Either b c) a) = return $ View distl (Either (Prod b a) (Prod c a))
 where
  distl :: Rep (Either b c, a) (Either (b,a) (c,a))
  distl = Rep {from = undistl, to = distl}
    where distl (Left y,x) = Left (y,x)
          distl (Right y,x) = Right (y,x)
          undistl (Left (y,x)) = (Left y,x)
          undistl (Right (y,x)) = (Right y,x)
prodsum _ = Nothing

summap :: Rule
summap (Map (Either a b) c) = return $ View eithermap (Prod (Map a c) (Map b c))
  where
    eithermap :: Rep (Map (Either a b) c) (Map a c, Map b c)
    eithermap = Rep {to = unpeither, from = peither}
    unpeither = ( mapKeysMonotonic (either id undefined) . Map.fromDistinctAscList >< 
                  mapKeysMonotonic (either undefined id) . Map.fromDistinctAscList
                ) . List.partition (isLeft . fst) . Map.toAscList
    peither = Map.fromDistinctAscList . uncurry (++) . 
              (fmap (Left >< id) . Map.toAscList >< fmap (Right >< id) . Map.toAscList)
summap _ = Nothing





----------------------------------------
-- ** Split, join

-- | Split and join.
mapprodmap :: Rule
mapprodmap (Map a (Prod b (Map c d))) 
  = return $ View (mapcomposite a) (Prod (Map a b) (Map (Prod a c) d))
 where
  -- Precondition to from: all keys a in second table must exist in the first
  mapcomposite :: Type a -> Rep (Map a (b, Map c d)) (Map a b, Map (a,c) d)
  mapcomposite t = Rep {to = unnjoin, from = njoin}
    where unnjoin = (Map.map fst /\ to (mapcomp t) . Map.map snd)
          njoin = Map.fromDistinctAscList . uncurry myzip . (Map.toAscList >< Map.toAscList . from (mapcomp t))
          myzip l [] = fmap (id >< (id /\ const Map.empty)) l
          myzip ((x,y):l) ((z,w):m) | gcompare t x z == EQ = (x,(y,w)) : myzip l m
                                    | otherwise = (x,(y,Map.empty)) : myzip l ((z,w):m)
  unnjoin :: Map a (Map b c) -> Map (a,b) c
  unnjoin =  Map.fromDistinctAscList . concat . 
             List.map ((\ (x,l) -> List.map ((const x /\ id) >< id) l) . 
             (id >< Map.toAscList)) . 
             Map.toAscList
  njoin :: Type a -> Map (a,b) c -> Map a (Map b c) 
  njoin t = Map.fromDistinctAscList . 
            fmap (fst.fst.head /\ Map.fromDistinctAscList . fmap (snd >< id)) . 
            groupBy aux . Map.toAscList
    where aux x y = gcompare t (fst (fst x)) (fst (fst y)) == EQ
  mapcomp :: Type a -> Rep (Map a (Map c d)) (Map (a,c) d)
  mapcomp t = Rep {to = unnjoin, from = njoin t}
mapprodmap _ = Nothing

----------------------------------------
-- ** Recursion

-- | Type of tables.
type Table f = (Map Int (f Int), Int)

-- | Refine fix-point datatype to table type.
fixastable :: Rule
fixastable (Mu f) = do
  (View (Rep t f) fInt) <- applyF f Int
  let rep = Rep {to = (Map.map t >< id) . fixtotable, from = tabletofix . (Map.map f >< id)}
  return (View rep (Prod (Map Int (fInt)) Int)) 
   where
    fixtotable :: Dist f => Mu f -> Table f
    fixtotable mu = execState (mcata aux mu) (Map.empty,-1)
      where 
        aux :: f Int -> State (Table f) Int
        aux i = do
          (t,k) <- get
          put (Map.insert (k+1) i t, k+1)
          return (k+1)
    -- Precondition to abs: non circular and all keys exist
    tabletofix :: Functor f => Table f -> Mu f
    tabletofix = ana aux
      where 
        aux :: Functor f => Table f -> f (Table f)
        aux (t,k) = fmap (\k -> (t,k)) (fromJust $ Map.lookup k t)
fixastable _ = Nothing

-- | Unlift a functor while applying it to a type.
--   Previously: removefunctors.
applyF :: Fctr f -> Type a -> Maybe (View (Type (f a)))
applyF ID x        = return (View (Rep {to = unId, from = Id}) x)
applyF (K a) x     = return (View (Rep {to = unConst, from = Const}) a)
applyF (g:*:h) x   = do
  (View (Rep g2gx gx2g) gx) <- applyF g x
  (View (Rep h2hx hx2h) hx) <- applyF h x
  let rep = Rep {to = \(Pair x y) -> (g2gx x,h2hx y), from = \(x,y) -> Pair (gx2g x) (hx2h y)}
  return (View rep (Prod gx hx))
applyF (g:+:h) x = do
  (View (Rep g2gx gx2g) gx) <- applyF g x
  (View (Rep h2hx hx2h) hx) <- applyF h x
  let rep = Rep {to = (g2gx-|-h2hx) . aux, from = xua . (gx2g-|-hx2h)}
  return (View rep (Either gx hx))
      where 
        aux (Inl x) = Left x
        aux (Inr x) = Right x
        xua (Left x)  = Inl x
        xua (Right x) = Inr x
applyF (g:@:h) x   = do 
  (View (Rep to from) hx) <- applyF h x
  let rep = Rep (fmap to . unComp) (Comp . fmap from)
  return (View rep (g hx))

-- | Derive the functor for a given type.  
class Functor f => FunctorOf t f | t -> f where
  functorOf :: Type t -> Fctr f
instance FunctorOf X ID where
  functorOf _ = ID
instance FunctorOf Int (K Int) where
  functorOf _ = K Int
instance FunctorOf () (K ()) where
  functorOf _ = K One
instance (FunctorOf a f, FunctorOf b g)
 => FunctorOf (a,b) (f:*:g) where
  functorOf (Prod a b) = (functorOf a :*: functorOf b)
instance (FunctorOf a f, FunctorOf b g)
 => FunctorOf (Either a b) (f:+:g) where
  functorOf (Either a b) = (functorOf a :+: functorOf b)
instance FunctorOf a f
 => FunctorOf [a] ([]:@:f) where
  functorOf (List a) = List :@: functorOf a


-----------------------------------------------------------------------------
-- * Associativity and commutativity

-- | Associate product to the right. One step.
assocr :: Rule
assocr (Prod (Prod a b) c) = return $ View aux (Prod a (Prod b c))
    where aux = Rep {to = fst.fst /\ snd><id, from = id><fst /\ snd.snd}
assocr _ = fail "Cannot apply!"

-- | Associate product to the left. One step.
assocl :: Rule
assocl (Prod a (Prod b c)) = return $ View aux (Prod (Prod a b) c)
    where aux = Rep {from = fst.fst /\ snd><id, to = id><fst /\ snd.snd}
assocl _ = fail "Cannot apply!"

-- | Put maps to the right and non-maps to the left.
--   Needs to be redesigned....
prods :: Rule
prods = many (many (once assocr) >>> once swap) >>> many (once assocl)
 where
   swap :: Rule
   swap (Prod a (Map b c)) = fail "Cannot apply!"
   swap (Prod a (Prod (Map b c) d)) = fail "Cannot apply!"
   swap (Prod (Map a b) c) = return $ View (Rep {from = snd/\fst, to = snd/\fst}) (Prod c (Map a b))
   swap _ = fail "Cannot apply!"

-- | Right unit for product.
prodone :: Rule
prodone (Prod a One) = return $ View rep a
 where
  rep :: Rep (a,()) a
  rep = Rep {to = fst, from = \x -> (x,())}
prodone _ = fail "Not applicable"


-----------------------------------------------------------------------------
-- * Miscelaneous rules

-- | Refine an n-sum of taged ones to an integer.
--   In other words, interpret the n-sum as an enumeration type.
enum2int :: Rule
enum2int = (many (once eitherRightAssoc)) >>> e2i 0 where
  e2i :: Int -> Rule
  e2i i (Tag tg One) = do
    -- return (View (Rep (\() -> i) (\i' -> if i==i' then () else error ("Invalid enum: "++(show i)))) Int)
    return (View (Rep (\() -> i) (ifeq () (error "Enum invalid") i)) Int)
  e2i i (Either (Tag tg One) b) = do
    (View rep b) <- e2i (i+1) b
    Eq <- teq Int b
    -- return (View (Rep (either (\() -> i) (\b -> ((to rep) b)+1)) (\i' -> if i==i' then Left () else Right ((from rep) i'))) b)
    return (View (Rep (either (\() -> i) ((to rep))) (ifeq (Left ()) (Right . (from rep)) i)) b)
  e2i _ _ = Nothing
ifeq l r i i' = if i==i' then l else r i'

-- | Associate sums to the right.
--   Used by 'enum2int' to discover enumeration types.
eitherRightAssoc :: Rule
eitherRightAssoc (Either (Either a b) c) = return (View (Rep to from) (Either a (Either b c)))
  where
      to (Left (Left a)) = Left a
      to (Left (Right b)) = Right (Left b)
      to (Right c) = Right (Right c)
      from (Left a) = Left (Left a)
      from (Right (Left b)) = Left (Right b)
      from (Right (Right c)) = Right c
eitherRightAssoc _ = Nothing

-- | Remove all tags from a type represention. This is useful to
--   allow later rules to assume no tags are present.
removetags :: Rule
removetags = many (once aux)
    where aux (Tag n x) = Just (View idrep x)
	  aux _ = Nothing

-----------------------------------------------------------------------------
-- * Representation of recursive datatypes.

----------------------------------------
-- ** Standard functors

infixr 5 :+:
infixr 6 :*:
infixr 9 :@:

newtype ID a        = Id {unId :: a}             deriving Eq
newtype K b a       = Const {unConst :: b}       deriving Eq
data (g :+: h) a    = Inl (g a) | Inr (h a)      deriving Eq
data (g :*: h) a    = Pair (g a) (h a)           deriving Eq
newtype (g :@: h) a = Comp {unComp :: g (h a)}   deriving Eq

instance Functor ID where
    fmap f = Id . f . unId

instance Functor (K a) where
    fmap f = Const . id . unConst

instance (Functor f, Functor g) => Functor (f :+: g) where
    fmap f (Inl x) = Inl (fmap f x)
    fmap f (Inr x) = Inr (fmap f x)

instance (Functor f, Functor g) => Functor (f :*: g) where
    fmap f (Pair x y) = Pair (fmap f x) (fmap f y)

instance (Functor g, Functor h) => Functor (g :@: h)
    where fmap f = Comp . fmap (fmap f) . unComp

----------------------------------------
-- ** Fixpoints and recursion patterns

-- | Explicit fixpoint operator
newtype Mu f = In {out :: f (Mu f)}

-- This instance requires -fallow-undecidable-instances.
instance Eq (f (Mu f)) => Eq (Mu f) where
  (In fmf) == (In gmg) = fmf==gmg

-- | Catamorphism or fold
cata :: Functor f => (f a -> a) -> Mu f -> a
cata g = g . fmap (cata g) . out

-- | Anamorphism or unfold
ana :: Functor f => (a -> f a) -> a -> Mu f
ana g = In . fmap (ana g) . g

-- | 
nu :: Functor g => (forall a. f a -> g a) -> Mu f -> Mu g
nu t = In . fmap (nu t) . t . out

-- |
nu' :: Functor f => (forall a. f a -> g a) -> Mu f -> Mu g
nu' t = In . t . fmap (nu' t) . out

----------------------------------------
-- ** Functors and monads

-- | Class of functors that distribute over monads.
class Functor f => Dist f where
    dist :: Monad m => f (m a) -> m (f a)

instance Dist ID where
    dist (Id x) = do y <- x
		     return (Id y)

instance Dist (K a) where
    dist (Const x) = return (Const x)

instance (Dist f, Dist g) => Dist (f :+: g) where
    dist (Inl x) = do y <- dist x
		      return (Inl y)
    dist (Inr x) = do y <- dist x
		      return (Inr y)

instance (Dist f, Dist g) => Dist (f :*: g) where
    dist (Pair l r) = do x <- dist l
			 y <- dist r
			 return (Pair x y)

instance (Dist g, Dist h) => Dist (g :@: h) where
    dist (Comp x) = do y <- return $ fmap dist x
		       z <- dist y
		       return (Comp z)

instance Dist [] where
    dist = sequence


-- | Monadic catamorphism.
mcata :: (Monad m, Dist f) => (f a -> m a) -> Mu f -> m a
mcata g = cata (\x -> dist x >>= g)


-----------------------------------------------------------------------------
-- * Auxiliary types and functions

-- | Determine whether the left alternative is present.
isLeft :: Either a b -> Bool
isLeft = either (const True) (const False)

-- | Type of non-empty lists
data List1 a    -- Mu ((K a) :+: ((K a) :*: ID))
  = Singleton a 
  | Cons1 { head1 :: a, tail1 :: (List1 a) }

----------------------------------------
-- ** Pointfree operators

infix 5 ><
infix 4 /\
infix 3 -|-
infix 2 \/

(f >< g) (x,y) = (f x, g y)
(/\) f g x = (f x, g x)
(f -|- g) (Left x) = Left (f x)
(f -|- g) (Right x) = Right (g x)
(\/) f g (Left x) = f x
(\/) f g (Right x) = g x


-----------------------------------------------------------------------------
-- * Deriving type representations from types.

-- | Typeable class of types for which we can derive representations.
class Typeable a where
    typeof :: Type a

instance Typeable Int where
    typeof = Int

instance Typeable () where
    typeof = One

instance Typeable Bool where
    typeof = Bool

instance Typeable Char where
    typeof = Char

instance (Typeable a, Typeable b) => Typeable (a,b) where
    typeof = Prod typeof typeof

instance (Typeable a, Typeable b) => Typeable (Either a b) where
    typeof = Either typeof typeof

instance (Typeable a, Typeable b) => Typeable (Map a b) where
    typeof = Map typeof typeof

instance Typeable String where
    typeof = String

instance Typeable a => Typeable [a] where
    typeof = List typeof

instance Typeable a => Typeable (Set a) where
    typeof = Set typeof

instance Typeable HList.HNil where
  typeof = HNil
  
instance (Typeable a, Typeable l) => Typeable (HList.HCons a l) where
  typeof = HCons typeof typeof
    
-----------------------------------------------------------------------------
-- * Using dynamic types

-- | Dynamic types
data Dynamic where
    Dyn :: Type a -> a -> Dynamic

-- | Apply a generic function to a dynamically typed value.
applyDyn :: (forall a . Type a -> a -> b) -> Dynamic -> b
applyDyn f (Dyn ta a) = f ta a

-- | Turn a function on dynamics into a generic function.
dynApply :: (Dynamic -> b) -> Type a -> a -> b
dynApply f ta a = f (Dyn ta a)

instance Show Dynamic where
    show (Dyn a x) = "Dyn "++(show a)++" "++(gshow a x)
    
-- | Variation on 'forth' that does not need static knowledge of the
--   target type, and produces a dynamically typed value instead.
forthDyn :: View (Type a) -> a -> Maybe Dynamic
forthDyn (View rep b) x = return (Dyn b (to rep x))

-- | Variation on 'back' that does not need static knowledge of the
--   original type, and produces a dynamically typed value instead.
backDyn :: View (Type a) -> Dynamic -> Maybe a
backDyn (View rep b) (Dyn c x) = do
  Eq <- teq b c
  return (from rep x)


-----------------------------------------------------------------------------




