{-# OPTIONS -fglasgow-exts -fallow-overlapping-instances #-}

module Data.Unification where

import Data.Generics
import Data.Maybe
import Control.Monad.State

-- Unification a la Typing Haskell in Haskell but using Generics

class HasVariables a
    where isVar :: a -> Bool

type Subst a = [(a,a)]

instance Show a => Show (Subst a)
    where show c = let aux = map (\(v,e) -> (show v) ++ " := " ++ (show e)) c
		       in foldr (\a b -> a ++ "\n" ++ b) "" aux

empty :: Subst a
empty = []

unit :: a -> a -> Subst a
unit v e = [(v, e)]

vars :: (HasVariables a, Data a) => a -> [a]
vars = everything (++) (mkQ [] aux)
    where aux :: HasVariables a => a -> [a]
          aux i | isVar i = [i]
		| otherwise = []

apply :: (HasVariables a, Eq a, Data a, Data b) => Subst a -> b -> b
apply s e = foldr (\x f -> everywhere (mkT (aux x)) f) e s
    where aux :: (HasVariables a, Eq a) => (a,a) -> a -> a
	  aux (v,a) b | (isVar b) && (b == v) = a
		      | otherwise = b

infix 4 @@

(@@) :: (HasVariables a, Eq a, Data a) => Subst a -> Subst a -> Subst a
s @@ q = (apply s q) ++ s


unify :: (HasVariables a, Eq a, Data a) => a -> a -> Maybe (Subst a)
unify = unify'
    where  unify' :: (HasVariables a, Eq a, Data a,
		      HasVariables b, Eq b, Data b,
		      HasVariables c, Eq c, Data c) 
		  => a -> b -> Maybe (Subst c)
	   unify' l r | isVar l = unifyVar l r
		      | isVar r = unifyVar r l
		      | (toConstr l) == (toConstr r) = execStateT (gzipWithM (mkMM (aux l r)) l r) empty
		      | otherwise = fail "Terms do not unify"
           aux :: (HasVariables a, Eq a, Data a, 
		   HasVariables b, Eq b, Data b,
		   HasVariables c, Eq c, Data c) 
	       => a -> b -> a -> b -> StateT (Subst c) Maybe b
	   aux _ _ l r = do s1 <- get
			    let s2 = unify' (apply s1 l) (apply s1 r)
			    guard (isJust s2)
			    put ((fromJust s2) @@ s1)
			    return r

unif :: (HasVariables a, Eq a, Data a) => a -> a -> Maybe (Subst a)
unif l r | isVar l = unifyVar l r
	 | isVar r = unifyVar r l
	 | (toConstr l) == (toConstr r) = execStateT (tmapM' aux l r) empty
	 | otherwise = fail "Terms do not unify"

aux :: (HasVariables a, Eq a, Data a) => a -> a -> StateT (Subst a) Maybe a
aux l r = do s1 <- get
	     let s2 = unif (apply s1 l) (apply s1 r)
             guard (isJust s2)
	     put ((fromJust s2) @@ s1)
	     return r

tmapM' :: (Data a, Data b, Monad m) => (b -> b -> m b) -> a -> a -> m a
tmapM' f = gzipWithM (mkMM f)

mkMM :: (Typeable a, Typeable b, Typeable c, Typeable d, Monad m) 
     => (a -> b -> m b) -> c -> d -> m d
mkMM f x y = case (cast x, cast y)
	     of (Just x', Just y') -> do r <- (f x' y')
					 return (fromJust (cast r))
		_  -> return y

unifyVar :: (Data a, HasVariables b, Eq b, Data b, Data c) 
         => a -> b -> Maybe (Subst c)
unifyVar v e | (isVar e) && (e == fromJust (cast v)) = return empty
	     | (fromJust (cast v)) `elem` (vars e) = fail "Terms do not unify"
             | otherwise = return (unit (fromJust (cast v)) (fromJust (cast e)))

-- Example

data E = V Int | A | B E E | C Int 
	 deriving (Show,Eq,Typeable,Data)

instance HasVariables E
    where isVar (V _) = True
	  isVar _     = False

e1 = B A (V 1)

e2 = B (B (V 1) (V 2)) (V 3)

e3 = B (B A (B A A)) (V 2)

e4 = V 1

e5 = C 4

