-----------------------------------------------------------------------------
-- |
-- Module      :  Language.HaLex.Dfa
-- Copyright   :  (c) Joo Saraiva 2005
-- License     :  LGPL
-- 
-- Maintainer  :  jas@di.uminho.pt
-- Stability   :  provisional
-- Portability :  portable
--
-- Deterministic Finite Automata in Haskell: Examples
-- 
-- Code Included in the Lecture Notes on 
--      Language Processing (with a functional flavour).   
--
-----------------------------------------------------------------------------

module Language.HaLex.DfaUtils ( lowerLetters
                               , islower
                               , capitalLetters
                               , isupper
                               , isletter 
                               , digits
                               , isdigit
                               , intDen
                               , isint
                               , space
                               , isspace
                               , ident
                               , isident
                               , identWithSpacesAtEnd
                               ) where
-----------------------------------------------------------------------------
import List
import Char
import Language.HaLex.RegExp
import Language.HaLex.Dfa
import Language.HaLex.Ndfa
import Language.HaLex.RegExp2Fa
import Language.HaLex.FaOperations
import Language.HaLex.FaAsDiGraph
import Language.HaLex.Minimize
import Language.HaLex.RegExpAsDiGraph
import Language.HaLex.Fa2RegExp
import qualified Language.HaLex.FaClasses as Fa
-----------------------------------------------------------------------------

asciiTable = map chr [0..255]


-- | 'Dfa' defining the language of lower letters

lowerLetters :: Dfa Int Char
lowerLetters = Dfa asciiTable [1,2,3] 1 [2] delta
  where delta 1 x | x `elem` ['a'..'z'] = 2
        delta _ _                       = 3

-- | Predicate to define whether a ascii symbol is a lower letter or not

islower :: Char -> Bool
islower l = dfaaccept lowerLetters [l]

-- | 'Dfa' defining the language of capital letters

capitalLetters :: Dfa Int Char
capitalLetters = Dfa asciiTable [1,2,3] 1 [2] delta
  where delta 1 x | x `elem` ['A'..'Z'] = 2
        delta _ _                       = 3

-- | Predicate to define whether a ascii symbol is a capital letter or not
isupper l = dfaaccept capitalLetters [l]

-- | Predicate to define whether a ascii symbol is a letter or not

isletter :: Char -> Bool
isletter    l = dfaaccept (ndfa2dfa (lowerLetters `unionDfa` capitalLetters)) [l]

isletter''' x = (islower x) || (isupper x)
isletter'   l = ndfaaccept ( lowerLetters `unionDfa` capitalLetters) [l] 


-- | 'Dfa' defining the language of digits

digits :: Dfa Int Char
digits = Dfa asciiTable [1,2,3] 1 [2] delta
  where delta 1 x | x `elem` ['0'..'9'] = 2
        delta _ _                       = 3

-- | Predicate to define whether a ascii symbol is a digit or not

isdigit :: Char -> Bool
isdigit d = dfaaccept digits [d]


-- | 'Dfa' defining the language of integer numbers

intDen :: Dfa Int Char
intDen = Dfa asciiTable [1,2,3,4] 1 [3] delta
  where delta 1 '+'                     = 2
        delta 1 '-'                     = 2
        delta 2 x | x `elem` ['0'..'9'] = 3 
        delta 3 x | x `elem` ['0'..'9'] = 3 
        delta _ _                       = 4

-- | Predicate to define whether a list of ascii symbols form a integer number or not

isint :: [Char] -> Bool
isint = dfaaccept intDen


-- | 'Dfa' defining the language of spaces (newlines, tabs and spaces)

space :: Dfa Int Char
space  = Dfa asciiTable [1,2,3] 1 [2] delta
  where delta 1 x | x `elem` [' ','\n','\t'] = 2
        delta _ _                            = 3

-- | Predicate to define whether a ascii symbol is a space or not

isspace :: Char -> Bool
isspace s = dfaaccept space [s]


-- | 'Dfa' defining a language of identifiers

ident :: Dfa Int Char
ident = Dfa asciiTable [1,2,3,4] 1 [3] delta
  where delta 1 x | isletter  x  = 2
        delta 2 _                = 3
        delta 3 _                = 3
        delta _ _                = 4
        
isident :: [Char] -> Bool
isident = dfaaccept ident
        
-- | 'Dfa' defining a language of identifiers with spaces at the end.

identWithSpacesAtEnd :: Dfa Int Char
identWithSpacesAtEnd =  Dfa asciiTable [1,2,3,4,5] 1 [3,4] delta
  where delta 1 x | isletter x  = 2                      
        delta 2 _               = 3

        delta 3 x | isspace  x  = 4
                  | otherwise   = 3

        delta 4 x | isspace x   = 4

        delta _ _               = 5
        

