-----------------------------------------------------------------------------
-- |
-- Module      :  Language.HaLex.Edfa
-- Copyright   :  (c) Jooo Saraiva 2001,2002,2003,2004,2005
-- License     :  LGPL
-- 
-- Maintainer  :  jas@di.uminho.pt
-- Stability   :  provisional
-- Portability :  portable
--
-- Extended Finite Automata in Haskell. 
--
-- Code Included in the Lecture Notes on 
--      Language Processing (with a functional flavour).   
--
-----------------------------------------------------------------------------

module Language.HaLex.Edfa ( 
              -- * Data type 
                Edfa (..)
              -- * Acceptance
              , edfaaccept 
              -- * Transformation
              , transitionTableEdfa 
              , nonEmptytransitions
              , edfa2ndfa 
              , renameEdfa
              , addMove2Edfa
              -- * Conversion
              , dfa2edfa
              , ndfa2edfa
              , re2edfa
              -- * Printing
              , edfa2graphviz
              , edfa2graphviz2file
              -- * Elimination
              , elimStateEdfa
              , minimizeEdfa 
              , minimizeEdfa'
              , lstminEdfa2lstEdfaWithSize
              , minimizeRegExp 
              , regExp2Edfas
              , elimLstStatesEdfa
              , elimLstStatesEdfa'
              , initialEdfaStates
              , addNewStartStEdfa
              , addNewFinalStEdfa 
              , edfanumberIncomingArrows
              , edfanumberOutgoingArrows
              , edfadestinationsFrom
              , sizeOfMinimizedEdfa
           ) where

import List
import Language.HaLex.Util
import Language.HaLex.RegExp
import Language.HaLex.Ndfa
import Language.HaLex.Dfa
import Language.HaLex.RegExp2Fa
import Language.HaLex.Minimize
import Language.HaLex.Fa2RegExp
import Language.HaLex.FaAsDiGraph

-----------------------------------------------------------------------------
-- * Data type 

-- | The type of Extended Deterministic Finite Automata. Parameterized with
--   the type @st@ of states.
data Edfa st sy = Edfa [sy]                      -- Vocabulary
                       [st]                      -- Finite set of states
                       st                        -- The start state
                       [st]                      -- The set of final states
                       (st -> st -> RegExp sy )  -- Transition Function

-----------------------------------------------------------------------------
-- * Printing

-- | Print a 'Edfa' as a Haskell function.

instance (Show st, Show sy) => Show (Edfa st sy) where
      showsPrec p (Edfa v q s z delta) =
                showString ("edfa = Edfa v q s z delta") .
                showString ("\n  where \n\t v = ") .
                showList v . 
                showString ("\n\t q = ") .
                showList q . 
                showString ("\n\t s = ") .
                shows s . 
                showString ("\n\t z = ") .
                showList z . 
                showString ("\n\t -- delta :: st -> st -> RegExp sy \n") .
                showEdfaDelta q delta

-- | Helper function to show the transition function of a 'Edfa'.

showEdfaDelta :: (Show st, Show sy) 
             => [st] -> (st -> st -> RegExp sy) -> [Char] -> [Char]

showEdfaDelta q d = foldr (.) (showChar '\n') f
  where
    f     = zipWith3 showF m n q'
    (m,n) = unzip l
    q'    = map (uncurry d) l
    l     = [(a,b) | a <- q , b <- q]
         
    showF st st' Empty = \x -> x                  -- to not show "empty" transitions 
    showF st st' re = showString("\t delta ") .
                      shows st .
                      showChar(' ') .
                      shows st' .
                      showString(" = ") .
                      shows re .
                      showChar('\n')

-----------------------------------------------------------------------------
-- * Transformation

-- | Produce the transition table of a given 'Edfa'. 
--   Given a 'Edfa', it returns a list of triples of the form
-- @
--                 (Origin,RegExp,Destination)
-- @
-- defining all the transitions of the 'Edfa'.
--
transitionTableEdfa :: Edfa st sy                -- ^ Automaton
                    -> [(st,RegExp sy,st)]       -- ^ Transition table
transitionTableEdfa (Edfa v q s z delta) = [ (ao , delta ao ad , ad)  
                                           | ao <- q 
                                           , ad <- q
                                           ]

-- | Produce the transition table filtering out the Empty transitions

nonEmptytransitions tt = filter notEmptytrans (transitionTableEdfa tt)
  where notEmptytrans (_,Empty,_) = False
        notEmptytrans (_,_,_)     = True
 




edfa2ndfa :: (Eq st , Ord sy) => Edfa st sy -> Ndfa Int sy
edfa2ndfa edfa@(Edfa vs qs ss zs d) = ttNdfa2Ndfa (vs ,q',[s'],z',ntt)
  where s'        = posInList ss qs
        z'        = [ posInList z qs  | z <- zs ]
        q'        = [ 0 .. n-1]

        ttedfa    = nonEmptytransitions edfa
        ttedfa'   = map (\(o,re,d) -> (posInList o qs, re ,posInList d qs)) ttedfa
        (ntt,n)   = fntt' ttedfa' (length qs)


fntt' :: Ord sy
      => [(Int, RegExp sy, Int)] -> Int -> ([(Int, Maybe sy, Int)], Int)

fntt' []    n        = ([],n)
fntt' ((o,er,d):t) n = ((o,Nothing,head s) : (head z,Nothing,d) : 
                        (transitionTableNdfa ndfa) ++ tt' , n'')
  where (ndfa@(Ndfa vs qs s z delta), n')  = regExp2Ndfa' (extREtoRE er) n
        (tt'  , n'')                       = fntt' t n'

         

posInList :: Eq a => a -> [a] -> Int
posInList z vs =  let (Just x) = elemIndex z vs 
                  in x



-- | Rename an 'Edfa'.
--

renameEdfa :: Eq st 
           => Edfa st sy        -- ^ Automaton 
           -> Int               -- ^ Initial state ID
           -> Edfa Int sy       -- ^ Renamed automaton
renameEdfa (Edfa v q s z delta) istid = Edfa v q' s' z' delta' 
  where newSts      =  zipWith (\ a b -> (a,b)) q [istid .. ]
        q'          =  old2new newSts q
        s'          =  head $ old2new newSts [s]
        z'          =  old2new newSts z
        delta' o d  =  delta (lookupSnd newSts o) (lookupSnd newSts d) 


old2new :: Eq st => [(st,Int)] -> [st] -> [Int]
old2new nsts sts = map (lookupFst nsts) sts

lookupFst :: Eq st => [(st,Int)] -> st -> Int
lookupFst nsts ost = snd $ head (filter (\ (a,b) -> a == ost) nsts)

lookupSnd :: [(st,Int)] -> Int -> st
lookupSnd nsts nst = fst $ head (filter (\ (a,b) -> b == nst) nsts)



-- | Add a transition to an automaton
--

addMove2Edfa :: (Eq st , Eq sy)
             => Edfa st sy                 -- ^ Automaton
             -> (st, st, RegExp sy)        -- ^ Transition
             -> Edfa st sy                 -- ^ New Automaton
addMove2Edfa (Edfa v q s z delta) (sto,std,re) = Edfa v q' s z delta'
  where q' = q `union` [sto,std]
        delta' o d | o == sto && d == std = 
                              if delta o d == Empty then re else Or re (delta o d)
                   | otherwise            = delta o d



-----------------------------------------------------------------------------
-- * Acceptance

edfaaccept :: (Ord sy , Eq st) 
           => Edfa st sy         -- ^ Automaton 
           -> [sy]               -- ^ Input symbols
           -> Bool
edfaaccept edfa symb = ndfaaccept (edfa2ndfa edfa) symb


-----------------------------------------------------------------------------
-- * Construction 

-- | Construct a 'Edfa' from a 'Dfa'

dfa2edfa :: Eq st 
         => Dfa st sy                  -- ^ Deterministic Finite Automaton
         -> Edfa st sy                 -- ^ Extended Deterministic Finite Automaton
dfa2edfa (Dfa v q s z delta) = Edfa v q s z delta'
 where delta' o d = regExpFromTo delta v o d


-- | Construct a 'Edfa' from a 'Ndfa'
--   (It assumes that the 'Ndfa' has a single start state...)
--   (To solve this we can use the renameNdfa starting at a giving int and adding
--    a not used int as the new initial state, moving by epsilon to all the others)

ndfa2edfa :: Eq st 
          => Ndfa st sy                -- ^ Non-Deterministic Finite Automaton
          -> Edfa st sy                -- ^ Extended Deterministic Finite Automaton
ndfa2edfa (Ndfa v q s z delta) = Edfa v q s' z delta'
 where delta' o d = ndfaregExpFromTo delta v o d
       s'         = head s   -- ???


re2edfa :: Ord sy => RegExp sy -> Edfa Int sy
re2edfa = dfa2edfa . beautifyDfa . regExp2Dfa . extREtoRE

-----------------------------------------------------------------------------
-- * Printing

-- | Print a 'Edfa' in GraphViz


edfatographviz :: (Show (RegExp sy)) 
        => Edfa st sy            -- ^ Automaton
        -> [Char]                -- ^ Graph's name
        -> [Char]                -- ^ Node's shape
        -> [Char]                -- ^ Orientation
        -> (st -> [Char])        -- ^ Show function to print the state ids
        -> [Char]
edfatographviz edfa@(Edfa v q s z delta) name shape orientation showState = 
  "digraph " ++ name ++ " {\n " 
             ++ "rankdir = " ++ orientation ++ " ;\n " 
             ++ (showElemsListPerLine (showStates q)) ++ "\n " 
             ++ (showInitialState s) ++ "\n " 
             ++ (showElemsListPerLine (showFinalStates' z))
             ++ (showElemsListPerLine (showEdfaArrows edfa))
             ++ "node [shape=none, lavel=initialState, style = invis];\n"
             ++ "\n}"
  where 
    showElemsListPerLine :: [String] -> String
    showElemsListPerLine []    = ""
    showElemsListPerLine (h:t) = ((showString h) "\n ") ++ 
                                 (showElemsListPerLine t) 

    showStates qs = [(showState q) ++ 
                     " [shape=" ++ shape ++ " , label=" ++ (showState q) ++ " ,color=black];" 
                    | q <- qs 
--                    , not (ndfaIsStDead q v z delta ) 
                    ]
 
    showInitialState  s = (showState s) 
                          ++ " [shape=" ++ shape ++ " , label= " ++ (showState s)
                          ++ ", color=green];\n " 

    showFinalStates' zs = [ (showState z) 
                          ++ " [shape=double" ++ shape ++" , color=red];" | z <- zs ]

    showEdfaArrows edfa
       = map (\ (o,l,d) -> genOneArrow (showState o) (show l) (showState d))
             (nonEmptytransitions edfa)



-- Creating the incoming arrows for the initial states
-- (for each state we create an invisible node and a arrow connecting to the initial one)


    createInitialArrows [] []         = " "
    createInitialArrows (x:xs) (y:ys) = (show x) ++ " -> " ++ (showState y) ++ 
                                        "[color = green];\n" ++ 
                                        createInitialArrows xs ys



-- | Print a 'Edfa' in GraphViz
edfa2graphviz :: (Show st, Show (RegExp sy)) 
              => Edfa st sy                         -- ^ Automaton
              -> [Char]                             -- ^ Name of Graph\module
              -> [Char]                             -- ^ Dot representation
edfa2graphviz edfa name = edfatographviz edfa name "circle" "LR" show 


-- | Print a 'Edfa' in GraphViz in a file
edfa2graphviz2file edfa name = writeFile (name++".dot") (edfa2graphviz edfa  name) 




-- | Compute the number of incoming arrows for a given state
--

edfanumberIncomingArrows :: (Eq st , Eq sy)
                         => (st -> st -> RegExp sy)    -- ^ Transition Function
                         -> [st]                       -- ^ Set of States
                         -> st                         -- ^ Destination State
                         -> Int                        -- ^ Number of Arrows
edfanumberIncomingArrows delta qs dest = 
        length  [ q
                | q <- qs 
                , delta q dest /= Empty
                ]


-- | Compute the number of outgoing arrows for a given state

edfanumberOutgoingArrows :: (Ord st , Eq sy)
                         => (st -> st -> RegExp sy )   -- ^ Transition Function
                         -> [st]                       -- ^ States
                         -> st                         -- ^ Origin State
                         -> Int                        -- ^ Number of Arrows

edfanumberOutgoingArrows delta qs o = length (edfadestinationsFrom delta qs o)




edfadestinationsFrom delta qs o = [ q
                                  | q <- qs 
                                  , delta o q /= Empty ]



-----------------------------------------------------------------------------
-- * Elimination

-- | Minimize an 'Edfa', that is, minimizes the number of states of a 'Edfa'

minimizeEdfa :: (Eq st , Eq sy) 
             => Edfa st sy                  -- ^ Automaton
             -> Edfa Int sy                  -- ^ Minimized Automaton
minimizeEdfa edfa = elimLstStatesEdfa edfa' elimStates
  where (edfa',elimStates) = initialEdfaStates edfa



initialEdfaStates :: (Eq st , Eq sy) 
                  => Edfa st sy              -- ^ Automaton
                  -> (Edfa Int sy , [Int])   -- ^ (Inital 'Edfa',States to be removed)
initialEdfaStates edfa = (Edfa v' q' s' z' d' , elimStates)
  where Edfa v' q' s' z' d' = addNewFinalStEdfa $ addNewStartStEdfa edfa
        elimStates          = q' \\ (s':z') 


-- | Compute all the minimized 'Edfa' obtained from a given 'Edfa' 
--   It uses the brute force algorithm that considers all possible 
--   permutations of state elimination.

minimizeEdfa' :: (Eq st , Eq sy) 
              => Edfa st sy                 -- ^ Automaton
              -> [Edfa Int sy]              -- ^ List of Minimized Automaton
                                            --   (it uses all possible permutations
                                            --    of state elimination)
minimizeEdfa' edfa = map (elimLstStatesEdfa edfa') (permutations elimStates)
  where (edfa',elimStates) = initialEdfaStates edfa



-- | Pairs a given minimized 'Edfa' with its size


minEdfa2EdfaWithSize edfa = (sizeOfMinimizedEdfa edfa , edfa)


-- | Pairs a given list of minimized 'Edfa' with their size'


lstminEdfa2lstEdfaWithSize = map minEdfa2EdfaWithSize 



-- | Computes the size of a minimized 'Edfa'
--   That is, returns the size of the regular expression labelling its single
--   transition from start to the final state.

sizeOfMinimizedEdfa (Edfa v q s z d) | length z > 1 = Nothing
                                     | otherwise    = Just size
  where size = sizeRegExp (d s (head z))


-- | Given a list of minimized 'Edfa' it returns the one defining the smallest 'RegExp'

minEdfaFromLst lstedfa = minEdfa
  where edfaLstWithSize = map (\x -> (sizeOfMinimizedEdfa x,x)) lstedfa
        sizesRegExp     = map (\(x,y) -> case x of { (Just v) -> v ; Nothing -> 0 }) 
                              edfaLstWithSize
        minimumSizeRE   = minimum sizesRegExp
        (Just pos)      = elemIndex minimumSizeRE sizesRegExp
        minEdfa         = lstedfa !! pos
        

-- | Compute the set of all possible 'Edfa' from a 'RegExp'

regExp2Edfas :: Eq sy => RegExp sy -> [Edfa Int sy]
regExp2Edfas re  = minimizeEdfa' $ dfa2edfa $ minimizeDfa $ regExp2Dfa re


-- | Compute the smallest regular expressions for a given one

minimizeRegExp :: Eq sy 
               => RegExp sy        -- Original Regular Expression
               -> RegExp sy        -- Simplified Regular Expression
minimizeRegExp re =  d s (head z)
  where edfaLst           = regExp2Edfas re 
        (Edfa v q s z d)  = minEdfaFromLst edfaLst 




-- | Eliminate a State of an 'Edfa' 
--   Given an 'Edfa' and one of its states q, computes an equivalent 'Edfa' constructed 
--   according to the following rule
--
--     delta' o d = (delta o d) + (delta o q) (delta q q)* (delta q d)
--

elimStateEdfa :: (Eq st , Eq sy)
              => Edfa st sy           -- ^ Automaton
              -> st                   -- ^ State to be removed
              -> Edfa st sy           -- ^ Construted Automaton

elimStateEdfa (Edfa v q s z delta) st = Edfa v' q' s' z' delta'
  where  v' = v
         q' = delete st q   
         s' = s
         z' = z
         delta' o d = simplifyRegExp 
                      ((delta o d) `Or` 
                       ((delta o st) `Then` (Star (delta st st)) `Then` (delta st d)))

-- | Eliminate a List of States of an 'Edfa'
--   Given an initial 'Edfa' and a list of states to be removed, it returns
--   an equivalent 'Edfa' with those states eliminated

elimLstStatesEdfa :: (Eq st , Eq sy)
                  => Edfa st sy           -- ^ Automaton
                  -> [st]                 -- ^ List of states to be removed
                  -> Edfa st sy           -- ^ Construted Automaton

elimLstStatesEdfa edfa []       = edfa
elimLstStatesEdfa edfa (st:sts) = elimLstStatesEdfa edfa' sts
  where edfa' = elimStateEdfa edfa st 


elimLstStatesEdfa' :: (Eq st , Eq sy)
                   => Edfa st sy           -- ^ Automaton
                   -> [st]                 -- ^ List of states to be removed
                   -> [Edfa st sy]         -- ^ Sequence of Construted Automaton
elimLstStatesEdfa' edfa []       = [edfa]
elimLstStatesEdfa' edfa (st:sts) = edfa : (elimLstStatesEdfa' edfa' sts)
  where edfa' = elimStateEdfa edfa st 



-- | Add a new start state to an 'Edfa'
--   We need to add a new start state to the 'Edfa' when 
--          - the start state of the 'Edfa' is also a final state
--          - the number of incoming arrows of the start state is greater than 0


addNewStartStEdfa :: (Eq st, Eq sy) 
                  => Edfa st sy             -- ^ Automaton
                  -> Edfa Int sy            -- ^ Construted Automaton

addNewStartStEdfa edfa@(Edfa v q s z d) 
  | s `elem` z || edfanumberIncomingArrows d q s > 0 = edfa'''
  | otherwise  = renameEdfa edfa 1                    -- To not force type of Edfa to Int
  where Edfa v' q' s' z' d' = renameEdfa edfa 2
        edfa'               = Edfa v' (1:q') 1 z' d'             -- Add new start state 1
        edfa''              = addEmptyMoves edfa' 1
        edfa'''             = addMove2Edfa edfa'' (1,s',Epsilon) -- Add Epsilon Transition
                                                                 -- from new start 1 to former start s'  



-- | Add a new final state to an 'Edfa'
--   We need to add a new final state to the 'Edfa' when 
--          - the 'Edfa' has more than one final state
--          - the single final state has a number of outcoming arrows greater than 0


addNewFinalStEdfa :: (Ord st, Eq sy) 
                  => Edfa st sy           -- ^ Automaton
                  -> Edfa Int sy          -- ^ Construted Automaton

addNewFinalStEdfa edfa@(Edfa v q s z d) 
  | length z > 1 || edfanumberOutgoingArrows d q (head z) > 0 = edfa'''
  | otherwise  = renameEdfa edfa 1                     -- To not force type of Edfa to Int
  where Edfa v' q' s' z' d' = renameEdfa edfa 1
        newFinalState       = 1+length q'
        edfa'               = Edfa v' (newFinalState:q') s' [newFinalState] d'  
        edfa''              = addEmptyMoves edfa' newFinalState
        edfa'''             = addEpsilonMoves edfa'' z' newFinalState        


addEpsilonMoves edfa [] _                = edfa
addEpsilonMoves edfa (st:sts) newFinalSt = addEpsilonMoves edfa' sts newFinalSt
  where edfa' = addMove2Edfa edfa (st , newFinalSt , Epsilon)

-- | Helper function to make a 'Edfa' complete
--

addEmptyMoves (Edfa v q s z d) st = Edfa v q s z d' 
  where d' sto std | sto == st   = Empty             
                   | std == st   = Empty
                   | otherwise   = d sto std



