-----------------------------------------------------------------------------
-- |
-- Copyright   :  (c) Joost Visser 2004
-- License     :  LGPL
-- 
-- Maintainer  :  joost.visser@di.uminho.pt
-- Stability   :  experimental
-- Portability :  portable
--
-- This module implements import chasing for Haskell.
--
-----------------------------------------------------------------------------

module Language.Haskell.ChaseImports where

-------------------------------------------------------------------------------    

import Language.Haskell.ModuleCollection
import Language.Haskell.Syntax
import Language.Haskell.Parser
import Language.Haskell.Pretty
import Language.Haskell.SyntaxTermInstances
import Language.Haskell.SyntaxAccessors
import qualified Language.Haskell.Literate as Literate

import StrategyLib

import IO
import System
import System.Directory
import Monad
import Control.Monad.Identity

-------------------------------------------------------------------------------    
-- * Import chasing for Haskell

-- | Chase modules, given a searchpath and module name, apply a transformation
--   function, and write the resulting modules to a given output directory.
hsIOwrapChasingArgs 
  :: (FilePath, ModuleName, FilePath)  -- ^ Path, module name, output directory
  -> (ModuleCollection -> IO ModuleCollection) -- ^ Transformation function
  -> IO ()
hsIOwrapChasingArgs (pth,m,d) transform
  = do errLn "Import chasing..."
       errLn ("Start from:  "++m)
       errLn ("Search path: "++pth)
       errLn ("Writing results in:"++d)
       let dirs = breakPath pth
       ms <- hsChaseFrom dirs m
       ms' <- transform ms
       mapM_ (writeModule d) ms'
       
-- | Start chasing from a single module (wrapper).
hsChaseFrom :: [FilePath]
            -> ModuleName 
            -> IO ModuleCollection
hsChaseFrom dirs m
  = hsChaseWith dirs [m] [] [] onModule onMissing
 where
  onModule m a = return $ a++[m]
  onMissing _ = return
  
-- | Haskell import chase algorithm (worker).
hsChaseWith :: [FilePath]    -- ^ Directories to search 
            -> [ModuleName]  -- ^ Modules still to find (todo)
            -> [ModuleName]  -- ^ Modules already found (done)
            -> a             -- ^ Accumulator
            -> (ModuleEntry -> a -> IO a)  -- ^ Action taken on modules found. 
	        -> (ModuleName  -> a -> IO a)  -- ^ Action taken on missing modules.
	        -> IO a
hsChaseWith dirs todo done accu onModule onMissing
 = chase todo done accu
   where
    chase [] done accu
      = do errLn "Import chasing complete."
           return accu
    chase (m:ms) done accu
      | m `elem` done = chase ms done accu
      | otherwise     = processFile `mplus` skipFile
     where
       processFile 
         = do sin <- readHsFile dirs m
              case (parseWrap' (fromLiterate sin)) of
                (Left pin)  
                  -> do let is = getImports pin
                        accu' <- onModule (m,is,pin) accu
                        chase (ms++is) (m:done) accu'
                (Right msg) 
                  -> do errLn ("Failed to parse "++m++": "++msg)
                        accu' <- onMissing m accu
                        chase ms (m:done) accu'     
       skipFile    
         = do errLn ("Did not find module "++m++": skipping.")
              accu' <- onMissing m accu
              chase ms (m:done) accu'


-- | Start chasing from all haskell modules that can be found recursively 
--   inside the given top directories.
hsChaseDownDirs :: [FilePath]           -- ^ Search path for import chasing
                -> [FilePath]           -- ^ Directories to find top modules
                -> IO ModuleCollection
hsChaseDownDirs searchPath dirs = do
  files <- findHsFiles dirs
  maybeModules <- mapM parseFile files
  let moduleEntries = foldr (\x xs -> maybe xs (:xs) x) [] maybeModules
  let todo = concatMap (\(_,is,_) -> is) moduleEntries
  let done = map (\(m,_,_) -> m) moduleEntries
  hsChaseWith (searchPath++dirs) todo done moduleEntries onModule onMissing
 where
  onModule m a = return $ a++[m]
  onMissing _ = return
  parseFile fileName = do
    content <- readFile fileName
    case (parseWrap' (fromLiterate content)) of
      (Left m) -> 
         return $ Just (getModuleName m, getImports m,m)
      (Right msg) -> do 
         errLn ("Failed to parse "++fileName++": "++msg)
         return Nothing

-------------------------------------------------------------------------------    
-- * Handle literate scripts

fromLiterate :: String -> String
fromLiterate str
  = fromLit (isLit str) str
  
isLit :: String -> Bool
isLit 
  = any isLitLine . lines
    where 
      isLitLine ('>':_) = True
      isLitLine _       = False

-- Rest is taken from DrIFT

-- NB we don't do the latex-style literate scripts currently.
fromLit :: Bool -> String -> String
fromLit True txt 
  = case Literate.process txt of
      ([],s) -> s
      (e,_) -> error e
fromLit False txt = txt

toLit :: Bool -> String -> String
toLit True = unlines . map (\l -> '>':l)  . lines
toLit False = id       
 
isLiterate :: String -> Bool
isLiterate = any ((=='>'). head) . words    
  
------------------------------------------------------------------------------- 
-- * Haskell I-O Auxilliaries

-- | Parse a String as a Haskell module. Returns either the parsed module
--   or an error message.
parseWrap' :: String -> Either HsModule String
parseWrap' x
  = case (parseModule x) of
      (ParseOk x)              -> Left x
      (ParseFailed srcLoc msg) -> Right (msg++" at "++(show srcLoc))

-- | Find haskell files inside a list of directories
findHsFiles :: [FilePath] -> IO [FilePath]
findHsFiles dirs = do
  let findCmd = "find "++unwords dirs++" -iname \"*.hs\" -or -iname \"*.lhs\" > find.results"
  system findCmd
  findResults <- readFile "find.results"
  return $ lines findResults

-- | Read a Haskell file with given name from given path.
readHsFile :: [FilePath] -> ModuleName -> IO String
readHsFile dirs modulename
  = do errLn ("Starting search for: "++modulename)
       readFileSearching dirs (m2fname modulename) ["hs","lhs"]
       
-- | Write a given module to file, in the given directory.
writeModule :: FilePath       -- ^ Output directory.
            -> ModuleEntry    -- ^ Module to be written.
            -> IO ()
writeModule dir (n,is,m)
  = do let content = (prettyPrint m)++"\n"
       let relName = (m2fname n)++".hs"
       let absName = dir++"/"++relName 
       errLn ("Writing module "++n++" to file "++absName)
       createDirectoryForced dir (noBasename relName)
       writeFile absName content

-- | Transform a Haskel module name (containing dots) into a file name (containing slashes).
m2fname :: ModuleName -> FilePath
m2fname = map dot2slash where dot2slash '.' = '/'
                              dot2slash c   = c

-- | Transform a file name (containing slashes) into a Haskel module name (containing dots).
fname2m :: FilePath -> ModuleName
fname2m = map slash2dot . noExtension 
  where 
    slash2dot '/' = '.'
    slash2dot c   = c
  
------------------------------------------------------------------------------- 
-- * General I-O Auxilliaries

-- | Report an error on stderr.
errLn :: String -> IO ()
errLn str = hPutStrLn stderr str

-- | Obtain the directory part of a file name.
noBasename :: FilePath -> FilePath
noBasename fname
  =  reverse (dropWhile (/='/') (reverse  fname))
  
-- | Obtain the file name without extension.
noExtension :: FilePath -> FilePath
noExtension fname
  = reverse (pop $ dropWhile (/='.') (reverse  fname))
    where pop [] = []
          pop (_:xs) = xs

-- | Search for a file in given directories with alternative extensions
readFileSearching :: [FilePath] -> FilePath -> [String] -> IO String
readFileSearching dirs basename exts
  = do results <- mapM tryReadFile fnames
       case dropWhile hasFailed results of
         ((Right (fc,fn)):_) 
	     -> errLn ("Read file: "++fn) >> return fc
	 _   -> errLn ("Could not find file: "++basename) >> mzero                  
    where 
      fnames = [d++'/':basename++'.':e | d <- dirs, e <- exts]
      hasFailed (Left _) = True
      hasFailed _        = False
      tryReadFile fn = try ( readFile fn >>= \fc -> return (fc,fn) )

-- | Create a directory, and its parents if necessary.
createDirectoryForced :: FilePath    -- ^ Directory in which to create the new directories 
                      -> FilePath    -- ^ Directory to create (may contain slashes).
                      -> IO ()
createDirectoryForced superdir dir
  = let segments = filter (/="") (breakFile dir)
        dirs = map (sepWith "/") [ take n segments | n <- [1..length segments] ]
        create d 
          = do let dir = superdir++"/"++d
               exists <- doesDirectoryExist dir 
               if exists then return ()
                         else createDirectory dir
        sepWith x [] = [] 
        sepWith x xs = foldl1 (\a b -> (a++x++b)) xs 
    in sequence_ (map create dirs)

-- | break a file name into its segments, i.e. the pieces between slashes.    
breakFile :: FilePath -> [String]
breakFile = breakWith (=='/')
       
-- | Break path argument into directory names. i.e. the pieces between colons.
breakPath :: String -> [FilePath]
breakPath = breakWith (==':')

-- | Break a list into segments, using a predicate on elements to indicate break points.
--   The break points themselves will be discarded.
breakWith :: (a -> Bool) -> [a] -> [[a]] 
breakWith f x = case break f x of
	(p,(_:pp)) -> p: breakWith f pp
	(p,[]) -> [p]

-------------------------------------------------------------------------------    
	
-- Test

testOnUMinhoHaskellLibrariesWithContrib :: IO ModuleCollection
testOnUMinhoHaskellLibrariesWithContrib
  = hsChaseDownDirs 
     (breakPath ":./../contrib/HUnit-1.0:./../contrib/StrategyLib-4.0-beta/library:./../contrib/StrategyLib-4.0-beta/models/drift-default:./../contrib/Sdf2Haskell-2.2/generator:./../contrib/haterm-1.4/library:./../contrib/HaXml-1.12/src") 
     ["."] 

-------------------------------------------------------------------------------    
