-----------------------------------------------------------------------------
-- |
-- Copyright   :  (c) Joost Visser 2004
-- License     :  LGPL
-- 
-- Maintainer  :  joost.visser@di.uminho.pt
-- Stability   :  experimental
-- Portability :  portable
--
-- This module implements some metrics on Haskell sources.
--
-----------------------------------------------------------------------------

module Language.Haskell.Metrics where

import Language.Haskell.Syntax
import Language.Haskell.SyntaxAccessors
import Language.Haskell.ModuleGraph
import Language.Haskell.ChaseImports
import Language.Haskell.FreeNames (hsFreeAndDeclared)

import Data.GraphR 
import Data.Bag
import Data.Set(size)
import Data.Metrics

-----------------------------------------------------------------------------

-- | Compute system-level metrics for a given list of modules.
systemMetrics :: [HsModule] -> [Metric]
systemMetrics ms 
  = map mkMetric [nOPm,nOm,cIs,wCs,tMs,bMs,timp,lev]
    where
      mkMetric (d,n,v) = Metric d n v
      nOPm = ("Number of Haskell 98 modules parsed with success","PM",
              show $ length ms)
      modGraph = buildModuleGraph ms
      nOm = ("Number of modules in module graph","M",
             show $ size $ ent modGraph)
      cIs = ("Circular import groups","NSSCC",
             show $ size $ strongNonSingletonComponentSet modGraph)
      wCs = ("Connected import groups (weak)","WCC",
             show $ size $ weakComponentSet modGraph)
      tMs = ("Number of top modules","TOP",
             show $ size $ topNodes modGraph)
      bMs = ("Number of bottom modules","BOT",
             show $ size $ bottomNodes modGraph)
      timp = ("Tree impurity of module graph","TIMP",
              (show $ treeImpurity modGraph)++" %")
      sCM = calculateStrongComponentMetrics modGraph
      lev  = ("Number of strong components in module graph","SCC",
              show $ componentCount sCM)

-- | Compute module-level metrics for a given list of modules.
moduleMetrics :: [HsModule] -> [GroupMetric]
moduleMetrics ms = map mkGroupMetric ms
 where
   modGraph = buildModuleGraph ms
   (fi,fo)  = fanInOut modGraph
   mkGroupMetric m = GroupMetric mn [fI,fO,fN,dN]
    where
     mn = getModuleName m
     fI = Metric "Number of client modules" "CLI" (show $ lookupBag fi mn)
     fO = Metric "Number of imported modules" "IMP" (show $ lookupBag fo mn)
     (fn,dn) = head $ hsFreeAndDeclared m
     fN = Metric "Number of free names" "FREE" (show $ length fn)
     dN = Metric "Number of declared names" "DECL" (show $ length dn)
     

-- | Run a metric report generator on all Haskell modules that can be
--   found by import chasing from files in given directories.
runMetrics :: String          -- ^ Search path
           -> FilePath        -- ^ Name of output file
           -> IO ()
runMetrics path resultFileName = do
  let dirs = breakPath path
  mc <- hsChaseDownDirs dirs dirs 
  putStrLn $ "Computing metrics ..."
  let ms = map (\(_,_,m) -> m) mc
  let moduleMs = moduleMetrics ms
  writeFile resultFileName (printMetricsCSV moduleMs)
  putStrLn (unlines $ printMetricsList $ systemMetrics ms)
  putStrLn (unlines $ printMetricsList $ summaryMetrics moduleMs)
  putStrLn $ "Per-module metrics written to "++resultFileName


-- | Takes a list of pairs (filename and a list of group metrics) and creates
--   a string representation of a Comma Separated Values file (CSV).
printMetricsCSV :: [GroupMetric] -> String
printMetricsCSV a = unlines $ map (concatWithSep ", ") (header:metrics)
   where header  = "FILE":(printGroupMetricHeaderRow [ head a ])
         metrics = map (\m@(GroupMetric f ms) -> f:(printMetricsRow ms)) a

-- | Concatenate strings with separator.
concatWithSep :: String -> [String] -> String
concatWithSep a []    = ""
concatWithSep a [s]   = s
concatWithSep a (h:t) = h ++ a ++ concatWithSep a t

-----------------------------------------------------------------------------


{- Wish list:

On module graph:

- How many modules?
- Depth and width of module graph?
- How many imports / exports per module?
- Cycles?

Per module:

- How many top-level declarations
  - How many functions?
  - How many types?
  - How many classes?
  - How many instances?
  
- Depth and width of each function?
- How many recursive / mutally recursive functions / types?
- Functions without type?

- How many instances per class?
- How many "undefined", "error", "fail" ?
- How many "quickCheck" properties?

Per function definition

- How many nested function declarations?
- How many nested applications?


-}
