-----------------------------------------------------------------------------
-- |
-- Copyright   :  (c) Universidade do Minho, 2004
-- License     :  LGPL
-- 
-- Maintainer  :  patrickmachado@gmail.com
-- Stability   :  experimental
-- Portability :  portable
--
--- Extract Package & Class Graph from Java files.
--- This graph shows relations between and among classes and packages,
--- i.e. inheritance, containment (nesting), imports.
--
-----------------------------------------------------------------------------

module Language.Java.CallGraph where

-------------------------------------------------------------------------------

import Language.Java.Syntax
import Language.Java.Util
import Language.Java.ChaseImports
import Language.Java.Declarations
import Language.Java.TypeInference
import List hiding (union)

import StrategyLib
import Control.Monad.Identity
import Data.Monoid
import Data.Graph.Slicing
import Data.Relation.MapFromPairs
import Data.Relation.SetOfPairs 
import qualified Data.Map as Map
import qualified Data.Set as Set
import Data.Relation.GraphViz

-------------------------------------------------------------------------------
-- * Representation

-- | The type of conditional call graphs.
type CallGraph
  = LRel CGNode CGNode CGEdgeType
  
-- | The type of call graph nodes.
data CGNode = CGClass ClassName
			| CGMethod  ClassName MethodName [ParameterType]
				deriving (Eq, Ord, Show, Read)

-- | The type of edges types in call graphs.
type CGEdgeType = String

-- | Empty call graph.  
emptyCG :: CallGraph
emptyCG = Map.empty

-- | Append two call graphs to form a new one.
appendCG :: CallGraph -> CallGraph -> CallGraph
appendCG = Map.union

-- | Call graphs are monoids.
instance Monoid CallGraph where
  mempty  = emptyCG
  mappend = appendCG

{-
instance Show CGNode where
	showsPrec _ (CGClass a) = showString a
	showsPrec _ (CGMethod c m ps) = showString c . showChar '.' 
			. showString m . showChar '(' . showPars ps . showChar ')'
		where
			showPars [] = id
			showPars [x] = showString x
			showPars (h:t) = showString h . showChar ',' . showPars t
-}
-- | Add an edge of given type to a call graph.
addEdgeToCG :: CGNode -> CGNode -> CGEdgeType
            -> CallGraph -> CallGraph
addEdgeToCG (CGClass "") _ _ gr   = gr
addEdgeToCG x y tp gr   =  Map.insert (x,y) tp gr

-- | Add a call edge to a call graph.
addCallEdge         :: CGNode -> CGNode -> CallGraph -> CallGraph
addCallEdge s t cg  = addEdgeToCG s t "call" cg

-- | Add a nesting edge to a call graph.
addNestEdge         :: CGNode -> CGNode -> CallGraph -> CallGraph
addNestEdge s t cg  = addEdgeToCG s t "nest" cg



--  | Test funcion. Takes the file tests/Foo.java. to be run from the libraries folder (make ghci) 
doit :: IO ()
doit = do
    --cus <- javaChaseImports ["Language/Java"] ["test.Foo"]
--    cus <- parseJavaFile "/tmp/labeler/AsyncCall.java"
--    cus <- parseJavaFile "/tmp/labeler/Img.java"
--    cus <- parseJavaFile "/tmp/labeler/Interface.java"
--    cus <- parseJavaFile "/tmp/labeler/Labeler.java"
--    cus <- parseJavaFile "/tmp/labeler/LabelerClient.java"
--    cus <- parseJavaFile "/tmp/labeler/LabelerImpl.java"
--    cus <- parseJavaFile "/tmp/labeler/NoInterface.java"
--    cus <- parseJavaFile "/tmp/labeler/Seq.java"
--    cus <- parseJavaFile "/tmp/labeler/Server.java"
--    cus <- parseJavaFile "/tmp/labeler/Vis.java"
--    cus <- parseJavaFile "/tmp/labeler/Zip.java"
--    cus <- parseJavaFile "/tmp/labeler/jpeg2img.java"
      cus <- parseJavaFile "/tmp/Bat.java"
      decs <- collectMembers "" cus
      ccg <- java2ccg (CGClass "") decs cus
      print ccg >> writeFile "yo.dot" (cg2dot "ola" ccg)


-- * API

-- | Main conversion function 
java2ccg :: (Term a, MonadPlus m) => CGNode -> Declarations -> a -> m CallGraph
java2ccg node@(CGClass cname) decs = applyTU $ stop_tdTU worker1
	where
		worker1 = failTU 
			`adhocTU` class2ccg 
			`adhocTU` constructor2ccg 
			`adhocTU` method2ccg 
			`adhocTU` decs2ccg 

		class2ccg (Class1 _ ident _ _ body) 
			= do 
				cg <- java2ccg (CGClass ident) decs body
				return $ addNestEdge node (CGClass ident) cg

		constructor2ccg 
			(Modifier_s_ConstructorDeclarator_Throws_opt_ConstructorBody 
			_ (Comma5 (Identifier ident) pars) _ body)
			= do
				let tpars = getParameterTypes pars
				lvars <- getLocalDecls body
				let newdec = decs `appendDec` lvars
				let mnode = CGMethod cname ident tpars
				cg <- java2ccg mnode newdec body
				return $ addNestEdge node mnode cg
				
		method2ccg (MethodHeader_MethodBody header body) 
			= do 
				ident <- getMethodName header
				tpars <- getMethodParameters header >>= return . getParameterTypes
				lvars <- getLocalDecls body
				lvarspars <- getLocalDecls header
				let newdec = decs `appendDec` lvars `appendDec` lvarspars
				let mnode = CGMethod cname ident tpars
				cg <- java2ccg mnode newdec body
				return $ addNestEdge node mnode cg
		decs2ccg (FieldDeclaration1 fd)
			= do
				let constr = CGMethod cname cname []
				cg <- java2ccg constr decs fd
				return $ addNestEdge node constr cg
		decs2ccg (MethodDeclaration x) = java2ccg node decs x

java2ccg node@(CGMethod cname _ _) decs = applyTU $ stop_tdTU worker2
	where 
		worker2 = failTU `adhocTU` call2ccg
		call2ccg (NameMethodInvocation ident args)
			= do 
				cg <- java2ccg node decs args 
				let (a,b) = sepname ident
				let typeof = maybe "unknown" id $ findType a cname decs 
				let targs = args2str $ findTypeOfArgs decs cname args
				return $ addCallEdge node (CGMethod typeof b targs) cg
		call2ccg (SuperMethodInvocation _ args)
			= do 
				cg <- java2ccg node decs args 
				return $ addCallEdge node node cg -- FIXME
		call2ccg (PrimaryMethodInvocation prim ident args)
			= do 
				cg <- java2ccg node decs (prim, args) 
				let xit = findTypeOf decs cname prim
				let tp = maybe "unknown" (\x -> if (x=="") then "unknown" else x) xit 
				let targs = args2str $ findTypeOfArgs decs cname args
				return $ addCallEdge node (CGMethod tp ident targs) cg


args2str :: [Maybe TypeName] -> [TypeName]
args2str = map (maybe "unknown" (\x -> if (x=="") then "unknown" else x))

-- | Given a relation, compute the sub graph of a package graph that holds
--   only edges that occur in the relation.
projectCG :: Rel CGNode CGNode -> CallGraph -> CallGraph
projectCG r cg
  = (lrel "call" $ sub $ rel (=="call") cg)
    `Map.union`
	(lrel "nest" $ sub $ rel (=="nest") cg)
    where
      sub r' = projectWith (\x y -> (x,y) `Set.member` r) r'

--superClass :: PackageGraph -> TypeName -> Maybe TypeName
--superClass (_,_,inh,_) tn = lookup tn $ pairs inh

-------------------------------------------------------------------------------
-- * Helpers for dot generation

-- | Print a call graph to dot file content.
cg2dot :: String -> CallGraph -> String
cg2dot n cg
  = lrel2dot n printEdge cg
  
-- | Print a call graph to dot file content.
lrel2dot :: String                                             -- ^ Graph name
         -> (((CGNode, CGNode), CGEdgeType) -> DotStatement)   -- ^ Function that prints edges
         -> CallGraph                                          -- ^ Call graph 
         -> String                                             -- ^ Dot file content
lrel2dot n prtEdge callGraph
  = "digraph "++n++" {\n"++(unlines stmts)++"}\n"
    where
      stmts     = nodeStmts++edgeStmts
      nodeStmts = map printNode (Set.toList (entities callGraph))
      edgeStmts = map prtEdge (Map.toList callGraph)

-- | Print edge as dot statement      
printEdge :: ((CGNode, CGNode), CGEdgeType) -> DotStatement
printEdge (nodes,"call") = mkEdge "color=red" $ edge2str nodes
printEdge (nodes,"nest") = mkEdge "color=tan" $ edge2str nodes
printEdge (nodes,"const") = mkEdge "color=green" $ edge2str nodes
printEdge (nodes,_)      = mkEdge "color=grey" $ edge2str nodes

-- | Print node as a dot statement.
printNode :: CGNode -> DotStatement
printNode node 	| isInvalid node = mkNode "shape=box, height=.1, fontcolor=red" (cgnode2str node)
				| otherwise = mkNode "shape=box, height=.1" (cgnode2str node)

-- | Create tuple of string representations for edge.
edge2str :: (CGNode, CGNode) -> (ClassName, ClassName)
edge2str (a,b) = (cgnode2str a, cgnode2str b)

-- | Create string representation of node.
cgnode2str :: CGNode -> ClassName
cgnode2str (CGClass c) = c
cgnode2str (CGMethod c n pars) = c ++ "." ++ n ++ "(" ++ (concat (intersperse "," pars)) ++ ")"

-- | Test whether the node has any unresolved parts.
isInvalid :: CGNode -> Bool
isInvalid (CGClass _) = False
isInvalid (CGMethod c _ pars) = c == "unknown" || "unknown" `elem` pars

-------------------------------------------------------------------------------
