-----------------------------------------------------------------------------
-- |
-- Copyright   :  (c) Universidade do Minho, 2004
-- License     :  LGPL
-- 
-- Maintainer  :  joost.visser@di.uminho.pt
-- Stability   :  experimental
-- Portability :  portable
--
-- Recursively find and parse Java files.
--
-----------------------------------------------------------------------------

module Language.Java.ChaseImports where

import Language.Java.Syntax
import Language.Java.SyntaxATermConvertibleInstances()
import Language.Java.SyntaxTermInstances()
import Language.Java.Util

import StrategyLib
import SGLR

import Control.Monad.Identity



-----------------------------------------------------------------------------

-- | Recursively find and parse Java files. 
--   Java import chasing is implemented by instantiation of Strafunski's
--   generic import chasing function.
javaChaseImports :: [FilePath]              -- ^ Class path
                 -> [ChaseName]             -- ^ Class names
                 -> IO [CompilationUnit]    -- ^ Parsed files.
javaChaseImports clspth clsnames
  = chaseWith clspth clsnames [] [] parse get_imports onModule onMissing
    where
      parse clspth clsname  = parseJavaFileOnPath clspth clsname
      get_imports cu		    = javaGetImportsQualified cu
      onModule mn is m cus	= return (m:cus)
      onMissing	mn cus		= return cus

javaGetImportsQualified :: CompilationUnit -> [ChaseName]
javaGetImportsQualified cu@(CU pkg_decl _ _)
  = map addQualifier (javaGetImports cu)
    where
      addQualifier name = maybeResolve importMap pkgName name
      importMap = mkImportMapCU cu
      pkgName   = packDecl2ident pkg_decl

-- | Extract all imported class names from a compilation unit.
javaGetImports :: CompilationUnit -> [ChaseName]
javaGetImports (CU pkg_decl imports types)
  = map qname2str used_cls_names
    where
      used_cls_names 
        = runIdentity (applyTU (full_tdTU worker) (imports,types))

      worker   = constTU [] `adhocTU` getClsOrIfcType
                            `adhocTU` getArrType
                            `adhocTU` getImportedType

      getClsOrIfcType (Name name)             = return [name]
      getArrType (ArrayType0 name)            = return [name]
      getImportedType (Import_semicolon name) = return [name]

-- | Find and parse a java file. The result is either a
--   parsed compilation unit, or an error message.
parseJavaFileOnPath :: [FilePath]       -- ^ Class path 
                    -> ChaseName        -- ^ Class name
                    -> IO (Either CompilationUnit String)
parseJavaFileOnPath clspth clsname
  = do tableName <- findFile tblpth "Java" ["tbl"]  
       fileName <- findFile clspth relativeName ["java"]
       cu <- sglr tableName fileName "CompilationUnit"
       return (Left cu)
    where
      relativeName = map dot2slash clsname
      dot2slash '.' = '/'
      dot2slash c   = c
      tblpth = ["."]

-- | Find and parse a java file. The result is either a
--   parsed compilation unit, or an error message.
parseJavaFile :: FilePath        -- ^ full file name
              -> IO (Either CompilationUnit String)
parseJavaFile fileName
  = do tableName <- findFile tblpth "Java" ["tbl"]  
       cu <- sglr tableName fileName "CompilationUnit"
       return (Left cu)
    where
      tblpth = ["."]

----------------------------------------------------------------------------



