-----------------------------------------------------------------------------
-- |
-- Copyright   :  (c) Universidade do Minho, 2004
-- License     :  LGPL
-- 
-- Maintainer  :  patrickmachado@gmail.com
-- Stability   :  experimental
-- Portability :  portable
--
-----------------------------------------------------------------------------

module Language.Java.Declarations where

-----------------------------------------------------------------------------

import Language.Java.Syntax
import Language.Java.Util

import StrategyLib
import MonadState()
import Control.Monad.Identity
import Data.Monoid
import qualified Data.Map as Map
import Data.Map(Map)
import Data.Relation.SetOfPairs

-----------------------------------------------------------------------------

-- | A repository of variable declarations at a certain point
--  Maybe TypeName -> The class where the variable is declared (Nothing if al local scope)
--  VarName -> The field/variable being declared
--  TypeName -> The type of the variable/field

-- A Collection of  
type GlobalDeclarations = Map (ClassName, MemberName) TypeName

type LocalDeclarations = Map VarName TypeName

type Declarations = (GlobalDeclarations, LocalDeclarations)

-- | The type of a class in a declaration
-- type TypeName = String
type TypeName = String
type ClassName = String
-- | The name of a variable
type VarName = String

-- | The name of a method
type MethodName = String

-- | The type of a parameter in the method header
type ParameterType = TypeName

-- | The Member Name in a declaration
data MemberName = Field VarName
			| Method MethodName [ParameterType]
				deriving (Show, Eq, Ord, Read)


-- | Declarations are monoids
instance Monoid Declarations where
	mempty  = emptyDec
	mappend = appendDec

-- * Declarations API

-- | Empty declaration
emptyDec :: Declarations
emptyDec = (Map.empty, Map.empty)

-- | Append two declarations
appendDec :: Declarations -> Declarations -> Declarations
appendDec (g1,l1) (g2,l2) = (Map.union g1 g2, Map.union l1 l2)

-- | Add a global declararion
addGlobalDec :: Declarations -> (ClassName, MemberName) -> TypeName -> Declarations
addGlobalDec (g,l) key rt = (Map.insert key rt g,l)

-- | Add a local declararion
addLocalDec :: Declarations -> VarName -> TypeName -> Declarations
addLocalDec (g,l) key rt = (g, Map.insert key rt l)

-- | Get the type of a local variable
getLocalType :: (MonadPlus m) => Declarations -> VarName -> m TypeName
getLocalType (_,l) a = maybe mzero return (Map.lookup a l)

-- | Get the type of a global (class scope) member (Field or Method)
getGlobalType :: (MonadPlus m) => Declarations -> ClassName -> MemberName -> m TypeName
getGlobalType (g,_) tn vn = maybe mzero return $ Map.lookup (tn, vn) g

-- | General funcion for getting a variable's or member's type
getType :: (MonadPlus m) => Declarations -> ClassName -> MemberName -> m TypeName
getType decs tn fi@(Field f) = getLocalType decs f `mplus` getGlobalType decs tn fi
getType decs tn m@(Method _ _) = getGlobalType decs tn m


-- * Declarations Manipulation
-- | Generic function to collect global variable declarations
collectMembers :: (Term a, MonadPlus m) => ClassName -> a -> m Declarations
collectMembers p = applyTU (stop_tdTU worker)
	where 
		worker = failTU 
			`adhocTU` classes 
			`adhocTU` fields 
			`adhocTU` methods 

		classes(Class1 _ ident _ _ body) 
			= collectMembers ident body
		--classes _ = return emptyDec

		fields (Comma_semicolon _ tp vs) 
			= do
				l <- getVars vs
				let tpof = type2str tp
				let f x xs = addGlobalDec xs (p,x) tpof
				return $ foldr f emptyDec l
 		--fields _ = return emptyDec

		-- Here we only account for methods returning something
		-- void methods are irrelevant
		methods (Modifier_s_Type_MethodDeclarator_Throws_opt _ 
			ret (Comma2 name ps) _) 
			= do 
				let pts = getParameterTypes ps
				return $ addGlobalDec emptyDec (p,Method name pts) (type2str ret) 
		methods _ = return emptyDec


-- | Collects local variable declarations in a method
getLocalDecls :: (Term a, MonadPlus m) => a -> m Declarations
getLocalDecls = applyTU $ stop_tdTU worker
	where 
		worker = failTU `adhocTU` local `adhocTU` localpars
		-- local (Comma (ReferenceType (ClassOrInterfaceType (Name n))) vs)
		local (Comma tp vs)
			= do
				l <- getVars vs
				let f (Field x) xs = addLocalDec xs x (type2str tp)
				    f _ _          = error "Method constructor not handled!"
				return $ foldr f emptyDec l
		--local _ = return emptyDec

		localpars (Type_VariableDeclaratorId tp v)
			= do
				l <- getVars v
				return $ foldr (\(Field x) r -> addLocalDec r x (type2str tp)) emptyDec l


-- | Collect a local variable
getVars :: (Term a, MonadPlus m) => a -> m [MemberName]
getVars = applyTU $ stop_tdTU worker
	where 
	  worker = failTU `adhocTU` vars
	  vars (Identifier1  name) = return [Field name]
	  vars _ = return []

-- | Get parameter list from a term. Must be applied to a method declaration,
-- otherwise, erroneous results mus be expected
getMethodParameters :: (Term a, MonadPlus m) => a -> m [FormalParameter]
getMethodParameters = applyTU (stop_tdTU (failTU `adhocTU` pars))
	where
		pars (Comma2 _ ps) = return ps
		pars _ = return []

-----------------------------------------------------------------------------
