-----------------------------------------------------------------------------
-- |
-- Copyright   :  (c) Universidade do Minho, 2004
-- License     :  LGPL
-- 
-- Maintainer  :  joost.visser@di.uminho.pt
-- Stability   :  experimental
-- Portability :  portable
--
--- Extract Package & Class Graph from Java files.
--- This graph shows relations between and among classes and packages,
--- i.e. inheritance, containment (nesting), imports.
--
-----------------------------------------------------------------------------

module Language.Java.PackageGraph where

-------------------------------------------------------------------------------

import Language.Java.Syntax
import Language.Java.SyntaxTermInstances()
import Language.Java.Util

import StrategyLib
import Control.Monad.Identity
import Data.Monoid
import Data.Graph.Slicing
import Data.Graph.Basics 
import qualified Data.Set as Set
import Data.Relation.GraphViz

-------------------------------------------------------------------------------
-- * Representation

-- | The type of package graphs is a 4-tuple of four relations: import, 
--   nesting, inherit, and implement.
type PackageGraph
  = (Rel PGNode PGNode, 
     Rel PGNode PGNode, 
     Rel PGNode PGNode, 
     Rel PGNode PGNode)
     
-- | The type of package graph nodes.
type PGNode
  = String
  
-- | Empty package graph.
emptyPG :: PackageGraph
emptyPG = (emptyRel, emptyRel, emptyRel, emptyRel)

-- | Appending two package graphs.
appendPG :: PackageGraph -> PackageGraph -> PackageGraph
appendPG (impo,nest,inhe,impl) (impo',nest',inhe',impl')
  = (impo `Set.union` impo',
     nest `Set.union` nest',
     inhe `Set.union` inhe',
     impl `Set.union` impl')

-- | Adding edges to the package graph.
addImport, addNesting, addInherit, addImplement
     :: PGNode -> PGNode -> PackageGraph -> PackageGraph
addImport s t (impo,nest,inhe,impl)     = (Set.insert (s,t) impo,nest,inhe,impl)
addNesting s t (impo,nest,inhe,impl)    = (impo,Set.insert (s,t) nest,inhe,impl)
addInherit s t (impo,nest,inhe,impl)    = (impo,nest,Set.insert (s,t) inhe,impl)
addImplement s t (impo,nest,inhe,impl)  = (impo,nest,inhe,Set.insert (s,t) impl)

-- | A package graph that can serve as legend.
legendPG :: PackageGraph
legendPG = (addImport "importing" "imported") $
           (addNesting "outer" "inner") $
           (addInherit "sub" "super") $
           (addImplement "implementation" "interface" ) $
           emptyPG

-- | Package graphs are monoids.
instance Monoid PackageGraph where
  mempty  = emptyPG
  mappend = appendPG

-- | Create a single relation out of a package graph.  
pgAsRel :: PackageGraph -> Rel Identifier Identifier
pgAsRel (impo,nest,inhe,impl)
  = impo `Set.union` nest `Set.union` inhe `Set.union` impl

-- | Given a relation, compute the sub graph of a package graph that holds
--   only edges that occur in the relation.
projectPG' :: Rel PGNode PGNode -> PackageGraph -> PackageGraph
projectPG' rel (impo,nest,inhe,impl)
  = (sub impo,sub nest,sub inhe,sub impl)
    where
      sub r = projectWith (\x y -> (x,y) `Set.member` rel) r 

-------------------------------------------------------------------------------
-- * Package graph generation

-- | Generate a package graph from a list of compilation units and print it.
--   in the GraphViz format.
java2pcg        :: (MonadPlus m) => [CompilationUnit] -> m String
java2pcg cus    =  do pg <- java2pcg' id "" cus
                      return (pg2dot "pg" pg)

-- | Generate a package graph from a Java fragment.
java2pcg' :: (Term x, MonadPlus m) 
          => (Identifier -> Identifier)   -- ^ Name resolver
          -> PGNode                       -- ^ Current parent.
          -> x                            -- ^ Java fragment. 
          -> m PackageGraph
java2pcg' resolver parent x
  = applyTU worker x
    where
      worker = allTU' worker
                 `adhocTU` cu2pcg
                 `adhocTU` class2pcg
                 `adhocTU` interface2pcg
                 `adhocTU` clsimp2pcg
                 `adhocTU` pkgimp2pcg
                 `adhocTU` type2pcg
      cu2pcg cu@(CU package imports types)
        = do ident  <- return (packDecl2ident package)
             let importMap = mkImportMapCU cu
             let newResolver = maybeResolve importMap ident
             idents <- collectDeclaredTypes types
             iPGs   <- mapM (\i -> java2pcg' resolver (ident++"."++i) imports) idents
             tPG    <- java2pcg' newResolver ident types
             return (foldr appendPG tPG iPGs)
      class2pcg (Class1 _ classname super interfaces body) 
        = do ident <- return (parent++"."++classname)
             sPG   <- return (super2pcg ident super)
             iPG   <- return (interfaces2pcg ident interfaces)
             bPG   <- java2pcg' resolver ident body
             return (addNesting parent ident (sPG `appendPG` iPG `appendPG` bPG))
      interface2pcg (Interface _ classname interfaces body) 
        = do ident <- return (parent++"."++classname)
             iPG   <- return (interfaces2pcg ident interfaces)
             bPG   <- java2pcg' resolver ident body
             return (addNesting parent ident (iPG `appendPG` bPG))
      clsimp2pcg (Import_semicolon name)
        = do let ident = qname2str name
             return (addImport parent ident emptyPG)
      pkgimp2pcg (Import_times_semicolon name)
        = do let ident = qname2str name
             return (addImport parent ident emptyPG)
      type2pcg (Name name)
        = return (addImport parent (resolver(qname2str name)) emptyPG)
             
      super2pcg cls super 
        = foldr (\qn pg -> addInherit cls (qname2str qn) pg) emptyPG (collectNames super)
      interfaces2pcg cls interfaces
        = foldr (\qn pg -> addImplement cls (qname2str qn) pg) emptyPG (collectNames interfaces)
              
-- | Collect all types declared in a Java fragment.
collectDeclaredTypes :: (MonadPlus m, Term t) => t -> m [Identifier]
collectDeclaredTypes x
  = applyTU (stop_tdTU worker) x
    where
      worker = failTU `adhocTU` (\(Class1 _ ident _ _ _)  -> return [ident]) 
                      `adhocTU` (\(Interface _ ident _ _) -> return [ident])

-- | Collect all names from a Java fragment.
collectNames :: Term x => x -> [Name]
collectNames x
  = runIdentity (applyTU (full_tdTU worker) x)
    where
      worker = constTU [] `adhocTU` (\name -> return [name])
 
-------------------------------------------------------------------------------
-- * Functions for Graphviz format generation.

-- | Print a Package Graph in dot format.
pg2dot     :: String         -- ^ Graph name
           -> PackageGraph   -- ^ Graph
           -> String         -- ^ Graphviz format.
pg2dot n (impo,nest,inhe,impl)
  = "digraph "++n++" {\n"++(unlines stmts)++"}\n"
    where
      entities = (ent impo)`Set.union`(ent nest)`Set.union`(ent inhe)`Set.union`(ent impl)
      pkgs = (dom nest) `Set.difference` (rng nest)
      clss = entities `Set.difference` pkgs
      pkgNodes = map (mkNode "shape=box, style=filled, color=tan") (Set.toList pkgs)
      clsNodes = map (mkNode "shape=box, color=green") (Set.toList clss)
      impoEdges = map (mkEdge "color=blue, arrowsize=0.5") (Set.toList impo)
      nestEdges = map (mkEdge "color=tan, arrowsize=0.5") (Set.toList nest)
      inheEdges = map (mkEdgeBack "dir=back, arrowsize=1.5, color=green") (Set.toList inhe)
      implEdges = map (mkEdgeBack "dir=back, arrowsize=1.5, color=green, style=dotted") (Set.toList impl)
      stmts = concat [pkgNodes,clsNodes,impoEdges,nestEdges,inheEdges,implEdges]

mkIfc "" = ""
mkIfc i  = quote i++" [ shape=box, color=green, style=dashed ]\n"
           
-------------------------------------------------------------------------------
