-----------------------------------------------------------------------------
-- |
-- Copyright   :  (c) Universidade do Minho, 2004
-- License     :  LGPL
-- 
-- Maintainer  :  joost.visser@di.uminho.pt
-- Stability   :  experimental
-- Portability :  portable
--
-- Auxiliaries.
--
-----------------------------------------------------------------------------

module Language.Java.Util where

import Language.Java.Syntax
import Language.Java.SyntaxTermInstances()
import Language.Java.SyntaxATermConvertibleInstances()
import Language.Java.SyntaxShowInstances()
import StrategyLib
import Control.Monad.Identity
import Monad()

import Char (toLower)
import qualified Data.Map as Map
import Data.Map(Map)
-----------------------------------------------------------------------------
-- * Java specific.

-- | Extract method name from its header
getMethodName   :: (Monad m, Term t) => t -> m Identifier
getMethodName header
  = case (applyTU (once_buTU (failTU `adhocTU` worker)) header) of
      Just ident  -> return ident
      Nothing     -> error "Method without a name?"
    where
      worker (Comma2 ident _) = return ident
      worker _                = mzero

-- | Fully qualified class name as string
qname2str   :: Name -> String
qname2str (Identifier_p idents) = concat (sepWith "." idents)
qname2str (Class idents)        = concat (sepWith "." ("class":idents))

-- | Convert a package declaration to a String.
packDecl2ident :: Maybe PackageDeclaration -> String
packDecl2ident Nothing                         = "NO PACKAGE"
packDecl2ident (Just (Package_semicolon name)) = qname2str name

-- | Break Java class name into identifier segments.
breakCls :: Identifier -> [Identifier]
breakCls = breakWith (=='.')

-- | Extract all imported class names from a compilation unit.
getDeclaredImports :: CompilationUnit -> [ChaseName]
getDeclaredImports (CU pkg_decl imports types)
  = map qname2str used_cls_names
    where
      used_cls_names 
        = runIdentity (applyTU (full_tdTU worker) imports)
      worker   
        = constTU [] `adhocTU` getImportedType
      getImportedType (Import_semicolon name) 
        = return [name]

-----------------------------------------------------------------------------
-- ** Import Map.maps.

-- | The type of import Map.maps. An import map maps identifiers to fully 
--   qualified identifiers.
type ImportMap
  = Map Identifier Identifier

mkImportMapCU cu
  = mkImportMap (getDeclaredImports cu)

-- | Create the import Map.map for a given list of imported classes.
mkImportMap :: [Identifier] -> ImportMap
mkImportMap [] = defaultImportMap
mkImportMap (qn:qns)
  = Map.insert n qn (mkImportMap qns)
    where
      n = last (breakCls qn)

-- | Captures the built-in scope (java.lang).
defaultImportMap :: ImportMap
defaultImportMap
  = Map.fromList (map (\c -> (c,"java.lang."++c)) javaLangClasses)
    where
      javaLangClasses 
        = interfaces142++classes142++exceptions142++errors142
        
-- | Turn an identifier into a qualified one
--   based on the given import Map.map, and the name of the
--   current package. If the identifier is qualified already
--   do nothing.
maybeResolve :: ImportMap -> Identifier -> Identifier -> Identifier
maybeResolve importMap pkgName clsName
  | isQualified clsName 
    = clsName
  | otherwise           
    = Map.findWithDefault (pkgName++"."++clsName) clsName importMap 
  
-- | Determine whether a given class name is fully qualified.
isQualified :: Identifier -> Bool
isQualified clsName
  = '.' `elem` clsName

-- | Turn a non-qualified identifier into a qualified one
--   based on the given import Map.map, and the name of the
--   current package.
resolveWithImportMap :: ImportMap -> Identifier -> Identifier -> Identifier
resolveWithImportMap importMap pkgStr name
  = Map.findWithDefault (pkgStr++"."++name) name importMap  

-- | Interfaces predefined in package java.lang in JDK 1.4.2.
interfaces142 :: [Identifier]
interfaces142
  = [ "CharSequence"
    , "Cloneable"
    , "Comparable"
    , "Runnable"
    ] 

-- | Classes predefined in package java.lang in JDK 1.4.2.
classes142 :: [Identifier]
classes142
  = [ "Boolean"
    , "Byte"
    , "Character"
    , "Character.Subset"
    , "Character.UnicodeBlock"
    , "Class"
    , "ClassLoader"
    , "Compiler"
    , "Double"
    , "Float"
    , "InheritableThreadLocal"
    , "Integer"
    , "Long"
    , "Math"
    , "Number"
    , "Object"
    , "Package"
    , "Process"
    , "Runtime"
    , "RuntimePermission"
    , "SecurityManager"
    , "Short"
    , "StackTraceElement"
    , "StrictMath"
    , "String"
    , "StringBuffer"
    , "System"
    , "Thread"
    , "ThreadGroup"
    , "ThreadLocal"
    , "Throwable"
    , "Void"
    ] 

-- | Exceptions predefined in package java.lang in JDK 1.4.2.
exceptions142 :: [Identifier]
exceptions142
  = [ "ArithmeticException"
    , "ArrayIndexOutOfBoundsException"
    , "ArrayStoreException"
    , "ClassCastException"
    , "ClassNotFoundException"
    , "CloneNotSupportedException"
    , "Exception"
    , "IllegalAccessException"
    , "IllegalArgumentException"
    , "IllegalMonitorStateException"
    , "IllegalStateException"
    , "IllegalThreadStateException"
    , "IndexOutOfBoundsException"
    , "InstantiationException"
    , "InterruptedException"
    , "NegativeArraySizeException"
    , "NoSuchFieldException"
    , "NoSuchMethodException"
    , "NullPointerException"
    , "NumberFormatException"
    , "RuntimeException"
    , "SecurityException"
    , "StringIndexOutOfBoundsException"
    , "UnsupportedOperationException"
    ] 

-- | Errors predefined in package java.lang in JDK 1.4.2.
errors142 :: [Identifier]
errors142
  = [ "AbstractMethodError"
    , "AssertionError"
    , "ClassCircularityError"
    , "ClassFormatError"
    , "Error"
    , "ExceptionInInitializerError"
    , "IllegalAccessError"
    , "IncompatibleClassChangeError"
    , "InstantiationError"
    , "InternalError"
    , "LinkageError"
    , "NoClassDefFoundError"
    , "NoSuchFieldError"
    , "NoSuchMethodError"
    , "OutOfMemoryError"
    , "StackOverflowError"
    , "ThreadDeath"
    , "UnknownError"
    , "UnsatisfiedLinkError"
    , "UnsupportedClassVersionError"
    , "VerifyError"
    , "VirtualMachineError"
    ]
    
-----------------------------------------------------------------------------
-- * Generic.

-- | Insert separators into a list.
sepWith :: a -> [a] -> [a]
sepWith _ [] = []
sepWith a [x] = [x]
sepWith a (x:xs) = x: a: sepWith a xs

-- | Break a list into segments, using a predicate on elements to 
--   indicate break points.
--   The break points themselves will be discarded.
breakWith :: (a -> Bool) -> [a] -> [[a]] 
breakWith f x = case break f x of
	(p,(_:pp)) -> p: breakWith f pp
	(p,[]) -> [p]

-- | Break a file name into its segments, i.e. the pieces between slashes.    
breakFile :: FilePath -> [String]
breakFile = breakWith (=='/')
       
-- | Break path argument into directory names. i.e. the pieces between colons.
breakPath :: String -> [FilePath]
breakPath = breakWith (==':')

-----------------------------------------------------------------------------
-- * For parsing

javaTable :: String
javaTable = "Language/Java/Java.tbl"

-----------------------------------------------------------------------------


-- * Helper funcions for call graph generation

-- | Function to determine the types of arguments in a function
-- declaration.

getParameterTypes :: [FormalParameter] -> [String]
getParameterTypes = map (\(Type_VariableDeclaratorId tp _) -> type2str tp)

type2str :: Type -> String
type2str (PrimitiveType pt) = primitive2str pt
type2str (ReferenceType rt) = reference2str rt

reference2str (ClassOrInterfaceType (Name ci)) = name2str ci
reference2str (ArrayType arr) = array2str arr
	where
		array2str (PrimitiveType1 pt) = primitive2str pt ++ "[]"
		array2str (ArrayType0 n) = name2str n ++ "[]"
		array2str (ArrayType1 arr) = array2str arr ++ "[]"

primitive2str :: PrimitiveType -> String
primitive2str Boolean = "boolean"
primitive2str (NumericType nt) = numeric2str nt
	where
		numeric2str (IntegralType it) = map toLower $ show it
		numeric2str (FloatingPointType ft) = map toLower $ show ft

name2str :: Name -> String
name2str = snd . sepname

sepname :: Name -> ([String], String)
sepname (Identifier_p l) = (init l, last l)
sepname (Class l) = (init l, last l)

-----------------------------------------------------------------------------

