module Language.Pointwise.Parser where

import Language.Pointwise.Syntax
import Language.Haskell.Syntax
import Language.Haskell.Pretty
import Data.Char

{- 
Parsing of a HsExp to a pointwise term.
It recognizes:
  G, G1, G2 ::=
      (G) | undefined | _L | inn G | out G
    | 'literal' | 'var' | fix G | (G1,G2)
    | fst G | snd G | Left G | RightG | G1 G2 | \ 'var' -> G 
    | case G of Left var1 -> G1 ; Right var2 -> G2
    | case G of Right var1 -> G1 ; Left var2 -> G2
    | case G of ...  
-}

mkVar  = HsVar . UnQual . HsIdent

hs2pw :: HsExp -> Maybe Term

hs2pw (HsParen e) = hs2pw e

-- unit -> "undefined" or "_L"
hs2pw (HsVar(UnQual(HsIdent "undifined"))) = return $ Unit
hs2pw (HsVar(UnQual(HsIdent "_L")))        = return $ Unit

-- Constants
hs2pw (HsApp (HsVar (UnQual (HsIdent "inn"))) exp)
  = hs2pw exp >>= return . In
hs2pw (HsApp (HsVar (UnQual (HsIdent "out"))) exp)
  = hs2pw exp >>= return . Out
hs2pw (HsLit lit) = return $ Const $ prettyPrint lit
hs2pw (HsVar(UnQual(HsIdent str)))    = return $ Var str
hs2pw (HsInfixApp e1 (HsQConOp (Special HsCons)) e2)
  = do t1 <- hs2pw e1
       t2 <- hs2pw e2
       return $ ((Const ":") :@: t1) :@: t2
hs2pw (HsList []) = return $ Const "[]"
hs2pw (HsList (x:xs))
  = do e <- hs2pw x
       es <- hs2pw (HsList xs)
       return $ ((Const ":") :@: e) :@: es

-- Recursion
hs2pw (HsApp (HsVar (UnQual (HsIdent "fix"))) exp) =
  do term <- hs2pw exp
     return $ Fix term

-- remaining
hs2pw (HsTuple [e1,e2]) =
   do t1 <- hs2pw e1
      t2 <- hs2pw e2
      return $ t1 :&: t2
hs2pw (HsApp (HsVar (UnQual (HsIdent "fst"))) e) =
   do t <- hs2pw e
      return $ Fst t
hs2pw (HsApp (HsVar (UnQual (HsIdent "snd"))) e) =
   do t <- hs2pw e
      return $ Snd t
hs2pw (HsApp (HsCon (UnQual (HsIdent "Left"))) e) =
   do t <- hs2pw e
      return $ Inl t
hs2pw (HsApp (HsCon (UnQual (HsIdent "Right"))) e) =
   do t <- hs2pw e
      return $ Inr t
hs2pw (HsApp e1 e2) =
   do t1 <- hs2pw e1
      t2 <- hs2pw e2
      return $ t1 :@: t2
hs2pw (HsLambda _ [HsPVar(HsIdent str)] e) =
   do t <- hs2pw e
      return $ Lam str t
-- in "case of"'s, guards fail and declarations are lost
hs2pw (HsCase e1
        [HsAlt _ (HsPApp (UnQual(HsIdent "Left"))
         [HsPVar(HsIdent str2)]) (HsUnGuardedAlt e2) _,
         HsAlt _ (HsPApp (UnQual(HsIdent "Right"))
         [HsPVar(HsIdent str3)]) (HsUnGuardedAlt e3) _]) =
   do t1 <- hs2pw e1
      t2 <- hs2pw e2
      t3 <- hs2pw e3
      return $ Case t1 (Lam str2 t2) (Lam str3 t3)
hs2pw (HsCase e1
        [HsAlt _ (HsPApp (UnQual(HsIdent "Right"))
         [HsPVar(HsIdent str3)]) (HsUnGuardedAlt e3) _,
         HsAlt _ (HsPApp (UnQual(HsIdent "Left"))
         [HsPVar(HsIdent str2)]) (HsUnGuardedAlt e2) _]) =
   do t1 <- hs2pw e1
      t2 <- hs2pw e2
      t3 <- hs2pw e3
      return $ Case t1 (Lam str2 t2) (Lam str3 t3)
hs2pw (HsCase e alts) =
   do t1 <- hs2pw e
      ts <- mapM alt2pws alts
      return $ Match t1 ts
  where alt2pws (HsAlt _ pat (HsUnGuardedAlt e) _) =
          do tp <- pat2pw pat
             te <- hs2pw e
             return (tp,te)
        alt2pws _ = fail "No guards allowed."
hs2pw (HsCon (UnQual (HsIdent x))) = return $ Const x
hs2pw t = fail $ "'"++prettyPrint t++
                 "' is not a valid pointwise term."


hsPat2HsExp :: HsPat -> HsExp
hsPat2HsExp (HsPVar hsName)     = HsVar $ UnQual hsName
hsPat2HsExp (HsPLit hsLiteral)  = HsLit hsLiteral
hsPat2HsExp (HsPNeg hsPat)      = HsNegApp . hsPat2HsExp $ hsPat
hsPat2HsExp (HsPInfixApp hsPat1 hsQName hsPat2) =
     let hsExp1 = hsPat2HsExp hsPat1
         hsExp2 = hsPat2HsExp hsPat2
         hsQOp = (if f hsQName  then HsQConOp else HsQVarOp) hsQName
     in HsInfixApp hsExp1 hsQOp hsExp2
 where
       f (Qual _ name) = g name
       f (UnQual name) = g name
       f (Special _  ) = True
       g (HsIdent name) = isUpper $ head name
       g (HsSymbol str) = isUpper $ head str
hsPat2HsExp (HsPApp hsQName []) = HsCon hsQName
hsPat2HsExp (HsPApp hsQName lHsPat) =
    foldl HsApp (HsCon hsQName) . map hsPat2HsExp $ lHsPat
hsPat2HsExp (HsPTuple lHsPat)   = HsTuple $ map hsPat2HsExp lHsPat
hsPat2HsExp (HsPList lHsPat)    = HsList  $ map hsPat2HsExp lHsPat
hsPat2HsExp (HsPParen hsPat)    = HsParen $ hsPat2HsExp hsPat
hsPat2HsExp (HsPRec hsQName lHsPatField) =
     HsRecConstr hsQName (map f lHsPatField)
 where
   f (HsPFieldPat hsQName hsPat) = HsFieldUpdate hsQName $ hsPat2HsExp hsPat
hsPat2HsExp (HsPAsPat hsName hsPat) = HsAsPat hsName (hsPat2HsExp hsPat)
hsPat2HsExp (HsPWildCard)           = mkVar "_L"
hsPat2HsExp (HsPIrrPat hsPat)       = HsIrrPat $ hsPat2HsExp hsPat


-- this approach may be changed...
pat2pw = hs2pw . hsPat2HsExp
