-----------------------------------------------------------------------------
-- |
-- Maintainer  :  Joost Visser
-- Stability   :  experimental
-- Portability :  portable
--
-- Some analyses on XML schemas.
--
-----------------------------------------------------------------------------

module Language.XMLSchema.Analysis where

import Data.Set as Set
import Data.Maybe
import StrategyLib
import Language.XMLSchema.XMLSchemaLib
import Control.Monad.Identity
import Data.GraphR
import Data.Map as Map
import Data.Graph.Metrics
import Data.Metrics hiding (name)
import Control.Monad.State

import Text.XML.HaXml.Types -- (Element(..),AttValue(..))
import Text.XML.HaXml.OneOfN -- (TwoOf2,SixOf13)
import Text.XML.HaXml.Xml2Haskell (Defaultable(..))

import System (system,ExitCode(..),getProgName)
import System.IO
import System.Process
import System.Directory
-- import System.Posix.Temp

import Debug.Trace 

-----------------------------------------------------------------------------
-- * Basic collection functions

-- | Collect all element names from a given XSD fragment.
collectElementNames :: Term xsd => xsd -> Set String
collectElementNames = Set.fromList . runIdentity . traversal
  where
    traversal = applyTU $ collect $ adhocTU (constTU []) worker
    --worker (Xs'element attrs _ _ _)
    worker attrs
      = return $ catMaybes [ xs'elementName attrs, xs'elementRef attrs ]

--

{-
getXsdUrls :: Text.XML.HaXml.Types.Element -> [String]
getXsdUrls (Text.XML.HaXml.Types.Elem _ attrs _) = foldr worker [] attrs
 where
   worker ('x':'m':'l':'n':'s':':':name,AttValue [Left url]) urls = (url++".xsd"):urls
   worker (a,v) urls = (a++"="++show v):urls
   worker _ urls     = urls

instance Show AttValue where
    showsPrec d (AttValue aa) = showParen (d >= 10)
              (showString "AttValue" . showChar ' ' . showsPrec 10 aa)

instance Show Reference where
    showsPrec d (RefEntity aa) = showParen (d >= 10)
              (showString "RefEntity" . showChar ' ' . showsPrec 10 aa)
    showsPrec d (RefChar aa) = showParen (d >= 10)
              (showString "RefChar" . showChar ' ' . showsPrec 10 aa)
-}

-- | Obtain the filenames, relative to a search directory,
--   from the import, include, and redefine constructs of
--   a given schema.
getSchemaLocations :: Schema -> [ChaseName]
getSchemaLocations (infile,_,schema,_,_)
  = Prelude.map fileNameRelToRoot locations
  where
    dir = noBasename infile
    locations = runIdentity $ applyTU (full_tdTU worker) schema
    worker = constTU [] 
              `adhocTU` onImportAttrs
              `adhocTU` onIncludeAttrs
              `adhocTU` onRedefineAttrs
    onImportAttrs attrs 
      = case (xs'importSchemaLocation attrs) of
          Nothing -> return []
          (Just "") -> return []
          (Just l)  -> return [l]
    onIncludeAttrs attrs = return [xs'includeSchemaLocation attrs]
    onRedefineAttrs attrs = return [xs'redefineSchemaLocation attrs]
    fileNameRelToRoot ('h':'t':'t':'p':':':'/':'/':f) = f
    fileNameRelToRoot f = simplifyFileName $ dir++f

-----------------------------------------------------------------------------
-- * Namespace management

-- | The type of universal resource identifiers used in XML namespaces.
type URI = String

-- | The type of prefixes used in XML documents to refer to XML namespaces.
type Prefix = String

-- | Type of XML namespace environments, which may store a default namespace,
--   a target namespace, and a map of prefixes to namespaces.
data NameSpaceEnv  = NameSpaceEnv {
   defaultNameSpace :: Maybe URI
 , targetNameSpace :: Maybe URI
 , prefixMap :: Map Prefix URI
 } deriving (Ord,Show,Eq)

-- | Create an empty name space environment.
emptyNameSpaceEnv :: NameSpaceEnv
emptyNameSpaceEnv = NameSpaceEnv {
   defaultNameSpace = Nothing
 , targetNameSpace = Nothing
 , prefixMap = Map.empty
} 

-- | Retrieve the URI associated to a given prefix from the given
--   namespace environment.
getNS :: NameSpaceEnv -> Prefix -> Maybe URI
getNS env p = Map.lookup p (prefixMap env)

getNS' env p = maybe (trace ("Unknown prefix: "++p++" in "++(show env)) Nothing) Just $ getNS env p

-- | Extract a namespace environment from the XML representation of an XML schema.
--   This function is defined on raw XML rather than the Xs'schema type, because
--   not all necessary information is present in that latter representation.
createNameSpaceEnv :: Element         -- ^ XML element that represent the body of an XML schema. 
                   -> NameSpaceEnv
createNameSpaceEnv (Elem _ attrs _) 
  = foldr update (insertXmlPrefix emptyNameSpaceEnv) attrs
 where
  update ("xmlns",AttValue [Left uri]) nse                          -- xmlns="http://..."
     = nse { defaultNameSpace = Just uri }
  update ('x':'m':'l':'n':'s':':':prefix,AttValue [Left uri]) nse   -- xmlns:foo="http://..."
     = nse { prefixMap = Map.insert prefix uri $ prefixMap nse }
  update ("targetNamespace",AttValue [Left uri]) nse                -- targetNamespace="http://..."
     = nse { targetNameSpace = Just uri }
  update _ nse
     = nse

insertXsPrefix :: NameSpaceEnv -> NameSpaceEnv
insertXsPrefix nse
  = nse { prefixMap = Map.insert "xs" "http://www.w3.org/2001/XMLSchema" (prefixMap nse) }

insertXmlPrefix :: NameSpaceEnv -> NameSpaceEnv
insertXmlPrefix nse
  = nse { prefixMap = Map.insert "xml" "http://www.w3.org/XML/1998/namespace" (prefixMap nse) }

-- | Inverse lookup of prefix that maps to the target namespace.
--   A run-time error is produced when the namespace is mapped to by more than one prefix.
targetPrefix :: NameSpaceEnv -> Maybe Prefix
targetPrefix NameSpaceEnv{ targetNameSpace=Just tns, prefixMap=mp }
  = Map.foldWithKey worker Nothing mp
  where
    worker pfx ns Nothing  = if ns==tns then Just pfx else Nothing
    worker _ ns (Just pfx) = if ns==tns then error "Two prefixes map to target name space!" else (Just pfx)
targetPrefix _ = Nothing

-- | Inverse lookup of prefix that maps to the XML Schema namespace, i.e. "http://www.w3.org/2001/XMLSchema".
--   A run-time error is produced when that namespace is mapped to by more than one prefix.
xmlSchemaPrefix :: NameSpaceEnv -> Maybe Prefix
xmlSchemaPrefix nse@NameSpaceEnv{ prefixMap=mp, defaultNameSpace=dns }
  = Map.foldWithKey worker start mp
  where
    start = if dns==Just tns then Just "" else Nothing
    tns = "http://www.w3.org/2001/XMLSchema"
    worker pfx ns Nothing  = if ns==tns then Just pfx else Nothing
    worker _ ns (Just pfx) = if ns==tns then error msg else (Just pfx)
    msg = "Two prefixes map to XMLSchema name space! "++(show nse)
xmlSchemaPrefix _ = Nothing

-- | Inverse lookup of prefixes that maps to the XML Schema namespace, 
--   i.e. "http://www.w3.org/2001/XMLSchema".
xmlSchemaPrefixes :: NameSpaceEnv -> [Prefix]
xmlSchemaPrefixes nse@NameSpaceEnv{ prefixMap=mp, defaultNameSpace=dns }
  = Map.foldWithKey worker start mp
  where
    start = if dns==Just tns then [""] else []
    tns = "http://www.w3.org/2001/XMLSchema"
    worker pfx ns pfxs  = if ns==tns then (pfx:pfxs) else pfxs
xmlSchemaPrefixes _ = []

-----------------------------------------------------------------------------
-- * Schema sets

-- | Type of a set of schemas (belonging to a single project, usually).
type SchemaSet = [Schema]

-- | Schema in various forms: file name, raw XML, parsed.
type Schema = (FilePath, Element, Xs'schema, NameSpaceEnv,FilePath)

-- | Retrieve the relative file name of a given schema.
fileSchema :: Schema -> FilePath
fileSchema (file,_,_,_,_) = file

-- | Retrieve the absolute file name of a given schema.
absoluteFileName :: Schema -> FilePath
absoluteFileName (_,_,_,_,file) = file

-- | Retrieve the raw XML representation of a given schema.
rawSchema :: Schema -> Element
rawSchema (_,element,_,_,_) = element

-- | Retrieve the parsed representation of a given schema.
parsedSchema :: Schema -> Xs'schema
parsedSchema (_,_,schema,_,_) = schema

nseSchema :: Schema -> NameSpaceEnv
nseSchema (_,_,_,nse,_) = nse

-----------------------------------------------------------------------------
-- * Graph representations of XML Schemas.

-- | Type of nodes of the XML Schema successor graph.
data XsdNode 
 = XsdNode { prefix :: Maybe Prefix
           , parent :: Maybe XsdNode
           , name :: String
           , kind :: XsdNodeKind
           } 
 deriving (Ord,Eq,Show) 

-- | Return the most distant parent of given node, which is the node
--   itself in case of a global node, and which is the global node
--   to which a node is local in the case of a local node.
urParent n@XsdNode{ parent=Nothing } = n
urParent XsdNode{ parent=Just p} = urParent p

-- | Enumeration type of the various kinds of XsdNode.
data XsdNodeKind = ElementKind | AttributeKind | TypeKind
                 | GroupKind | AttributeGroupKind | OtherKind deriving (Ord,Eq,Show)
 
-- | Pretty-print a given XsdNode.
--   Kind information is not present in the pretty-printed version.
showXsdNode :: XsdNode -> String
showXsdNode XsdNode{ prefix=pfx, parent=p, name=nm, kind=knd }
  = (showParents p) ++ (maybe "" (++":") pfx) ++ nm
    where  
      showParents :: Maybe XsdNode -> String
      showParents Nothing  = ""
      showParents p = "{"++(concat $ sepBy "." $ reverse $ listNames p)++"}"
      listNames Nothing = []
      listNames (Just (XsdNode{parent=p,name=nm})) = (nm:listNames p)  
      sepBy sep [] = []
      sepBy sep [x] = [x]
      sepBy sep (x:xs) = x:sep:(sepBy sep xs)

-- | Create a global XsdNode, with given prefix, name, and kind.
mkGlobalXsdNode 
 :: Maybe Prefix    -- ^ prefix to use when the name does not include one.
 -> String          -- ^ name
 -> XsdNodeKind     -- ^ kind
 -> XsdNode
mkGlobalXsdNode = mkXsdNode Nothing

-- | Create an XsdNode, with given parent, prefix, name, and kind. 
mkXsdNode 
 :: Maybe XsdNode   -- ^ parent
 -> Maybe Prefix    -- ^ prefix to use when the name does not include one.
 -> String          -- ^ name
 -> XsdNodeKind     -- ^ kind
 -> XsdNode
mkXsdNode prnt prefixDefault prefixedName knd
  = XsdNode { prefix = pfx, parent = prnt, name = nm, kind = knd  }
    where
      (x,y) = break (==':') prefixedName
      (pfx,nm) = if length y > 1 then (Just x,tail y) else (prefixDefault,x)

-- | Create an XsdNode, with given parent, prefix, name, and kind. 
mkXsdNodeUsed 
 :: Set XsdNode     -- ^ Chameleon nodes (declared by included schemas that have no target namespace)
 -> NameSpaceEnv    -- ^ To resolve prefix or provide default namespace.
 -> String          -- ^ name
 -> XsdNodeKind     -- ^ kind
 -> XsdNode
mkXsdNodeUsed chameleons nse prefixedName knd
  = if (pfx==(targetNameSpace nse))
       then maybe newNode id chamNode
       else newNode
    where
      uriDefault = defaultNameSpace nse
      (x,y) = break (==':') prefixedName
      (pfx,nm) = if length y > 1 then (getNS' nse x,tail y) else (uriDefault,x)
      newNode = XsdNode { prefix = pfx, parent = Nothing, name = nm, kind = knd  }
      chamNode = findChameleonNode chameleons nm knd

-- | Break an element name into its prefix and tag.
breakPrefixTag :: String -> (Prefix,String)
breakPrefixTag prefixedName = (pfx,nm)
 where
      (x,y) = break (==':') prefixedName
      (pfx,nm) = if length y > 1 then (x,tail y) else ("",x)
  
-----------------------------------------------------------------------------
-- * Successor graph construction

-- | Extract a map that associates to each schema file name
--   the set of global XSD nodes that are declared in it.
declaredXsdNodes :: SchemaSet -> Map FilePath (Set XsdNode)
declaredXsdNodes schemas = foldr worker Map.empty schemas
 where
  worker schema
   = Map.insert (fileSchema schema) (declaredXsdNodes1 schema)
  declaredXsdNodes1 schema
   = mkSet $ mapMaybe (onGlob . fst) glbls 
     where
      nse = nseSchema schema
      targetNS = targetNameSpace nse
      (Xs'schema attrs _ glbls) = parsedSchema schema
      onGlob (OneOf7 (Xs'simpleType attrs _ _)) = mkNode (xs'simpleTypeName attrs) TypeKind
      onGlob (TwoOf7 (Xs'complexType attrs _ _)) = mkNode (xs'complexTypeName attrs) TypeKind
      onGlob (ThreeOf7 (Xs'element attrs _ _ _)) = mkNode (xs'elementName attrs) ElementKind
      onGlob (FourOf7 (Xs'attribute attrs _ _)) = mkNode (xs'attributeName attrs) AttributeKind
      onGlob (FiveOf7 (Xs'attributeGroup attrs _ _ _)) = mkNode (xs'attributeGroupName attrs) AttributeGroupKind
      onGlob (SixOf7 (Xs'group attrs _ _)) = mkNode (xs'groupName attrs) GroupKind
      onGlob (SevenOf7 (x::Xs'notation)) = Nothing
      mkNode nm knd = maybe Nothing (\n -> Just $ mkGlobalXsdNode targetNS n knd) nm

-- | Extract a map that associates to each schema file name
--   the set of nodes that are visible via a chameleon include.
chameleonXsdNodes :: SchemaSet -> Map FilePath (Set XsdNode)
chameleonXsdNodes schemas
 = foldr worker Map.empty schemas
   where
     declaredNodes = declaredXsdNodes schemas
     chameleonFiles = chameleonSchemaFilenames schemas
     worker schema mp = Map.insert (fileSchema schema) nodes mp 
       where
         fileNames = Prelude.filter (`elem` chameleonFiles) $ getIncludedSchemaLocations schema
         nodes = foldr (\f s -> maybe s (Set.union s) $ Map.lookup f declaredNodes) Set.empty fileNames
      
-- | List file names of schemas without target namespace.
chameleonSchemaFilenames :: SchemaSet -> [FilePath]
chameleonSchemaFilenames schemas 
  = foldr worker [] schemas
    where
      worker schema files
        | (targetNameSpace $ nseSchema schema)==Nothing = (fileSchema schema):files
        | otherwise = files

-- | Obtain the filenames, relative to a search directory,
--   from the include constructs of
--   a given schema.
getIncludedSchemaLocations :: Schema -> [ChaseName]
getIncludedSchemaLocations (infile,_,schema,_,_)
  = Prelude.map (\l -> simplifyFileName $ (dir++l)) locations
  where
    dir = noBasename infile
    locations = runIdentity $ applyTU (full_tdTU worker) schema
    worker = constTU [] 
              `adhocTU` onIncludeAttrs
    onIncludeAttrs attrs = return [xs'includeSchemaLocation attrs]

findChameleonNode :: Set XsdNode -> String -> XsdNodeKind -> Maybe XsdNode
findChameleonNode nodes nm knd
 = Set.fold worker Nothing nodes
   where
    worker node found
      | (nm==(name node)) && (knd==(kind node))
        = maybe (Just node) reportDouble found
      | otherwise
        = found 
    reportDouble node
     = trace ("Multiple candidates to resolve Chameleon node "++(show node)) (Just node)

-- | Construct the successor graph of a given schema set.
successorGRAPH :: SchemaSet -> Gph XsdNode
successorGRAPH schemas = foldl (Set.union) (Set.empty) graphs
  where
    graphs = Prelude.map mkGraph schemas
    mkGraph schema 
     = successorGraph (maybe Set.empty id $ Map.lookup (fileSchema schema) chameleons) 
                      (nseSchema schema) (parsedSchema schema)
    chameleons = chameleonXsdNodes schemas

-- | Construct the successor graph of a given XSD fragment.
successorGraph 
  :: Term xsd 
  => Set XsdNode    -- ^ Chameleon nodes (declared by included schemas that have no target namespace)
  -> NameSpaceEnv   -- ^ namespace environment, needed to get prefixes right
  -> xsd            -- ^ schema, or subelement thereof
  -> Gph XsdNode    -- ^ successor graph
successorGraph chameleons nse 
 = successorGraph' chameleons nse Nothing

-- | Worker function for constructing the successor graph.
successorGraph' 
  :: Term xsd 
  => Set XsdNode    -- ^ Chameleon nodes (declared by included schemas that have no target namespace)
  -> NameSpaceEnv   -- ^ namespace environment, needed to get prefixes right
  -> Maybe XsdNode  -- ^ parent
  -> xsd            -- ^ schema, or subelement thereof
  -> Gph XsdNode    -- ^ successor graph
successorGraph' chameleons nse parent xsd = maybe emptyRel id (applyTU traversal xsd)
  where 
    targetNS  = targetNameSpace nse
    defaultNS = defaultNameSpace nse
    -- targetPfx = targetPrefix nse
    traversal = stop_tdTU $ failTU `adhocTU` onElement `adhocTU` onAttribute
                                   `adhocTU` onGroup `adhocTU` onAttributeGroup
                                   `adhocTU` onComplexType `adhocTU` onSimpleType
                                   `adhocTU` onRestriction `adhocTU` onExtension
                                   `adhocTU` onList `adhocTU` onUnion
    onElement (Xs'element attrs _ content _) 
      = worker' (xs'elementName attrs) 
               (xs'elementType attrs)
               (xs'elementRef attrs)
               content ElementKind
    onGroup (Xs'group attrs _ content) 
      = worker' (xs'groupName attrs) 
               Nothing
               (xs'groupRef attrs) 
               content GroupKind
    onAttribute (Xs'attribute attrs _ content) 
      = worker' (xs'attributeName attrs) 
               (xs'attributeType attrs)
               (xs'attributeRef attrs)  
               content AttributeKind
    onAttributeGroup (Xs'attributeGroup attrs _ content content') 
      = worker' (xs'attributeGroupName attrs) 
               Nothing
               (xs'attributeGroupRef attrs) 
               (content,content') AttributeGroupKind
    onComplexType (Xs'complexType attrs _ content) 
      = worker' (xs'complexTypeName attrs) Nothing Nothing content TypeKind
    onSimpleType (Xs'simpleType attrs _ content) 
      = worker' (xs'simpleTypeName attrs) Nothing Nothing content TypeKind
    onRestriction (Xs'restriction attrs _ content content') 
      = worker' Nothing 
               (xs'restrictionBase attrs) 
               Nothing
               (content,content') TypeKind
    onExtension (Xs'extension attrs _ content) 
      = worker' Nothing 
               (Just $ xs'extensionBase attrs) 
               Nothing
               content TypeKind
    onList (Xs'list attrs _ content)
      = worker' Nothing 
               (xs'listItemType attrs) 
               Nothing 
               content TypeKind
    onUnion (Xs'union attrs _ content)
      = foldM
          (\g t -> worker' Nothing (Just t) Nothing () TypeKind >>= \g' -> return $ g' `Set.union` g)
          (successorGraph' chameleons nse parent content)
          (maybe [] words $ xs'unionMemberTypes attrs)
    worker' (Just n) _ (Just r) _ _ = error $ "XSD component has both name ("++n++") and ref("++r++")"
    worker' name tpe ref content knd
      = return graph''
        where
          newParent = maybe parent (\nm -> Just $ mkXsdNode parent targetNS nm knd) name
          subGraph  = successorGraph' chameleons nse newParent content
          graph = case (newParent,tpe) of  -- Adding edge to type
                    (Just p, Just t) -> Set.insert (p, (mkXsdNodeUsed chameleons nse t TypeKind)) subGraph
                    otherwise        -> subGraph
          graph' = case (parent,name) of  -- Adding edge between new and old parent
                     (Just p, Just n) -> Set.insert (p,(mkXsdNode parent targetNS n knd)) graph
                     otherwise        -> graph
          graph'' = case (newParent,ref) of  -- Adding edge to ref
                      (Just p, Just r) -> Set.insert (p, (mkXsdNodeUsed chameleons nse r knd)) graph'
                      otherwise        -> graph'

-- | Derive the globals graph from a successor graph.
globalsGraph :: Gph XsdNode -> Gph (Set XsdNode)
globalsGraph g = partitionGraph groups g
  where
    groups = Set.fromList $ Map.elems $ splitByGlobals $ ent g
    
-- | Split the domain of a successor graph by top-level declaration.
splitByGlobals :: Set XsdNode -> Map XsdNode (Set XsdNode)
splitByGlobals ns 
 = Set.fold worker Map.empty ns
   where
     worker n mp = Map.insertWith Set.union (urParent n) (Set.singleton n) mp
     
-- | Derive import graph. Nodes are relative file names.
--   Edges represent import, include, and redefine relations.
importGraph :: SchemaSet -> Gph FilePath
importGraph schemas
 = foldr insrtN Set.empty schemas
   where
     insrtN s g = foldr (insrt1 s) g $ getSchemaLocations s
     insrt1 s f = Set.insert (fileSchema s,f)
     -- insrtN s g = foldr (insrt1 s) g $ getSchemaLocations s
     -- insrt1 s f g = maybe g (\s' -> Set.insert (s,s') g) $ Map.lookup f fileSchemaMap
     -- fileSchemaMap = foldr (\s -> Map.insert (fileSchema s) s) Map.empty schemas

-----------------------------------------------------------------------------
-- * Metrics     
     
-- | Compute metrics grouped by top-level declarations.
calcMetricsByToplevel :: Gph XsdNode -> GroupMetric'
calcMetricsByToplevel g 
  = calcCouplingMetrics "Globals graph metrics (coupling & coherence)" 
     (Set.fromList . Map.elems . splitByGlobals . ent) g

-- | Compute metrics grouped by strongly-connected components.
--   Not used.
calcMetricsByStrongComponents :: Ord a => Gph a -> GroupMetric'
calcMetricsByStrongComponents g 
  = calcCouplingMetrics "Coupling metrics for strong components"
      strongComponentSet g

-- | Compute graph metrics based on a given grouping function.
calcCouplingMetrics :: Ord a
                    => String                   -- ^ Title
                    -> (Gph a -> Set (Set a))   -- ^ Function to split nodes into groups
                    -> Gph a 
                    -> GroupMetric'
calcCouplingMetrics title split g = calcCouplingMetrics' title (split g) g 

calcCouplingMetrics' :: Ord a 
                     => String
                     -> Set (Set a)
                     -> Gph a
                     -> GroupMetric'
calcCouplingMetrics' title groups g =
  GroupMetric' title [
    mkMetric' "Efferent coupling (# of outgoing edges)" "Ce" (basicStats' groups mce),
    mkMetric' "Afferent coupling (# of outgoing edges)" "Ca" (basicStats' groups mca),
    mkMetric' "Internal coupling (# of internal edges)" "Ci" (basicStats' groups mci),
    mkMetric' "Instability based on coupling" "I" (basicStats' groups mi),
    mkMetric' "Coherence (ratio of internal edges)" "C" (basicStats' groups mc),
    mkMetric' "Tree impurity" "TI" (basicStats' groups mTI)
  ] 
  where
    -- mkMetric d n x = Left $ Metric d n x
    -- Prefix abbreviation with first letter of the graph name.
    mkMetric' d n x = Right $ Metric d ((head title):n) x
    (mce,mca,mci,mi,mc) 
      = Map.foldWithKey worker (Map.empty,Map.empty,Map.empty,Map.empty,Map.empty) cplngs
    worker gbl (ce,ca,ci,i,c) (mce,mca,mci,mi,mc) 
      = (Map.insert gbl ce mce, Map.insert gbl ca mca, Map.insert gbl ci mci, 
         Map.insert gbl i mi, Map.insert gbl c mc)
    cplngs = couplings groups g
    subgraphs = subGraphs groups g
    mTI = Map.foldWithKey (\gr g mp -> Map.insert gr (treeImpurity g) mp) Map.empty subgraphs


-----------------------------------------------------------------------------
-- * XML-agnostic schema size

calcXmlAgnosticMetrics :: [FilePath] -> IO GroupMetric
calcXmlAgnosticMetrics fileNames = do
  (lines,bytes) <- foldM (\(l,b) f -> wordCount f >>= \(l',_,b') -> return (l+l',b+b')) (0,0) fileNames
  return $ GroupMetric "XML-agnostic schema size" [
     Metric "Number of lines" "Lines" (show lines),
     Metric "Number of bytes" "Bytes" (show bytes)
   ]

wordCount :: FilePath -> IO (Int,Int,Int)
wordCount fname = do
  -- let tmpFileName = "/tmp/wordcount.tmp"
  -- let wcCommand = "wc "++fname++" > "++tmpFileName
  -- exitOK $ system wcCommand
  -- result <- readFile tmpFileName
  -- fileHandle <- openFile tmpFileName ReadWriteMode
  (tmpFileName,fileHandle) <- openTempFile "/tmp" "wordcount.tmp"
  -- putStrLn ("Made temporary file: "++tmpFileName)
  processHandle <- runProcess "wc" [fname] Nothing Nothing Nothing (Just fileHandle) Nothing
  exitOK $ waitForProcess processHandle
  -- fileHandle <- openFile tmpFileName ReadMode
  -- result <- hGetContents fileHandle
  -- hClose fileHandle
  result <- readFile tmpFileName
  -- putStrLn (show result)
  let [lns,wrds,bts,_] = words result
  removeFile tmpFileName
  -- putStrLn ("Removed file: "++tmpFileName)
  return (read lns,read wrds,read bts)

-- | Ensure that the given computation exits with an error code
--   that indicates successful execution.       
exitOK :: IO ExitCode -> IO ()
exitOK io
  = do exitCode <- io
       case exitCode of
         (ExitSuccess)   -> return ()
	 (ExitFailure code) -> fail $ "Exit code: " ++ show code


-----------------------------------------------------------------------------
-- * XSD-agnostic schema size metrics

_NODE_ = "NODE"
_Element_ = "Element"
_Attribute_ = "Attribut"
_ANN_ = "ANN"
_Documentation_ = "Document"
_AppInfo_ = "AppInfo"

calcXsdAgnosticMetrics :: [Element] -> GroupMetric
calcXsdAgnosticMetrics elements = GroupMetric "XSD-agnostic schema size" [
   mkMetric "Number of all XML nodes (both attributes and elements)" _NODE_,
   mkMetric "Number of XML elements" _Element_,
   mkMetric "Number of XML attributes" _Attribute_,
   mkMetric "Number of all XML nodes for annotation" _ANN_,
   mkMetric "Number of documentation nodes" _Documentation_,
   mkMetric "Number of appinfo nodes" _AppInfo_
 ]
 where
  mkMetric msg k = Metric msg (showKey k) (get k featureMap)
  get k m = show $ maybe 0 id (Map.lookup k m)

  featureMap = execState (mapM_ (countElem [_NODE_]) elements) Map.empty
  countElem kinds (Elem name attrs contents) = do
    incrementBy (length attrs) _Attribute_
    increment _Element_
    mapM_ (incrementBy (1+length attrs)) kinds
    let kinds' = case (snd $ breakPrefixTag name) of
          "annotation" -> _ANN_:kinds
          "documentation" -> _Documentation_:kinds
          "appinfo" -> _AppInfo_:kinds
          _ -> kinds
    mapM_ (countContent kinds') contents
  countContent kinds (CElem element) = countElem kinds element
  countContent kinds _               = increment _Element_ >> mapM_ increment kinds

-----------------------------------------------------------------------------
-- * XSD-aware counts

_ELg_ = "ELg"
_CTg_ = "CTg"
_STg_ = "STg"
_MGg_ = "MGg"
_AGg_ = "AGg"
_ATg_ = "ATg"
_GLOBAL_ = "GLOBAL"

_ELl_ = "ELl"
_CTl_ = "CTl"
_STl_ = "STl"
_MGl_ = "MGl"
_AGl_ = "AGl"
_ATl_ = "ATl"
_LOCAL_ = "LOCAL"

calcXsdAwareMetrics :: [Xs'schema] -> GroupMetric
calcXsdAwareMetrics schemas = GroupMetric "XSD-aware counts" [
   mkMetric "Number of global element declarations" _ELg_,
   mkMetric "Number of global complex-type definitions" _CTg_,
   mkMetric "Number of global simple-type definitions" _STg_,
   mkMetric "Number of global model-group definitions" _MGg_,
   mkMetric "Number of global attribute-group declarations" _AGg_,
   mkMetric "Number of global attribute declarations" _ATg_,
   mkMetric "Number of global declarations and definitions" _GLOBAL_,
   mkMetric "Number of local element declarations" _ELl_,
   mkMetric "Number of local complex-type definitions" _CTl_,
   mkMetric "Number of local simple-type definitions" _STl_,
   mkMetric "Number of local model-group definitions" _MGl_,
   mkMetric "Number of local attribute-group declarations" _AGl_,
   mkMetric "Number of local attribute declarations" _ATl_,
   mkMetric "Number of local declarations and definitions" _LOCAL_
 ]
 where
  mkMetric msg k = Metric msg (showKey k) (get k featureMap)
  get k m = show $ maybe 0 id (Map.lookup k m)

  globals = concatMap (\(Xs'schema attrs _ glbls) -> glbls) schemas
  featureMap = execState (mapM_ (countGlobal.fst) globals) Map.empty
  
  countGlobal (OneOf7 (x::Xs'simpleType)) = increment _STg_ >> increment _GLOBAL_ >> countLocal x
  countGlobal (TwoOf7 (x::Xs'complexType)) = increment _CTg_ >> increment _GLOBAL_ >> countLocal x
  countGlobal (ThreeOf7 (x::Xs'element)) = increment _ELg_ >> increment _GLOBAL_ >> countLocal x
  countGlobal (FourOf7 (x::Xs'attribute)) = increment _ATg_ >> increment _GLOBAL_ >> countLocal x
  countGlobal (FiveOf7 (x::Xs'attributeGroup)) = increment _AGg_ >> increment _GLOBAL_ >> countLocal x
  countGlobal (SixOf7 (x::Xs'group)) = increment _MGg_ >> increment _GLOBAL_ >> countLocal x
  countGlobal (SevenOf7 (x::Xs'notation)) = nop
  
  countLocal x = applyTU (allTU (\_ _ -> ()) () $ full_tdTU worker) x
  worker = constTU () 
            `adhocTU` (\(_::Xs'simpleType) -> increment _STl_ >> increment _LOCAL_)
            `adhocTU` (\(_::Xs'complexType) -> increment _CTl_ >> increment _LOCAL_)
            `adhocTU` (\(_::Xs'element) -> increment _ELl_ >> increment _LOCAL_)
            `adhocTU` (\(_::Xs'attribute) -> increment _ATl_ >> increment _LOCAL_)
            `adhocTU` (\(_::Xs'attributeGroup) -> increment _AGl_ >> increment _LOCAL_)
            `adhocTU` (\(_::Xs'group) -> increment _MGl_ >> increment _LOCAL_)

-----------------------------------------------------------------------------
-- * McCabe complexity for XSD

_MCC_ = "MCC"

mcCabeComplexityMetrics :: SchemaSet -> GroupMetric
mcCabeComplexityMetrics schemas = GroupMetric "McCabe complexity" [
   mkMetric "McCabe complexity" _MCC_
 ]
 where
  mkMetric msg k = Metric msg (showKey k) (get k featureMap)
  get k m = show $ maybe 0 id (Map.lookup k m)

  featureMap = execState (mapM_ countMcCabeXml05 schemas) Map.empty

-- | We define MCC for XSD as follows:
--   * A choice-model group with n branches contributes n to a schemas MCC.
--   * A schema particle that admits the minOccurs and maxOccurs attributes
--     contributes 1 to a schemas MCC, if the values of the minOccurs and 
--     maxOccurs attributes are distinct. (We recall that the defaults for 
--     these attributes are minOccurs = maxOccurs = 1.)
--   * If h is the name of the head of a substitution group with n > 0 non-abstract
--     participants, potentially counting h itself (if it is non-abstract), then
--     each element reference to h contributes n to a schemas MCC.
--   * If t is the name of a user-defined (as opposed to primitive) type with 
--     n > 0 non-abstract subtypes (including derived types and t itself if it is 
--     non-abstract), then each type reference to t contributes n to a schemas MCC.
--   * Each nillable attribute with value true contributes 1 to a schemas MCC.
--   * The number of root-element declarations is added to a schemas MCC.
-- NOT DONE YET
countMcCabeXml05 :: Schema -> State FeatureMap ()
countMcCabeXml05 schema = applyTU (full_tdTU worker) xsd
 where
  nse = nseSchema schema
  xsd = parsedSchema schema
  worker = constTU () 
            `adhocTU` onChoice_Braches
            `adhocTU` onElementAttrs_Occurs `adhocTU` onGroupAttrs
            `adhocTU` onChoiceAttrs `adhocTU` onSequenceAttrs
            `adhocTU` onAnyAttrs `adhocTU` onAllAttrs
            `adhocTU` onNillable
            
  -- Branches of choice
  onChoice_Braches (Xs'choice _ _ branches) = incrementBy (length branches) _MCC_
  
  -- Occurs attributes
  doOccurs minOccurs maxOccurs attrs
   = if (maybe "1" id $ minOccurs attrs)==(maybe "1" id $ maxOccurs attrs)
      then nop 
      else increment _MCC_
  onElementAttrs_Occurs = doOccurs xs'elementMinOccurs xs'elementMaxOccurs
  onGroupAttrs = doOccurs xs'groupMinOccurs xs'groupMaxOccurs
  onChoiceAttrs = doOccurs xs'choiceMinOccurs xs'choiceMaxOccurs
  onSequenceAttrs = doOccurs xs'sequenceMinOccurs xs'sequenceMaxOccurs
  onAnyAttrs = doOccurs (default2maybe . xs'anyMinOccurs) (default2maybe . xs'anyMaxOccurs)
    where 
      default2maybe (NonDefault a) = Just a
      default2maybe (Default a)    = Just a
  onAllAttrs = doOccurs allMinOccurs allMaxOccurs
    where
      allMinOccurs attrs = maybe Nothing (const $ Just "1") (xs'allMinOccurs attrs) 
      allMaxOccurs attrs = maybe Nothing (const $ Just "1") (xs'allMaxOccurs attrs)

  -- Nillable="true"
  onNillable Xs'element_nillable_false  = nop
  onNillable Xs'element_nillable_true   = increment _MCC_

      
-----------------------------------------------------------------------------
-- * XSD feature metrics

-- | Type of a mapping from feature keys to associated counts.
type FeatureMap = Map Key Int

-- | Feature keys are simply strings.
type Key = String

-- | Constants to use as keys in feature map.
_Sequence_ = "Sequence"
_Choice_ = "Choice"
_All_ = "All"

_Restriction_ = "Restrict"
_List_ = "List"
_Union_ = "Union"
_Pattern_ = "Pattern"
_Enumeration_ = "Enumerat"
_EnumGroup_ = "EnumGrou"
 
_DTD_Occurs_ = "DTDOccur"
_Non_DTD_Occurs_ = "NonDTDOc"
_Nillable_ = "Nillable"
_Fixed_ = "Fixed"
_Default_ = "Default"
_Required_ = "Required"
_Prohibited_ = "Prohibit"
_Redundant_ = "Redundan"
 
_Abstract_ = "Abstract"
_SubstitutionGroup_ = "SubstGrp"
_SimpleTypeRestriction_ = "SiTyRest"
_SimpleContentRestriction_ = "SiCoRest"
_SimpleContentExtension_ = "SiCoExt"
_ComplexContentRestriction_ = "CoCoRest"
_ComplexContentExtension_ = "CoCoExt"
_Redefine_ = "Redefine"
_Block_ = "Block"
_Final_ = "Final"

_Mixed_ = "Mixed"

_Any_ = "Any"
_AnyAny_ = "AnyAny"
_AnyOther_ = "AnyOther"
_AnyTarget_ = "AnyTarg"
_AnyLocal_ = "AnyLocal"
_AnyURL_ = "AnyURL"
_AnyA_ = "AnyA"
_AnyAAny_ = "AnyAAny"
_AnyAOther_ = "AnyAOthe"
_AnyATarget_ = "AnyATarg"
_AnyALocal_ = "AnyALoca"
_AnyAURL_ = "AnyAURL"

_Unique_ = "Unique"
_Key_ = "Key"
_KeyRef_ = "KeyRef"

_FileCount_ = "FileCoun"
_SingleFile_ = "SingleFi"
_MultipleFile_ = "MultiFi"
_NoTarget_ = "NoTarget"
_Chameleon_ = "Chameleo"
_Import_ = "Import"
_Include_ = "Include"

-- | Turn a key into a string.
showKey :: Key -> String
showKey k = k

-- | Calculate XSD feature metrics for a given schema.
calcXsdFeatureMetrics :: SchemaSet -> GroupMetric
calcXsdFeatureMetrics schemas = GroupMetric "XSD feature metrics" [ 
   -- Model group operators
   mkMetric "Number of sequence operators" _Sequence_,
   mkMetric "Number of choice operators"   _Choice_,
   mkMetric "Number of all operators"      _All_,
   
   -- Simple type features
   mkMetric "Number of restriction types"  _Restriction_,
   mkMetric "Number of list types"         _List_,
   mkMetric "Number of union types"        _Union_,
   mkMetric "Number of pattern types"      _Pattern_,
   mkMetric "Number of enumeration constants" _Enumeration_,
   mkMetric "Number of enumeration groups" _EnumGroup_,
   
   -- Occurrence features
   mkMetric "Number of DTD occurs"         _DTD_Occurs_,
   mkMetric "Number of Non-DTD occurs"     _Non_DTD_Occurs_,
   mkMetric "Number of nillable elments"   _Nillable_,
   mkMetric "Number of fixed=\"...\""      _Fixed_,
   mkMetric "Number of default=\"...\""    _Default_,
   mkMetric "Number of use=\"required\""   _Required_,
   mkMetric "Number of use=\"prohibited\"" _Prohibited_,
   mkMetric "Number of specified defaults" _Redundant_,
   
   -- Subtyping features
   mkMetric "Number of abstract elements and types" _Abstract_,
   mkMetric "Number of substitution groups" _SubstitutionGroup_,
   mkMetric "Number of simple-type restrictions" _SimpleTypeRestriction_,
   mkMetric "Number of complex-type extensions; simple content" _SimpleContentExtension_,
   mkMetric "Number of complex-type extensions; complex content" _ComplexContentExtension_,
   mkMetric "Number of complex-type restrictions; simple content" _SimpleContentRestriction_,
   mkMetric "Number of complex-type restrictions; complex content" _ComplexContentRestriction_,
   mkMetric "Number of redefines" _Redefine_,
   mkMetric "Number of blockDefault/block" _Block_,
   mkMetric "Number of finalDefault/final" _Final_,
   
   -- Mixed content   
   mkMetric "Number of mixeds" _Mixed_,

   -- Extensibility (wildcards)
   mkMetric "Number of anys" _Any_,
   mkMetric "Number of any - ##any" _AnyAny_,
   mkMetric "Number of any - ##other" _AnyOther_,
   mkMetric "Number of any - ##targetnamespace" _AnyTarget_,
   mkMetric "Number of any - ##local" _AnyLocal_,
   mkMetric "Number of any - URL" _AnyURL_,
   mkMetric "Number of anyAttributes" _AnyA_,
   mkMetric "Number of anyAttribute - ##any" _AnyAAny_,
   mkMetric "Number of anyAttribute - ##other" _AnyAOther_,
   mkMetric "Number of anyAttribute - ##targetnamespace" _AnyATarget_,
   mkMetric "Number of anyAttribute - ##local" _AnyALocal_,
   mkMetric "Number of anyAttribute - URL" _AnyAURL_,
   
   -- Identity constraints
   mkMetric "Number of uniques" _Unique_,
   mkMetric "Number of keys" _Key_,
   mkMetric "Number of keyrefs" _KeyRef_,
   
   -- Modularization
   Metric "Number of files" _FileCount_ (show fileCount),
   Metric "Number of single-file projects"   _SingleFile_   (show $ if fileCount==1 then 1 else 0),
   Metric "Number of multiple-file projects" _MultipleFile_ (show $ if fileCount>1 then 1 else 0),
   mkMetric "Number of files without target namespace" _NoTarget_,
   Metric "Number of projects that use chameleon includes" _Chameleon_ (show $ chameleon),
   mkMetric "Number of imports" _Import_,
   mkMetric "Number of includes" _Include_
   
 ]
 where
   mkMetric msg k = Metric msg (showKey k) (get k featureMap)
   featureMap = execState (mapM_ extractXsdFeatureMap schemas) Map.empty
   get k m = show $ maybe 0 id (Map.lookup k m)
   get' k m = maybe 0 id (Map.lookup k m)
   fileCount = length schemas
   chameleon = if (get' _Include_ featureMap > 0 && get' _NoTarget_ featureMap > 0) then 1 else 0

-- | Extract a feature map from a given schema.
extractXsdFeatureMap :: Schema -> State FeatureMap ()
extractXsdFeatureMap schema = do
    applyTU (full_tdTU worker) xsd 
    when (Nothing==targetNameSpace nse) (increment _NoTarget_)
  where
      nse = nseSchema schema
      xsd = parsedSchema schema
      worker = constTU () 
                `adhocTU` onSequence `adhocTU` onChoice `adhocTU` onAll 
                `adhocTU` onList `adhocTU` onUnion `adhocTU` onPattern 
                `adhocTU` onEnumeration `adhocTU` onNillable  
                `adhocTU` onRestriction
                `adhocTU` (fuse [onElementAttrs_Occurs,onElementAttrs_Fixed,
                                 onElementAttrs_Default,onElementAttrs_Subst])
                `adhocTU` (fuse [onAttributeAttrs_Fixed,onAttributeAttrs_Default])
                `adhocTU` onGroupAttrs
                `adhocTU` onChoiceAttrs `adhocTU` onSequenceAttrs
                `adhocTU` onAllAttrs `adhocTU` onAnyAttrs
                `adhocTU` onMaxExclusiveAttrs `adhocTU` onMinExclusiveAttrs
                `adhocTU` onMaxInclusiveAttrs `adhocTU` onMinInclusiveAttrs
                `adhocTU` onPrecisionAttrs `adhocTU` onScaleAttrs
                `adhocTU` onLengthAttrs `adhocTU` onMinLengthAttrs `adhocTU` onMaxLengthAttrs
                `adhocTU` onEncodingAttrs `adhocTU` onPeriodAttrs `adhocTU` onDurationAttrs
                `adhocTU` onAttributeUse
                
                `adhocTU` onSchemaElementFormDefault `adhocTU` onSchemaAttributeFormDefault
                `adhocTU` onAnyProcessContents `adhocTU` onAnyAttributeProcessContents
                
                `adhocTU` onElementAbstract `adhocTU` onComplexTypeAbstract
                `adhocTU` onSimpleType `adhocTU` onSimpleContent `adhocTU` onComplexContent
                `adhocTU` onRedefine 
                `adhocTU` onElementAttrs_BlockFinal `adhocTU` onComplexTypeAttrs_BlockFinal
                `adhocTU` onSchemaAttrs_BlockFinal
                
                `adhocTU` onComplexTypeMixed `adhocTU` onComplexContentMixed
                
                `adhocTU` onAnyAttrs_Any 
                `adhocTU` onAnyAttributeAttrs_Any
                
                `adhocTU` onImport `adhocTU` onInclude
                
      -- model group operators
      onSequence (Xs'sequence attrs ann content) = increment _Sequence_
      onChoice (Xs'choice attrs ann content)     = increment _Choice_
      onAll (Xs'all attrs ann content)           = increment _All_
      
      -- simple type features
      onRestriction (Xs'restriction attrs ann (TwoOf2 (_,content)) _) = do
        when (foldr (\c b -> b || isSixOf13 c) False content) (increment _EnumGroup_)
        increment _Restriction_
      onRestriction _ = increment _Restriction_
      onList (Xs'list attrs ann content)         = increment _List_
      onUnion (Xs'union attrs ann content)       = increment _Union_
      onPattern (Xs'pattern attrs ann)           = increment _Pattern_
      onEnumeration (Xs'enumeration attrs ann)   = increment _Enumeration_
      isSixOf13 (SixOf13 (e::Xs'enumeration)) = True
      isSixOf13 _                             = False
      
      -- DTD and non-DTD occurs
      onElementAttrs_Occurs = doOccurs xs'elementMinOccurs xs'elementMaxOccurs
      onGroupAttrs = doOccurs xs'groupMinOccurs xs'groupMaxOccurs
      onChoiceAttrs = doOccurs xs'choiceMinOccurs xs'choiceMaxOccurs
      onSequenceAttrs = doOccurs xs'sequenceMinOccurs xs'sequenceMaxOccurs
      doOccurs minOccurs maxOccurs attrs
        = case (minOccurs attrs, maxOccurs attrs) of
            (Just "0", Just "unbounded") -> increment _DTD_Occurs_
            (Just "1", Just "unbounded") -> increment _DTD_Occurs_ >> increment _Redundant_
            (Just "0", Just "1")         -> increment _DTD_Occurs_ >> increment _Redundant_
            (Just "1", Just "1")         -> increment _DTD_Occurs_ >> incrementBy 2 _Redundant_
            (Nothing,Nothing)            -> nop
            (Just "1",_)                 -> increment _Non_DTD_Occurs_ >> increment _Redundant_
            (_,Just "1")                 -> increment _Non_DTD_Occurs_ >> increment _Redundant_
            _                            -> increment _Non_DTD_Occurs_
      onAnyAttrs = doOccurs (default2maybe . xs'anyMinOccurs) (default2maybe . xs'anyMaxOccurs)
      default2maybe (NonDefault a) = Just a
      default2maybe (Default a)    = Just a
      onAllAttrs = doOccurs allMinOccurs allMaxOccurs
      allMinOccurs attrs = maybe Nothing (const $ Just "1") (xs'allMinOccurs attrs) 
      allMaxOccurs attrs = maybe Nothing (const $ Just "1") (xs'allMaxOccurs attrs)
      
      -- nillable="true"
      onNillable Xs'element_nillable_false  = increment _Redundant_
      onNillable Xs'element_nillable_true   = increment _Nillable_

      -- fixed="..." and specified default fixed="true"
      onMaxExclusiveAttrs = doFixed xs'maxExclusiveFixed Xs'maxExclusive_fixed_true 
      onMinExclusiveAttrs = doFixed xs'minExclusiveFixed Xs'minExclusive_fixed_true 
      onMaxInclusiveAttrs = doFixed xs'maxInclusiveFixed Xs'maxInclusive_fixed_true 
      onMinInclusiveAttrs = doFixed xs'minInclusiveFixed Xs'minInclusive_fixed_true 
      onPrecisionAttrs    = doFixed xs'precisionFixed Xs'precision_fixed_true 
      onScaleAttrs        = doFixed xs'scaleFixed Xs'scale_fixed_true 
      onLengthAttrs       = doFixed xs'lengthFixed Xs'length_fixed_true 
      onMinLengthAttrs    = doFixed xs'minLengthFixed Xs'minLength_fixed_true 
      onMaxLengthAttrs    = doFixed xs'maxLengthFixed Xs'maxLength_fixed_true 
      onEncodingAttrs     = doFixed xs'encodingFixed Xs'encoding_fixed_true 
      onPeriodAttrs       = doFixed xs'periodFixed Xs'period_fixed_true 
      onDurationAttrs     = doFixed xs'durationFixed Xs'duration_fixed_true 
      doFixed fixed true attrs = case fixed attrs of
        Nothing      -> nop
        (Just x)     -> increment _Fixed_ >>
                        when (x/=true) (increment _Redundant_)
      onElementAttrs_Fixed   = doFixed xs'elementFixed "true"
      onAttributeAttrs_Fixed = doFixed xs'attributeFixed "true"
        
      -- default="..."  
      onElementAttrs_Default attrs = doDefault xs'elementDefault attrs
      onAttributeAttrs_Default attrs = doDefault xs'attributeDefault attrs
      doDefault dflt attrs = if (dflt attrs)==Nothing then nop else increment _Default_
      
      -- use="required" and use="prohibited" and specified default use="optional"
      onAttributeUse attr = case attr of
        Xs'attribute_use_required   -> increment _Required_ 
        Xs'attribute_use_prohibited -> increment _Prohibited_
        Xs'attribute_use_optional   -> increment _Redundant_

      -- specified defaults (incomplete)
      {-
        * attr(finalDefault,optional,['']).
        * attr(blockDefault,optional,['']).
        * attr(attributeFormDefault,optional,[unqualified]).
        * attr(elementFormDefault,optional,[unqualified]).
        * attr(minOccurs,optional,['1']).
        * attr(maxOccurs,optional,['1']).
        * attr(use,optional,[optional]).
        * attr(mixed,optional,[false]).
        * attr(abstract,optional,[false]).
        * attr(nillable,optional,[false]).
        * attr(namespace,optional,['##any']).
        * attr(processContents,optional,[strict]).
        * attr(fixed,optional,[false]).
      -}
      onSchemaElementFormDefault (NonDefault Xs'schema_elementFormDefault_unqualified)
        = increment _Redundant_
      onSchemaElementFormDefault _ = nop
      onSchemaAttributeFormDefault (NonDefault Xs'schema_attributeFormDefault_unqualified)
        = increment _Redundant_
      onSchemaAttributeFormDefault _ = nop
      onAnyProcessContents (NonDefault Xs'any_processContents_strict)
        = increment _Redundant_
      onAnyProcessContents _ = nop
      onAnyAttributeProcessContents (NonDefault Xs'anyAttribute_processContents_strict)
        = increment _Redundant_
      onAnyAttributeProcessContents _ = nop
      
      
      -- abstract elements and types
      onElementAbstract Xs'element_abstract_true  = increment _Abstract_
      onElementAbstract Xs'element_abstract_false = increment _Redundant_
      onComplexTypeAbstract Xs'complexType_abstract_true  = increment _Abstract_
      onComplexTypeAbstract Xs'complexType_abstract_false = increment _Redundant_
      
      -- substitution groups
      onElementAttrs_Subst attrs 
        = if (xs'elementSubstitutionGroup attrs)==Nothing 
            then nop 
            else increment _SubstitutionGroup_
    
      -- simple type restriction
      onSimpleType (Xs'simpleType attrs ann (OneOf3 (_::Xs'restriction))) 
        = increment _SimpleTypeRestriction_
      onSimpleType _ = nop
      -- complex type restriction and extension
      onSimpleContent (Xs'simpleContent attrs ann (OneOf2 (_::Xs'restriction)))
        = increment _SimpleContentRestriction_ 
      onSimpleContent (Xs'simpleContent attrs ann (TwoOf2 (_::Xs'extension)))
        = increment _SimpleContentExtension_
      onComplexContent (Xs'complexContent attrs ann (OneOf2 (_::Xs'restriction)))
        = increment _ComplexContentRestriction_ 
      onComplexContent (Xs'complexContent attrs ann (TwoOf2 (_::Xs'extension)))
        = increment _ComplexContentExtension_
    
      -- redefine
      onRedefine (_::Xs'redefine) = increment _Redefine_
      
      -- blockDefault / block and finalDefault / final
      onElementAttrs_BlockFinal attrs
        = doBlock (xs'complexTypeBlock attrs) (nop) >>
          doFinal (xs'complexTypeFinal attrs) (nop)
      onComplexTypeAttrs_BlockFinal attrs 
        = doBlock (xs'complexTypeBlock attrs) (nop) >>
          doFinal (xs'complexTypeFinal attrs) (nop)
      onSchemaAttrs_BlockFinal attrs
        = doBlock (default2maybe $ xs'schemaBlockDefault attrs) (increment _Redundant_) >>
          doFinal (default2maybe $ xs'schemaFinalDefault attrs) (increment _Redundant_)      
      doBlock (Just "") countDefault = increment _Block_ >> countDefault
      doBlock (Just _)  countDefault = increment _Block_
      doBlock _         countDefault = nop
      doFinal (Just "") countDefault = increment _Final_ >> countDefault
      doFinal (Just _)  countDefault = increment _Final_
      doFinal _         countDefault = nop
      
      -- mixed and specified default mixed="false"
      onComplexTypeMixed (Default (_::Xs'complexType_mixed)) = nop
      onComplexTypeMixed (NonDefault Xs'complexType_mixed_true) = increment _Mixed_
      onComplexTypeMixed (NonDefault Xs'complexType_mixed_false)
        = increment _Mixed_ >> increment _Redundant_
      onComplexContentMixed (_::Xs'complexContent_mixed) = increment _Mixed_

      -- wildcards
      onAnyAttrs_Any attrs = increment _Any_ >> doAny (xs'anyNamespace attrs)
      doAny (Default _) = nop
      doAny (NonDefault "##any") = increment _AnyAny_ >> increment _Redundant_
      doAny (NonDefault "##other") = increment _AnyOther_
      doAny (NonDefault "##targetnamespace") = increment _AnyTarget_
      doAny (NonDefault "##local") = increment _AnyLocal_
      doAny (NonDefault _) = increment _AnyURL_
      onAnyAttributeAttrs_Any attrs = increment _AnyA_ >> doAnyA (xs'anyAttributeNamespace attrs)
      doAnyA (Default _) = nop
      doAnyA (NonDefault "##any") = increment _AnyAAny_ >> increment _Redundant_
      doAnyA (NonDefault "##other") = increment _AnyAOther_
      doAnyA (NonDefault "##targetnamespace") = increment _AnyATarget_
      doAnyA (NonDefault "##local") = increment _AnyALocal_
      doAnyA (NonDefault _) = increment _AnyAURL_
      
      onUnique (_::Xs'unique) = increment _Unique_
      onKey (_::Xs'key) = increment _Key_
      onKeyRef (_::Xs'keyref) = increment _KeyRef_
      
      onImport (_::Xs'import) = increment _Import_
      onInclude (_::Xs'include) = increment _Include_
                             
-- | Fuse several monadic functions into a single one.
--   Monadic bind is used for sequencing.
fuse :: Monad m => [a -> m ()]-> a -> m ()
fuse fs a = sequence_ $ Prelude.map ($ a) fs

-- | Do nothing.
nop :: Monad m => m ()
nop = return ()

-- | Increment a counter map in a state monad. 
increment :: (Ord k, MonadState (Map k a) m, Num a) => k -> m ()
increment key
  = modify inc
    where
      inc m = if Map.member key m 
                then Map.adjust (+1) key m
                else Map.insert key 1 m

-- | Increment a counter map in a state monad by a given number.
incrementBy :: (Ord k, MonadState (Map k a) m, Num a) => a -> k -> m ()
incrementBy n key
  = modify inc
    where
      inc m = if Map.member key m 
                then Map.adjust (+n) key m
                else Map.insert key n m

-----------------------------------------------------------------------------
-- * Helpers that deal with file and directory names.

-- | Removes redundant superdir segments from
--   a file path.
simplifyFileName :: FilePath -> FilePath
simplifyFileName fname = concat $ sepBy "/" $ simplify [] segments
 where
  segments = breakFile fname
  simplify (a:as) ("..":bs) = simplify as bs
  simplify as     (b:bs)    = simplify (b:as) bs
  simplify as     []        = reverse as
  sepBy sep [] = []
  sepBy sep [x] = [x]
  sepBy sep (x:xs) = x:sep:(sepBy sep xs)

-- | Obtain the directory part of a file name.
noBasename :: FilePath -> FilePath
noBasename fname
  =  reverse (dropWhile (/='/') (reverse  fname))

-- | Obtain the file name without extension.
noExtension :: FilePath -> FilePath
noExtension fname
  = reverse (pop $ dropWhile (/='.') (reverse  fname))
    where pop [] = []
          pop (_:xs) = xs

-- | Like 'noExtension' except that it also works on file names
--   that do not have an extension to begin with.
noExtension' :: FilePath -> FilePath
noExtension' fname
 = case reverse (breakFile fname) of
     (seg:segs) 
       | '.' `elem` fname -> segs2fname $ reverse ((noExtension seg):segs)
       | otherwise        -> fname
     _                    -> fname   
   where 
     segs2fname []       = ""
     segs2fname [sg]     = sg
     segs2fname (sg:sgs) = sg++"/"++(segs2fname sgs)

-- | Remove the protocol prefix from a filename.
noProtocol :: FilePath -> FilePath
noProtocol ('h':'t':'t':'p':':':'/':'/':fname) = fname
noProtocol fname = fname

-- | break a file name into its segments, i.e. the pieces between slashes.    
breakFile :: FilePath -> [String]
breakFile = breakWith (=='/')
       
-- | Break path argument into directory names. i.e. the pieces between colons.
breakPath :: String -> [FilePath]
breakPath = breakWith (==':')

-- | Break a list into segments, using a predicate on elements to indicate break points.
--   The break points themselves will be discarded.
breakWith :: (a -> Bool) -> [a] -> [[a]] 
breakWith f x = case break f x of
	(p,(_:pp)) -> p: breakWith f pp
	(p,[]) -> [p]
  
-----------------------------------------------------------------------------


 

