-----------------------------------------------------------------------------
-- |
-- Copyright   :  (c) Universidade do Minho, 2005
-- License     :  LGPL
-- 
-- Maintainer  :  Joo Saraiva - jas@di.uminho.pt
-- Stability   :  experimental
-- Portability :  portable
--
-- Table Combinators
--
-----------------------------------------------------------------------------


module PrettyPrint.Tables.Combinators 
       ( renderAscii
       , renderLaTeX
       , renderHtml
       , (<->)
       , (<|>) 
       , text
       , empty
       , hlist
       , hlist'
       , vlist
       , vlist'
       , matrix
       , asciiT'
       , matrix2table               -- to be rewritten
       , matrix2table'              -- to be rewritten
       , matrix2table''             -- to be rewritten
       , matrix2table'''            -- to be rewritten
       ) where

-----------------------------------------------------------------------------

import PrettyPrint.Tables.Data_Lazy
import PrettyPrint.Tables.Visfun_Lazy


infixr 4 <->
infixr 5 <|>


-----------------------------------------------------------------------------
-- * Building concrete tables

-- | Rendering the Table in ASCII


renderAscii :: PPTable                -- ^ Table
            -> IO ()
renderAscii t = putStrLn t'
  where t' = asciiT' (C_ConsPPTableR_1 t)

asciiT' ::  PPTableR              -- ^ Table
        -> [Char]                 -- ^ Table in Ascii 
asciiT' t =  showT ascii
  where (ascii,html,latex,vrml) = lrcEval t


asciiT t =  putStrLn $ asciiT' t      


-- | Rendering the Table in LaTeX


renderLaTeX :: PPTable               -- ^ Table
            -> IO ()
renderLaTeX t = putStrLn t'
  where t' = latexT' (C_ConsPPTableR_1 t)


latexT' :: PPTableR                  -- ^ Table
        -> String                    -- ^ Table in LaTeX 

latexT' t = show latex
  where (ascii,html,latex,vrml) = lrcEval t


latexT = putStrLn . latexT' 



-- | Rendering the Table in HTML


renderHtml :: PPTable               -- ^ Table
           -> IO ()
renderHtml t = putStrLn t'
  where t' = htmlT' (C_ConsPPTableR_1 t)


htmlT' :: PPTableR                  -- ^ Table
       -> [Char]                    -- ^ Table in Html
htmlT' t = showHtml html
  where (ascii,html,latex,vrml) = lrcEval t


htmlT = putStrLn . htmlT'


showT [] = ""
showT(h:t) = h ++ "\n" ++ showT t 

-----------------------------------------------------------------------------
-- * Basic Combinators


-- | Combining tables horizontally

(<->)  :: PPTable -> PPTable -> PPTable
t <-> b = C_AboveT_1 t b


-- | Combining tables vertically

(<|>) :: PPTable -> PPTable -> PPTable
l <|> r = C_BesidesT_1 l r


-- | Constructs a table from a text

text :: (Show t) => t -> PPTable
text t =  C_OneT_1 $ 
               C_RootTable_1 [C_AttrAny_1 "border" "true"] 
                   [(C_OneRow_1 [C_TableText_1 [] (C_CText_1 (show t))]  )]

-- | Constructs an empty text

empty :: PPTable
empty =  C_OneT_1 $ C_RootTable_1 [] []


-----------------------------------------------------------------------------
-- * More Advanced Combinators


-- | Constructs an horizontal table (ie, one row table) from a list

hlist :: (Show t) => [t] -> PPTable
hlist = foldr ((<|>) . text) empty 

hlist' :: [PPTable] -> PPTable
hlist' = foldr (<|>) empty 


-- | Constructs a vertical table (ie, one column table) from a list

vlist :: (Show t) => [t] -> PPTable
vlist = foldr ((<->) . text) empty 

vlist' :: [PPTable] -> PPTable
vlist' = foldr (<->) empty 



vlist'' :: (Show a) => [a] -> PPTable
vlist'' l = C_OneT_1 $ C_RootTable_1 [C_AttrAny_1 "border" "true",C_Align_1 "left"] l'
  where l' = [ C_OneRow_1 [C_TableText_1 [] (C_CText_1 (show e))]
             | e <- l 
             ] 


-- | Constructs matrix from a listof lists

matrix :: (Show t) => [[t]] -> PPTable
matrix m = vlist' $ map hlist m


-----------------------------------------------------------------------------
-- * Examples

ex1 :: PPTable
ex1 = (text "Joao") <-> (text "Saraiva")

ex2 = text "TABLE"


ex3 = ex2  <|> ex1 <|> ex2 <|> ex1 <|> ex2 

ex4 = ex2  <-> ex1 <-> ex2 <-> ex1 <-> ex1 

ex5 = ex4 <-> ex3 <-> ex4 <|> ex3


ex6 = vlist [1,2,3,4,5,7,2]
ex7 = hlist [1,2,3,4,5,7,2222]

ex8 = hlist' [ex6,ex7]

a = [ ["This","is","the","first","row"]
    , ["and","the","","second"]
    ]

ex9 = matrix a


ex91 = matrix [ ["Table"] ]

ex10 = ex8 <|> ex9

ex11 = matrix [["TABLE1"]] <|> (matrix [ ["TABLE2"] , ["TABLE2"]])



{-
a = "aaaa"  <|> "bbbb" <|> "cccc" <->
    "eeeee" <|> "gggg" <|> "aaaa" 
-}

list2VTable :: Show a => [a] -> Table
list2VTable l = C_RootTable_1 [C_AttrAny_1 "border" "true",C_Align_1 "left"] (list2VTable' l)
  where list2VTable' []    = []
        list2VTable' (h:t) = (C_OneRow_1 [C_TableText_1 [] (C_CText_1 (show h))]) : (list2VTable' t)



-- | Compute a 'Table' from a matrix of strings 
   
matrix2table :: [[String]] ->  PPTableR 
matrix2table m = C_ConsPPTableR_1 $ C_OneT_1 $ C_RootTable_1 [C_AttrAny_1 "border" "true"] c
 where c = map row m

       row a = C_OneRow_1 (map elem a)

       elem a = C_TableText_1 [] (C_CText_1 a)
                  
{-
Tables> asciiT $ matrix2table a
|---------------------------------|
|"This"|"is" |"the"|"first" |"row"|
|---------------------------------|
|"and" |"the"|""   |"second"|     |
|---------------------------------| 
-}


a' = [ [["This","is","the"],["first","row"]] 
     , [] 
     , [["and","the","","second"],["a"]]
     , [["bb","cc"],["aa"]]
     ]


matrix2tableList :: Show a => [[[a]]] -> PPTable 
matrix2tableList m = C_OneT_1 (C_RootTable_1 [C_AttrAny_1 "border" "true" 
                                               ,C_Align_1 "right"] c)
 where c = map row m

       row a = C_OneRow_1 (map elem a)

       elem a = C_NestedTable_1 (list2VTable a)
            


h = [["aa"],["bb","bbb"],["ccc"]]

b = [ (["xxx","yyyy"], [ ["a" ] , ["b","c" ] , ["d"]  ])
    , (["xxx","yyyy"], [ ["a" ] , ["b","c" ] , ["d"]  ])
    ]


-- matrix2tableList' :: [a] -> [(b,[c])] -> PPTable 
matrix2tableList' header body = matrix2tableList body''
  where body' = map (\(a,b) -> a : b) body
        body'' = ([[]] : header) : body'




header = ["mon","tue","wed","thu","fri","sat","sun"]
left   = ["", "9:00","10:00","11:00","12:00"]

body   = [ [ "fp", "" ,"pl","fp","afp","","" ]
         , [ "", "afp" ,"pl","fp","fp","","" ]
        ]




matrix2table' :: [String]                   -- ^ header
              -> [String]                   -- ^ first column
              -> [[String]]                 -- ^ other columns
              -> PPTableR                   -- ^ Table

matrix2table' header left body =  matrix2table matrix'
  where body'  = header : body
        left'  = left
        matrix = zip left' body'
        matrix' = map (\(a,b) -> a:b) matrix 
          


body' = [ ("9:00" ,[ "fp", "" ,"pl","fp","afp","","" ])
        , ("10:00",[ "", "afp" ,"pl","fp","fp","","" ])
        ]



matrix2table'' :: (Show a, Show b) 
               => [(a, [b])]              -- ^ Rows: List of (fst column,list of remaining columns)
               -> PPTableR                -- ^ Table

matrix2table'' body =  matrix2table body'
  where body' = map (\(a,b) -> (show a) : (map show b)) body






-- | Builds a tables given the header and the body of the table.
--           aaaaaaaaaaaaaa
--        (b,cccccccccccccc)
--        (b,cccccccccccccc)
 

matrix2table''' :: (Show a, Show b, Show c)
                => [a]                       -- ^ header
                -> [(b, [c])]                -- ^ Rows: List of (fst column,list of remaining columns)
                -> PPTableR                  -- ^ Table
matrix2table''' header body =  matrix2table body''
  where body' = map (\(a,b) -> (show a) : (map show b)) body
        body'' = ("" : (map show header) ) : body'

         

