

import Control.Monad
import Debug.Trace

data Type a where
  Int :: Type Int
  Prod :: Type a -> Type b -> Type (a,b)
  Either :: Type a -> Type b -> Type (Either a b)
  Tag :: String -> Type a -> Type a
  Ref :: String -> Type DynType
  RefDyn :: String -> Type Dynamic
  RefDyn' :: String -> Type a -> Type Dynamic
  LRef :: n -> Type (LDynamic n)
  CLRef :: cxt -> n -> Type (CLDynamic cxt n)
--  RefEq :: String -> (DynType -> Type x -> Maybe (Equals x x)) -> Type x
--   DT  :: DynType -> Type DynType

instance Show (Type a) where
  show Int = "Int"
  show (Prod a b) = "(Prod "++show a++" "++show b++")"
  show (Either a b) = "(Either "++show a++" "++show b++")"
  show (Tag s a) = "(Tag "++show s++" "++show a++")"
  show (Ref s) = "(Ref "++show s++")"
  show (RefDyn s) = "(RefDyn "++show s++")"

data Dynamic where
  Dyn :: Type a -> a -> Dynamic

{- IDEA
   - Identify types by static label (number) in context (hlist)
   - Allow references to types by label in type representation
   - Represented type is dynamic where the hidden type is constrained
     to be the type selected by the label
   - See Oleg's tags.
-}
data LDynamic n where
  LDyn :: n -> Type a -> a -> LDynamic n

data CLDynamic cxt n where
  CLDyn :: ElementAt n cxt (Type a) => cxt -> n -> Type a -> a -> CLDynamic cxt n
  
mkLDyn :: ElementAt n cxt (Type a) => cxt -> n -> Type a -> a -> LDynamic n
mkLDyn cxt n t a = LDyn n t a
  
instance Show Dynamic where
  show (Dyn t a) = gshow t a

gshow :: Type a -> a -> String
gshow Int i = show i
gshow (Prod a b) (x,y) = "("++gshow a x++","++gshow b y++")"
gshow (Either a b) (Left x) = "(Left "++gshow a x++")"
gshow (Either a b) (Right y) = "(Right "++gshow b y++")"
gshow (Tag s a) x = gshow a x
gshow (RefDyn s) d = show d

unDyn :: Type a -> Dynamic -> Maybe a
unDyn a (Dyn b x) = do
  Eq <- teq a b
  return x

unsafeUnDyn :: Type a -> Dynamic -> a
unsafeUnDyn a d@(Dyn a' _) = maybe err id $ unDyn a d
  where
    err = error $ "unsafeUnDyn: "++show d++" does not have type "++show a++" but type "++show a'
  
data Equals a b where
  Eq :: Equals a a

instance Show (Equals a b) where
  show = eqShow 
  
eqShow :: Equals a b -> String
eqShow Eq = "Eq"
  
--  Eqable :: (a->x) -> (b->x) -> Equals a b
    
teq :: Type a -> Type b -> Maybe (Equals a b)
teq Int Int = return Eq
teq (Prod a b) (Prod c d) = do
  Eq <- teq a c
  Eq <- teq b d
  return Eq
teq (Either a b) (Either c d) = do
  Eq <- teq a c
  Eq <- teq b d
  return Eq
-- teq (Tag s a) (Tag t b) = do
--   guard (s==t)
--   teq a b
teq (Tag s a) b = teq a b
teq a (Tag t b) = teq a b
teq (Ref s) (Ref t)
  | s==t      = return Eq -- fail "Cannot prove equality of referenced type!" 
  | otherwise = mzero
teq (RefDyn s) (RefDyn t)
  | s==t      = return Eq
  | otherwise = mzero
teq _ _ = mzero

resolveRef :: String -> Type a -> [DynType]
resolveRef _ Int = []
resolveRef r (Prod a b) = resolveRef r a ++ resolveRef r b
resolveRef r (Either a b) = resolveRef r a ++ resolveRef r b
resolveRef r (Tag s a)
  | r==s      = [DynType a]
  | otherwise = resolveRef r a
resolveRef _ (Ref _) = []
resolveRef _ (RefDyn _) = []

unroll :: Type a -> Maybe DynType
unroll t = unroll' t t
 
unroll' :: Type x -> Type a -> Maybe DynType
unroll' cxt t = unroll'' [] t
 where
  unroll'' :: [String] -> Type a -> Maybe DynType
  unroll'' done Int = return $ DynType Int
  unroll'' done (Tag s a) = do
    (DynType a') <- unroll'' (s:done) a
    return $ DynType (Tag s a') 
  unroll'' done (Ref s)
   | s `elem` done = return $ DynType (Ref s)
   | otherwise = case resolveRef s cxt of
       [DynType a] -> unroll'' done (Tag s a)
       []          -> error $ "Dangling reference: "++s 
       _           -> error $ "Ambiguous reference: "++s
  unroll'' done (Prod a b) = do
    (DynType a') <- unroll'' done a
    (DynType b') <- unroll'' done b
    return $ DynType (Prod a' b') 
  unroll'' done (Either a b) = do
    (DynType a') <- unroll'' done a
    (DynType b') <- unroll'' done b
    return $ DynType (Either a' b') 
    
unrollView :: Type x -> Type a -> Maybe (View (Type a))
unrollView cxt t = aux [] t
 where
  aux :: [String] -> Type a -> Maybe (View (Type a))
  aux done Int = return $ View id id Int
  aux done (Tag s a) = do
    (View t f a') <- aux (s:done) a
    return $ View t f (Tag s a')
  aux done (RefDyn s)
    | s `elem` done = return $ View id id (RefDyn s)
    | otherwise = case resolveRef s cxt of
        [DynType a] -> do
          (View t f a') <- aux done (Tag s a)
          return $ View (t . unsafeUnDyn a) (Dyn a . f) a'
  aux done (Prod a b) = do
    (View t f a') <- aux done a
    (View t' f' b') <- aux done b
    return $ View (t >< t') (f >< f') (Prod a' b')
  aux done (Either a b) = do
    (View t f a') <- aux done a
    (View t' f' b') <- aux done b
    return $ View (t -|- t') (f -|- f') (Either a' b')
  aux done (Ref s) = error $ "Shouldn't use unrollView on ref: "++s

(><) f g (a,b) = (f a, g b)
(-|-) f g (Left a) = Left (f a)
(-|-) f g (Right b) = Right (g b)

data View a where
  View :: (a->b) -> (b->a) -> Type b -> View (Type a)
   
   
data DynType where
  DynType :: Type a -> DynType
  
instance Show DynType where
  show (DynType a) = "(DynType "++show a++")"
  

teq' :: Type x -> Type a -> Type b -> Maybe (Equals a b)
teq' cxt a b = do
    (DynType a') <- unroll' cxt a
    (DynType b') <- unroll' cxt b
    undefined -- teq a' b'
  where
    teq :: Type a -> Type b -> Maybe (Equals a b)
    teq Int Int = return Eq
    teq (Prod a b) (Prod c d) = do
      Eq <- teq a c
      Eq <- teq b d
      return Eq
    teq (Either a b) (Either c d) = do
      Eq <- teq a c
      Eq <- teq b d
      return Eq
    teq (Tag s a) (Tag t b) = do
      guard (s==t)
      teq a b
    teq (Ref s) (Ref t)
      | s==t      = return Eq -- error "Cannot prove equality of referenced type!" 
      | otherwise = mzero
    teq (Ref s) b = case resolveRef s cxt of
      [a]          -> error $ "Close, but no sigar!" --- teq a b
      []           -> error $ "Cannot resolve reference: "++s
      _            -> error $ "Ambiguous reference: "++s
--    teq (RefEq s f) b = case resolveRef s cxt of
--      [a]          -> f a b
--      []           -> error $ "Cannot resolve reference: "++s
--       _            -> error $ "Ambiguous reference: "++s
    teq _ _ = mzero

  

data Typed a = Type a :| a

-- instance Eq a => Eq (Typed a) where
--   (_ :| y) == (_ :| w) = y==w

instance Eq (Typed a) where
  (x :| y) == (_ :| w) = gcompare x y w == EQ
  
gcompare :: Type a -> a -> a -> Ordering
gcompare = undefined
  
  
data A = A1 B | A2 Int
data B = B1 A

a = Tag "A" (Either (Tag "A1" (RefDyn "B")) (Tag "A2" Int))
b = Tag "B" (Tag "B1" (RefDyn "A"))
ab = Prod a b

v = a :| Left (Dyn b (Dyn a (Right 2)))

mkDyn :: Type cxt -> String -> Type a -> a -> Dynamic
mkDyn cxt s t v = case resolveRef s cxt of
       [DynType t'] -> maybe (error $ "Type from context does not match given type")
                        (\Eq -> Dyn t v)
                        $ teq t t'
       []          -> error $ "Dangling reference: "++s 
       _           -> error $ "Ambiguous reference: "++s


-- a' = Tag "A" (Either (Tag "A1" (RefDyn' "B" b')) (Tag "A2" Int))
-- b' = Tag "B" (Tag "B1" (RefDyn' "A" a'))

-- a' = Tag "A" (Either (Tag "A1" (CLRef ab' (S Z))) (Tag "A2" Int))
-- b' = Tag "B" (Tag "B1" (CLRef ab' Z))
-- ab' = (a',(b',()))

{- Occurs check problem
abCLRef = (aCLRef,(bCLRef,()))

aCLRef = Tag "A" (Either (Tag "A1" (CLRef abCLRef (S Z))) (Tag "A2" Int))
bCLRef = Tag "B" (Tag "B1" (CLRef abCLRef Z))
-}

{-  
class WFCxt cxt
instance WFCxt' cxt cxt => WFCxt cxt
class WFCxt' cxt cxt'
instance (
    WFType cxt e,
    WFCxt' cxt l
  ) => WFCxt' cxt (e,l)
instance (
    WFType cxt e
  ) => WFCxt' cxt (e,())
class WFType cxt t
instance WFType cxt (Type Int)
instance WFType cxt (Type (CLDynamic cxt n))
instance (
  WFType cxt (Type a),
  WFType cxt (Type b)
  ) => WFCxt (Type (a,b))
instance (
  WFType cxt (Type a),
  WFType cxt (Type b)
  ) => WFCxt (Type (Either a b))

enforceWFCxt :: WFCxt cxt => cxt -> cxt
enforceWFCxt = id
-}

{-
refSafe :: String -> Typed a -> Typed Dynamic 
refSafe s (t :| v) = (RefDyn' s t :| Dyn t v)

v' = a' :| Left (Dyn b' (Dyn a' (Right 2)))
-}

data Ok a = Ok a

-- REF hides the type of the value
data REF a where
  REF :: Type a -> x -> REF a

-- This constructor guarantees that the hidden type of the value matches the type representation.
newRef :: Type a -> a -> REF a
newRef t a = REF t a


forth :: View (Type a) -> a -> Dynamic
forth (View t _ b) a = Dyn b (t a)


tst :: Type x -> Type a -> Type b -> Maybe String
tst cxt x y = do
  (DynType x') <- unroll' cxt x
  (DynType y') <- unroll' cxt y
  Eq <- teq x' y'
  return "OK"

tstVw :: Type x -> Type a -> Type b -> Maybe String
tstVw cxt x y = do
  (View _ _ x') <- unrollView cxt x
  (View _ _ y') <- unrollView cxt y 
  Eq <- teq x' y'
  return "OK"

unView :: View a -> DynType
unView (View _ _ a) = DynType a



--- A bit of HList

class ElementAt idx l e | idx l -> e where elementAt :: idx -> l -> e
instance ElementAt Z (e,l) e         where elementAt _ (e,_) = e
instance ElementAt n l e
      => ElementAt (S n) (e',l) e    where elementAt (S n) (_,l) = elementAt n l

data Z = Z
newtype S n = S n



---

data CType cxt a where
  CInt :: CType cxt Int
  CProd :: CType cxt a -> CType cxt b -> CType cxt (a,b)
  CEither :: CType cxt a -> CType cxt b -> CType cxt (Either a b)
  CTag :: String -> CType cxt a -> CType cxt a
  CCLRef :: n -> CType cxt (CLDynamic cxt n)

a' = CTag "A" (CEither (CTag "A1" (CCLRef (S Z))) (CTag "A2" CInt))
b' = CTag "B" (CTag "B1" (CCLRef Z))
ab' = (a',(b',()))

enforceCxt :: cxt -> CType cxt a -> CType cxt a
enforceCxt cxt = id

{- Occurs check problem again!!
ab'' = (a'',(b'',()))
a'' = enforceCxt ab'' a'
b'' = enforceCxt ab'' b'
-}
